/*
	Wabbitcode - An IDE for the Z80 line of TI calculators
	Copyright (C) 2007 Donald Straney

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef G_OS_WIN32
#include <windows.h>
#define ASSEMBLER_NAME "spasm.exe"
#else
#include <sys/wait.h>
#define ASSEMBLER_NAME "spasm"
#endif

#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include "asmide.h"
#include "file.h"
#include "project.h"

//assembly status strings
#define ASSEMBLY_STOPPED_STR "Assembly was canceled\n"
#define ASSEMBLY_UNKNOWN_STR "Assembly status unknown\n"
#define ASSEMBLY_START_STR "Assembling %s...\n"
#define SYMTABLE_START_STR "Creating symbol table for %s...\n"
#define LISTING_START_STR "Creating code listing for %s...\n"
#define STAT_START_STR "Generating stats for %s...\n"
#define COUNT_START_STR "Counting code...\n"
#define COMMAND_RUN_STR "Running '%s'...\n"
#define COMMAND_SUCCESS_STR "Command succeeded\n"
#define COMMAND_FAILURE_STR "Command failed\n"
//size of output buffer for getting assembler output
#define OUTPUT_BUF_SIZE 100
//max args for command in build step
#define MAX_CMD_ARGS 100

//PID for assembler, if running
GPid asm_pid;
//output pipes from assembler
GIOChannel *asm_output, *asm_err;
//name of temp file for code counter
gchar *tempfilename;
//selected range before code counter was run
GtkTextIter selected_startiter, selected_enditer;
//current build step, or NULL if none
build_step *curr_build_step;
//set to TRUE if assembler was killed manually
gboolean assembler_killed;
//strings and colors for assembler exit codes
gchar *status_strings[] = {"Assembly succeeded!\n", "Assembly succeeded, with warnings\n",
                           "Errors encountered\n", "Stopped because of fatal error\n"};
output_color status_colors[] = {GREEN, ORANGE, RED, RED};

void callback_countcode_final (GPid, gint, gpointer);
void build_createsymtable (const gchar *);
void build_normal (const gchar *, gint);
gboolean run_command (const gchar *);


/*
 * Does some basic parsing on
 * a command and returns an
 * allocated argv
 */

gchar **parse_command (const char *command) {
	gboolean in_quotes;
	int curr_arg;
	gchar *curr_pos, *arg_start, *arg_end, **argv = malloc (sizeof (gchar *) * (MAX_CMD_ARGS + 1));

	in_quotes = FALSE;
	curr_arg = 0;
	curr_pos = command;
	while (*curr_pos) {
		if (curr_arg == MAX_CMD_ARGS) {
			show_error ("Too many arguments passed to command", "The extra ones will be ignored");
			break;
		}

		//check if this argument's in quotes
		if (*curr_pos == '\"') {
			curr_pos++;
			in_quotes = TRUE;
		}
		//save the its starting position
		arg_start = curr_pos;
		//then find the end
		if (in_quotes) {
			while (*curr_pos && *curr_pos != '\"') curr_pos++;
			arg_end = curr_pos;
			if (*curr_pos) curr_pos++;
		} else {
			while (*curr_pos && !isspace (*curr_pos)) curr_pos++;
			arg_end = curr_pos;
		}
		//now save the argument
		argv[curr_arg++] = g_strndup (arg_start, arg_end - arg_start);
		//and skip whitespace to get to the next one
		in_quotes = FALSE;
		while (*curr_pos && isspace (*curr_pos)) curr_pos++;
	}
	//terminate with a NULL
	argv[curr_arg] = NULL;

	return argv;
}


/*
 * Frees an argv created
 * with parse_command
 */

void free_command_argv (gchar **argv) {
	int i;
	for (i = 0; argv[i]; i++) g_free (argv[i]);
	g_free (argv);
}


/*
 * Makes sure file is saved
 * before assembling, returns
 * TRUE if successful
 */

gboolean check_file_saved (file *currfile) {
	GtkWidget *dialog;
	gchar *filename;
	gint response;

	//check if the file's been modified...
	if (gtk_text_buffer_get_modified (currfile->textbuffer) || currfile->path == NULL) {

		//if we always save before building, then just save and don't worry about it
		if (save_before_build)
			return save_file (currfile, FALSE);

		//otherwise, ask first
		if (currfile->path)
			filename = g_path_get_basename (currfile->path);
		else
			filename = g_strdup ("(Unsaved file)");

		dialog = gtk_message_dialog_new (GTK_WINDOW (window), GTK_DIALOG_MODAL,
			GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO,
			"You must save %s\nbefore assembling it", filename);
		gtk_dialog_add_button (GTK_DIALOG (dialog), "Yes, always", GTK_RESPONSE_ACCEPT);
		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
			"Would you like to save it now?");
		g_free (filename);

		response = gtk_dialog_run (GTK_DIALOG (dialog));
		if (response == GTK_RESPONSE_YES || response == GTK_RESPONSE_ACCEPT) {
			if (response == GTK_RESPONSE_ACCEPT) save_before_build = TRUE;
			gtk_widget_destroy (dialog);
			if (!save_file (currfile, FALSE))
				return FALSE;

		} else {
			gtk_widget_destroy (dialog);
			return FALSE;
		}
	}

	return TRUE;
}


/*
 * Starts assembly from current build step
 */

void do_build_step () {
	GList *curr_node;
	file *test_file;
	gchar *fullpath;

	//first make sure this file is saved
	//so we've got to check if it's in the list of open files
	for (curr_node = open_files; curr_node; curr_node = curr_node->next) {
		test_file = (file *) (curr_node->data);
		if (curr_build_step->filename && !strcmp (test_file->path, curr_build_step->filename)) {
			//if it's the same path, then make sure it's saved
			if (!check_file_saved (test_file))
				return;

			break;
		}
	}
	//if it's not open, then we can assume it's saved

	if (curr_build_step->filename) {
		if (g_path_is_absolute (curr_build_step->filename))
			fullpath = g_strdup (curr_build_step->filename);
		else
			fullpath = g_build_filename (project_path, curr_build_step->filename, NULL);
	}

	//now the simple part - call the right build function depending on the current action
	switch (curr_build_step->action) {
		case BUILD_NORMAL:
			build_normal (fullpath, curr_build_step->output_type);
			break;
		case BUILD_SYMTABLE:
			build_createsymtable (fullpath);
			break;
		case BUILD_COMMAND:
			run_command (curr_build_step->command);
	}

	g_free (fullpath);
}


/*
 * Called when status of assembler's
 * output message pipe changes,
 * prints messages in assembler
 * output window
 */

gboolean callback_get_assembler_output (GIOChannel *source, GIOCondition condition, gpointer data) {
	gchar output_buffer[OUTPUT_BUF_SIZE];
	gsize bytes_read;

	//if there was an error or there's nothing more to print, then exit
	if (condition == G_IO_ERR || condition == G_IO_HUP)
		return FALSE;

	//otherwise, read data from the IO channel
	if (g_io_channel_read_chars (source, output_buffer, OUTPUT_BUF_SIZE - 1, &bytes_read, NULL)
		!= G_IO_STATUS_NORMAL)

		return FALSE;

	//and print it
	output_buffer[bytes_read] = '\0';
	append_assembler_output (output_buffer, NONE);
	return TRUE;
}


/*
 * Called when assembler exits
 */

void callback_assembler_exited (GPid pid, gint status, gpointer output_file) {
	gboolean assembly_failed;
	gsize bytes_read;
	gchar output_buffer[OUTPUT_BUF_SIZE];
	gchar *dialog_message, *old_dialog_message;

	//first, flush the messages from the assembler
	while (TRUE) {
		if (g_io_channel_read_chars (asm_output, output_buffer, OUTPUT_BUF_SIZE - 1, &bytes_read, NULL)
			== G_IO_STATUS_NORMAL) {

			output_buffer[bytes_read] = '\0';
			append_assembler_output (output_buffer, NONE);
		} else {
			g_io_channel_unref (asm_output);
			break;
		}
	}

	//write the status message to the output window, and if there's an output file
	// to be opened, open it
#ifndef G_OS_WIN32
	if (!WIFEXITED (status)) {
		show_error ("The assembler crashed", "Please report this bug to burntfuse@gmail.com\n\
or the Wabbitcode forum at\nwww.revsoft.org and attach the code that\ncaused the crash.");
		assembly_failed = TRUE;
	} else
#endif
	if (assembler_killed) {
		append_assembler_output (ASSEMBLY_STOPPED_STR, RED);
		assembly_failed = TRUE;
	} else {
#ifndef G_OS_WIN32
		status = WEXITSTATUS (status);
#endif

		if (status <= 3)
			append_assembler_output (status_strings[status], status_colors[status]);
		else
			append_assembler_output (ASSEMBLY_UNKNOWN_STR, RED);

		assembly_failed = (status >= 2);
		if (!assembly_failed && output_file && !curr_build_step)
			open_file ((gchar *)output_file);
	}

	//then show that the assembler's no longer running
	update_assembler_outputwin (FALSE);

	if (output_file)
		g_free (output_file);

	//now check if there's anything on stderr - if there was, it's info to be displayed in a dialog
	//TODO: find some better way to do this?
	dialog_message = NULL;
	while (TRUE) {
		if (g_io_channel_read_chars (asm_err, output_buffer, OUTPUT_BUF_SIZE - 1, &bytes_read, NULL)
			!= G_IO_STATUS_NORMAL)

			break;

		output_buffer[bytes_read] = '\0';
		old_dialog_message = dialog_message;

		if (old_dialog_message) {
			dialog_message = g_strconcat (dialog_message, output_buffer, NULL);
			g_free (old_dialog_message);
		} else {
			dialog_message = g_strdup (output_buffer);
		}
	}
	g_io_channel_unref (asm_err);

	if (dialog_message) {
		show_info (dialog_message);
		g_free (dialog_message);
	}

	//(for future portability to Windows)
	g_spawn_close_pid (pid);

	//if we're doing this as part of a project build sequence, move on to the next step
	if (curr_build_step && curr_build_step->next && !assembly_failed) {
		curr_build_step = curr_build_step->next;
		do_build_step ();
	}
}


/*
 * Called when a command exits
 */

void callback_command_exited (GPid pid, gint status, gpointer data) {
	gsize bytes_read;
	gchar output_buffer[OUTPUT_BUF_SIZE];
	gboolean assembly_failed;

	//first, flush the messages on stdout...
	while (TRUE) {
		if (g_io_channel_read_chars (asm_output, output_buffer, OUTPUT_BUF_SIZE - 1, &bytes_read, NULL)
			== G_IO_STATUS_NORMAL) {

			output_buffer[bytes_read] = '\0';
			append_assembler_output (output_buffer, NONE);
		} else {
			g_io_channel_unref (asm_output);
			break;
		}
	}
	//...and stderr
	while (TRUE) {
		if (g_io_channel_read_chars (asm_err, output_buffer, OUTPUT_BUF_SIZE - 1, &bytes_read, NULL)
			== G_IO_STATUS_NORMAL) {

			output_buffer[bytes_read] = '\0';
			append_assembler_output (output_buffer, NONE);
		} else {
			g_io_channel_unref (asm_err);
			break;
		}
	}

	//write the status message to the output window
#ifdef G_OS_WIN32
	if (status == 0) {
#else
	if (WIFEXITED (status) && WEXITSTATUS (status) == 0) {
#endif
		append_assembler_output (COMMAND_SUCCESS_STR, GREEN);
		assembly_failed = FALSE;
#ifndef G_OS_WIN32
	} else if (assembler_killed) {
		append_assembler_output (ASSEMBLY_STOPPED_STR, RED);
		assembly_failed = TRUE;
#endif
	} else {
		append_assembler_output (COMMAND_FAILURE_STR, RED);
		assembly_failed = TRUE;
	}

	//show that the assembler's no longer running
	update_assembler_outputwin (FALSE);

	//(for future portability to Windows)
	g_spawn_close_pid (pid);

	//if we're doing this as part of a project build sequence, move on to the next step
	if (curr_build_step && curr_build_step->next && !assembly_failed) {
		curr_build_step = curr_build_step->next;
		do_build_step ();
	}
}


/*
 * Changes filename extension of path,
 * returns allocated string
 */

gchar *change_extension (const gchar *path, char *new_extension) {
	gchar *new_path;
	int pos;

	//now find the extension of the file
	for (pos = strlen (path) - 1; ; pos--) {
		if (pos == 0 || G_IS_DIR_SEPARATOR (path[pos])) {
			pos = strlen (path);
			break;
		} else if (path[pos] == '.') {
			pos++;
			break;
		}
	}

	//and create the new path
	new_path = (gchar *)g_malloc (sizeof (gchar) * (pos + strlen (new_extension) + 1));
	strncpy (new_path, path, pos);
	new_path[pos] = '\0';
	strcat (new_path, new_extension);

	return new_path;
}


/*
 * Builds an argument array
 * for running the assembler
 */

gchar **build_argv (const gchar *arg1, const gchar *arg2) {
	gchar **argv;
	int argv_size, curr_arg;
	GtkTreeIter iter;
	gboolean more_dirs;

	//create the array
	argv_size = 6 + 2 * gtk_tree_model_iter_n_children (GTK_TREE_MODEL (include_dirs), NULL);
	argv = g_slice_alloc (sizeof (gchar*) * argv_size);
	//fill in the constant parts
	argv[0] = ASSEMBLER_NAME;
	argv[1] = arg1;
	argv[2] = arg2;
	argv[argv_size - 3] = "-N";
	argv[argv_size - 2] = labels_casesensitive ? "-A" : NULL;
	argv[argv_size - 1] = NULL;

	//then fill in the include dirs
	curr_arg = 3;
	more_dirs = gtk_tree_model_get_iter_first (GTK_TREE_MODEL (include_dirs), &iter);
	while (more_dirs) {
		gchar *curr_dir;
		gtk_tree_model_get (GTK_TREE_MODEL (include_dirs), &iter, 0, &curr_dir, -1);
		argv[curr_arg++] = "-I";
		argv[curr_arg++] = curr_dir;
		more_dirs = gtk_tree_model_iter_next (GTK_TREE_MODEL (include_dirs), &iter);
	}

	return argv;
}


/*
 * Frees an argument array
 * created with build_argv
 */

void free_argv (gchar **argv) {
	int curr_arg = 3;

	//free include dirs
	while (argv[curr_arg] && strcmp (argv[curr_arg], "-N") != 0) {
		curr_arg++;
		g_free (argv[curr_arg++]);
	}

	//now free the array itself
	g_slice_free1 (sizeof (gchar*) * (curr_arg + 1), argv);
}


/*
 * Runs specified command,
 * returns TRUE on success
 */

gboolean run_command (const gchar *command) {
	gchar *outmessage, **argv;
	int asm_stdout, asm_stderr;
	GError *error = NULL;

	//set up arguments
	argv = parse_command (command);

	//try running the command
	if (!g_spawn_async_with_pipes (project_path, argv, NULL, G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_SEARCH_PATH,
		NULL, NULL, &asm_pid, NULL, &asm_stdout, &asm_stderr, &error)) {

		show_error ("Couldn't run command", error->message);
		free_command_argv (argv);
		g_error_free (error);
		return FALSE;
	}

	//if that worked, set up a callback for when it exits,
	free_command_argv (argv);
	assembler_killed = FALSE;
	g_child_watch_add (asm_pid, callback_command_exited, NULL);

	//show the assembler output window
	update_assembler_outputwin (TRUE);
	outmessage = g_strdup_printf (COMMAND_RUN_STR, command);
	append_assembler_output (outmessage, BLUE);
	g_free (outmessage);

	//and listen for output to print
#ifdef G_OS_WIN32
	asm_output = g_io_channel_win32_new_fd (asm_stdout);
	asm_err = g_io_channel_win32_new_fd (asm_stderr);
#else
	asm_output = g_io_channel_unix_new (asm_stdout);
	asm_err = g_io_channel_unix_new (asm_stderr);
#endif
	g_io_add_watch (asm_output, G_IO_IN | G_IO_ERR | G_IO_HUP, callback_get_assembler_output, NULL);
	g_io_add_watch (asm_err, G_IO_IN | G_IO_ERR | G_IO_HUP, callback_get_assembler_output, NULL);
	return TRUE;
}


/*
 * Runs assembler with specified
 * arg vector, output file,
 * callback (uses default if NULL),
 * and output message, returns
 * TRUE on success
 */

gboolean do_build (const gchar *message, gchar **argv, gchar *output_file, GChildWatchFunc callback) {
	GError *error = NULL;
	gchar *working_dir, *outmessage;
	gint asm_stdout, asm_stderr;
#ifdef G_OS_WIN32
	gchar *wd;
#endif

	//make sure weird stuff doesn't happen if a project isn't open
	if (!project_open || !build_sequence)
		curr_build_step = NULL;

	//get the directory of the file to use as the working dir for assembler
	//(first argument to SPASM is always the filename)
	working_dir = g_path_get_dirname (argv[1]);

#ifdef G_OS_WIN32
	//random path stuff
	wd = getcwd (NULL, -1);
	argv[0] = g_build_filename (wd, ASSEMBLER_NAME, NULL);
	g_free (wd);
#endif

	if (!g_spawn_async_with_pipes ((const gchar *)working_dir, argv, NULL, G_SPAWN_DO_NOT_REAP_CHILD |
		G_SPAWN_SEARCH_PATH, NULL, NULL, &asm_pid, NULL, &asm_stdout, &asm_stderr, &error)) {

		show_error ("Couldn't run assembler", error->message);
		g_error_free (error);
		g_free (working_dir);
#ifdef G_OS_WIN32
		g_free (argv[0]);
#endif
		return FALSE;
	}

#ifdef G_OS_WIN32
	g_free (argv[0]);
#endif

	//if that worked, set up a callback for when the assembler closes,
	assembler_killed = FALSE;
	if (callback)
		g_child_watch_add (asm_pid, callback, output_file);
	else
		g_child_watch_add (asm_pid, callback_assembler_exited, output_file);

	//show the assembler output window
	update_assembler_outputwin (TRUE);
	outmessage = g_strdup_printf (message, argv[1]);
	append_assembler_output (outmessage, BLUE);
	g_free (outmessage);

	//and listen for output to print
#ifdef G_OS_WIN32
	asm_output = g_io_channel_win32_new_fd (asm_stdout);
	asm_err = g_io_channel_win32_new_fd (asm_stderr);
#else
	asm_output = g_io_channel_unix_new (asm_stdout);
	asm_err = g_io_channel_unix_new (asm_stderr);
#endif
	g_io_add_watch (asm_output, G_IO_IN | G_IO_ERR | G_IO_HUP, callback_get_assembler_output, NULL);

	g_free (working_dir);
	return TRUE;
}


/*
 * Assembles file
 */

void build_normal (const gchar *filename, gint output_type) {
	gchar *output_filename, *new_extension, **argv;

	//set up the argument vector for the assembler
	new_extension = get_output_type_extension (output_type);
	output_filename = change_extension (filename, new_extension);
	g_free (new_extension);
	argv = build_argv (filename, output_filename);

	//and run it
	do_build (ASSEMBLY_START_STR, argv, NULL, NULL);
	free_argv (argv);
	g_free (output_filename);
}


/*
 * Assembles current file
 */

void callback_build (GtkWidget *widget, gpointer data) {

	if (project_open && build_sequence) {
		//if a project's open, first clear the output window
		clear_assembler_outputwin ();

		//then follow each build step
		curr_build_step = build_sequence;
		do_build_step ();

	} else {

		//make sure there's an open file first
		if (!currfile) {
			show_error ("There isn't an open file to assemble,\nand you haven't set how the project\n\
should compile.", "To set the project build sequence, select\n'Project options' from the 'Project' menu");
			return;
		}

		//if not or if there isn't a build sequence for the project,
		// just build the current file normally
		if (!check_file_saved (currfile))
			return;

		build_normal (currfile->path, output_type);
	}
}


/*
 * Runs code counter on selected text
 */

void callback_countcode (GtkWidget *widget, gpointer data) {
	gchar *code, *argv[] = {ASSEMBLER_NAME, NULL, "-C", "-N", NULL};
	int tempfile;
	GError *error = NULL;

	//first get the selected text
	gtk_text_buffer_get_selection_bounds (currfile->textbuffer, &selected_startiter, &selected_enditer);
	//if there isn't any selected text, ignore it
	if (gtk_text_iter_equal (&selected_startiter, &selected_enditer))
		return;

	code = gtk_text_buffer_get_text (currfile->textbuffer, &selected_startiter, &selected_enditer, FALSE);
	//now write it to a temp file
	tempfile = g_file_open_tmp ("wabbitcode-XXXXXX", &argv[1], &error);
	if (tempfile == -1) {
		show_error ("Couldn't create temporary file", error->message);
		g_error_free (error);
		return;
	}

	tempfilename = argv[1];
	close (tempfile);
	if (!g_file_set_contents (argv[1], code, -1, &error)) {
		show_error ("Couldn't write to temporary file", error->message);
		g_error_free (error);
		g_free (code);
		return;
	}
	g_free (code);

	//run the assembler
	do_build (COUNT_START_STR, argv, NULL, callback_countcode_final);
}

void callback_countcode_final (GPid pid, gint status, gpointer output_file) {
	callback_assembler_exited (pid, status, output_file);

	//remove the temp file
	g_unlink (tempfilename);
	g_free (tempfilename);
	//then re-select the text
	gtk_text_buffer_select_range (currfile->textbuffer, &selected_startiter, &selected_enditer);
}


/*
 * Generates symbol table for file
 */

void build_createsymtable (const gchar *filename) {
	gchar *output_file, **argv;

	//set up the argument vector for the assembler
	argv = build_argv (filename, "-L");
	output_file = change_extension (filename, "inc");

	//and run it
	if (!do_build (SYMTABLE_START_STR, argv, output_file, NULL))
		g_free (output_file);

	free_argv (argv);
}


/*
 * Generates symbol table for current file
 */

void callback_createsymtable (GtkWidget *widget, gpointer data) {

	//make sure the file's saved first
	if (!check_file_saved (currfile))
		return;

	build_createsymtable (currfile->path);
}


/*
 * Creates code listing
 */

void callback_createlisting (GtkWidget *widget, gpointer data) {
	gchar *output_file, **argv;

	//make sure the file's saved first
	if (!check_file_saved (currfile))
		return;

	//and set up the argument vector for the assembler
	argv = build_argv (currfile->path, "-T");
	output_file = change_extension (currfile->path, "lst");

	if (!do_build (LISTING_START_STR, argv, output_file, NULL))
		g_free (output_file);

	free_argv (argv);
}


/*
 * Shows program statistics
 */

void callback_printstats (GtkWidget *widget, gpointer data) {
	gchar **argv;

	//make sure the file's saved first
	if (!check_file_saved (currfile))
		return;

	//and set up the argument vector for the assembler
	argv = build_argv (currfile->path, "-S");

	//finally, run the assembler
	do_build (STAT_START_STR, argv, NULL, NULL);
	free_argv (argv);
}


/*
 * Kills assembler
 */

void callback_stop_assembler (GtkWidget *widget, gpointer data) {
	assembler_killed = TRUE;

#ifdef G_OS_WIN32
	TerminateProcess ((HANDLE) asm_pid, 1);
#else
	kill ((pid_t)asm_pid, 15);
	//TODO: wait a bit, and if it doesn't exit send 9 to finish it off for good (wahahahaha...)
#endif
}

