/*
	Wabbitcode - An IDE for the Z80 line of TI calculators
	Copyright (C) 2007 Donald Straney

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef G_OS_WIN32
#include <windows.h>
#include <unistd.h>
#endif

#include <gtk/gtk.h>
#include "asmide.h"
#include "file.h"
#include "settings.h"
#include "undo.h"
#include "project.h"


/*
 * Called when window's close button
 * is pressed - closes all files and
 * exits
 */

gboolean callback_close_window (GtkWidget *widget, gpointer data) {

	//is a project open?
	if (project_open) {
		//if so, do all the project close stuff
		if (!close_project ())
			return TRUE;
	} else {
		//otherwise, just try to close all files
		if (!remove_all_files ())
			return TRUE;
	}

	//if the user didn't cancel that, save the settings
	gtk_window_get_size (GTK_WINDOW (window), &winwidth, &winheight);
	save_settings ();
	//and exit
	return FALSE;
}


/*
 * Called when "quit" is selected from menu
 */

void callback_quit (GtkWidget *widget, gpointer data) {

	//so we don't end up with two copies of all the cleanup & exit stuff,
	// it just makes the window's close button callback do all the dirty work
	if (!callback_close_window (NULL, NULL))
		gtk_main_quit ();
}


/*
 * Called when the active tab in the notebook
 * is changed - updates line numbers and state
 * of undo/redo widgets for the active file,
 * sets currfile to point to active file,
 * sets window title
 */

void callback_switch_file (GtkNotebook *notebook, GtkNotebookPage *page,
						   guint page_num, gpointer data) {

	//TODO: move to main.c?
	GtkWidget *child;
	GList *searchfile;
	int i = 0;
	action *curr_action;

	//find the file struct for the new active file and make currfile point to it
	child = gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), (gint)page_num);
	for (searchfile = open_files; searchfile && ((file *)searchfile->data)->scrolledwin != child;
		searchfile = searchfile->next);
	//TODO: g_assert?
	currfile = (file *)searchfile->data;

	//then update all the other stuff
	update_statusbar (NULL);
	set_redo_sensitive (currfile->undo_stack_pos > 0);
	for (curr_action = currfile->undo_stack; i < currfile->undo_stack_pos;
		i++, curr_action = curr_action->next);
	set_undo_sensitive (curr_action && curr_action->next);

	gtk_window_set_title (GTK_WINDOW (window), gtk_label_get_text (GTK_LABEL (currfile->tablabel)));
}


/*
 * Creates new (empty) file
 */

void callback_new (GtkWidget *widget, gpointer data) {

	add_file ("", -1, NULL);
}


/*
 * Saves current file
 */

void callback_save (GtkWidget *widget, gpointer data) {

	save_file (currfile, FALSE);
}


/*
 * Saves current file under a different name
 */

void callback_saveas (GtkWidget *widget, gpointer data) {

	save_file (currfile, TRUE);
}


/*
 * Closes current file
 */

void callback_close (GtkWidget *widget, gpointer data) {

	if (data)
		remove_file ((file *)data);
	else
		remove_file (currfile);
}


/*
 * Shows dialog and lets user
 * pick a file to open
 */

void callback_open (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog;
	gint response;
	gchar *path;

	dialog = gtk_file_chooser_dialog_new ("Open File", GTK_WINDOW (window),
				GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

	if (project_open) {
		gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), project_path);
	} else if (currfile && currfile->path) {
		path = g_path_get_dirname (currfile->path);
		gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), path);
		g_free (path);
	}
	add_file_chooser_filters (GTK_FILE_CHOOSER (dialog));
	response = gtk_dialog_run (GTK_DIALOG (dialog));

	if (response != GTK_RESPONSE_ACCEPT) {
		gtk_widget_destroy (dialog);
		return;
	}

	path = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
	gtk_widget_destroy (dialog);

	//try to open the file
	open_file (path);

	g_free (path);
}


/*
 * Reverts to saved version of current file
 */

void callback_revert (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog;
	gchar *contents;
	gint response;
	guint len;

	if (!currfile->path)
		return;

	dialog = gtk_message_dialog_new (GTK_WINDOW (window),
				GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_WARNING,
				GTK_BUTTONS_YES_NO,
				"Are you sure you want to revert to\nthe saved version of this file?");
	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
		"You will lose any unsaved changes");
	response = gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);

	if (response == GTK_RESPONSE_NO)
		return;

	if (read_file (currfile->path, &contents, &len)) {
		gtk_text_buffer_set_text (currfile->textbuffer, contents, len);
		g_free (contents);
		gtk_text_buffer_set_modified (currfile->textbuffer, FALSE);
	}
}


/*
 * Copies selected text to clipboard
 */

void callback_copy (GtkWidget *widget, gpointer data) {

	gtk_text_buffer_copy_clipboard (currfile->textbuffer, gtk_clipboard_get (GDK_SELECTION_CLIPBOARD));
}


/*
 * Cuts selected text to clipboard
 */

void callback_cut (GtkWidget *widget, gpointer data) {

	gtk_text_buffer_cut_clipboard (currfile->textbuffer,
		gtk_clipboard_get (GDK_SELECTION_CLIPBOARD), TRUE);
}


/*
 * Pastes text from clipboard
 */

void callback_paste (GtkWidget *widget, gpointer data) {

	gtk_text_buffer_paste_clipboard (currfile->textbuffer,
		gtk_clipboard_get (GDK_SELECTION_CLIPBOARD), NULL, TRUE);
	gtk_widget_grab_focus (currfile->textview);
}


/*
 * Selects all text in file
 */

void callback_select_all (GtkWidget *widget, gpointer data) {
	GtkTextIter startiter, enditer;

	gtk_text_buffer_get_bounds (currfile->textbuffer, &startiter, &enditer);
	gtk_text_buffer_select_range (currfile->textbuffer, &startiter, &enditer);
	gtk_widget_grab_focus (currfile->textview);
}


/*
 * Changes case of
 * selected text
 */

void callback_change_case (GtkWidget *widget, gpointer type) {
	GtkTextIter startiter, curriter, enditer;
	gchar *old_text, *new_text;
	GString *curr_string;
	gunichar curr_char;
	gint len;

	//make sure there's a selection
	if (!gtk_text_buffer_get_selection_bounds (currfile->textbuffer, &startiter, &enditer))
		return;

	//and convert it
	switch ((case_type)type) {
		case LOWER:
			old_text = gtk_text_buffer_get_text (currfile->textbuffer, &startiter, &enditer, FALSE);
			new_text = g_utf8_strdown (old_text, -1);
			g_free (old_text);
			len = g_utf8_strlen (new_text, -1);
			break;
		case UPPER:
			old_text = gtk_text_buffer_get_text (currfile->textbuffer, &startiter, &enditer, FALSE);
			new_text = g_utf8_strup (old_text, -1);
			g_free (old_text);
			len = g_utf8_strlen (new_text, -1);
			break;
		case TITLE:
			curr_string = g_string_new (NULL);
			curriter = startiter;
			while (!gtk_text_iter_equal (&curriter, &enditer)) {
				curr_char = gtk_text_iter_get_char (&curriter);
				if (gtk_text_iter_starts_word (&curriter))
					curr_char = g_unichar_totitle (curr_char);
				else
					curr_char = g_unichar_tolower (curr_char);
				g_string_append_unichar (curr_string, curr_char);
				gtk_text_iter_forward_char (&curriter);
			}
			new_text = curr_string->str;
			len = curr_string->len;
			break;

		case INVERT:
			//TODO: use g_string_sized_new and get size with
			//gtk_text_iter_get_offset (&enditer) - gtk_text_iter_get_offset (&startiter)
			curr_string = g_string_new (NULL);
			curriter = startiter;
			while (!gtk_text_iter_equal (&curriter, &enditer)) {
				curr_char = gtk_text_iter_get_char (&curriter);
				if (g_unichar_islower (curr_char))
					curr_char = g_unichar_toupper (curr_char);
				else
					curr_char = g_unichar_tolower (curr_char);
				gtk_text_iter_forward_char (&curriter);
				g_string_append_unichar (curr_string, curr_char);
			}
			new_text = curr_string->str;
			len = curr_string->len;
			break;
	}
	//then replace the original text with the new converted string
	gtk_text_buffer_begin_user_action (currfile->textbuffer);
	gtk_text_buffer_delete (currfile->textbuffer, &startiter, &enditer);
	gtk_text_buffer_insert (currfile->textbuffer, &startiter, new_text, len);
	gtk_text_buffer_end_user_action (currfile->textbuffer);
	//restore the selection
	enditer = startiter;
	gtk_text_iter_backward_chars (&enditer, len);
	gtk_text_buffer_select_range (currfile->textbuffer, &startiter, &enditer);

	if ((case_type)type == LOWER || (case_type)type == UPPER)
		g_free (new_text);
	else
		g_string_free (curr_string, TRUE);
}


/*
 * Shows about dialog
 */

void callback_show_about (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog;
	gchar *license;
	const gchar *authors[] = {"Editor: Donald Straney", "SPASM (assembler): Spencer Putt, Donald Straney", NULL};
#ifdef G_OS_WIN32
	const gchar *artists[] = {"Stock icons from GNOME and Mist icon themes",
"File icons modified from GNOME icon theme", "Record icon by Tango project",
"Play icon modified from Tango project icon", "Clock icon by Donald Straney", NULL};
#else
	const gchar *artists[] = {"File icons modified from GNOME icon theme", "Record icon by Tango \
project", "Play icon modified from Tango project icon", "Clock icon by Donald Straney", NULL};
#endif

	dialog = gtk_about_dialog_new ();
	gtk_about_dialog_set_name (GTK_ABOUT_DIALOG (dialog), APP_NAME);
	gtk_about_dialog_set_version (GTK_ABOUT_DIALOG (dialog), VERSION);
	gtk_about_dialog_set_copyright (GTK_ABOUT_DIALOG (dialog), "Copyright Donald Straney 2008");
	gtk_about_dialog_set_comments (GTK_ABOUT_DIALOG (dialog), "\nAn IDE for the Z80 line of TI \
calculators\n\nSend comments/suggestions/flames\nto burntfuse@gmail.com\n\nA product of Revolution \
Software - \nsee our website at www.revsoft.org");
	if (g_file_get_contents (DOCDIR"/COPYING", &license, NULL, NULL)) {
		gtk_about_dialog_set_license (GTK_ABOUT_DIALOG (dialog), license);
		g_free (license);
	}
//	gtk_about_dialog_set_website (dialog, "http://www.revsoft.org");
	gtk_about_dialog_set_authors (GTK_ABOUT_DIALOG (dialog), authors);
	gtk_about_dialog_set_artists (GTK_ABOUT_DIALOG (dialog), artists);
	gtk_about_dialog_set_logo_icon_name (GTK_ABOUT_DIALOG (dialog), "wabbitcode");
	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
}


/*
 * Shows help files
 * in web browser
 */

void callback_show_help (GtkWidget *widget, gpointer filename) {
#ifdef G_OS_WIN32
	gchar *wd, *fullpath;

	wd = getcwd (NULL, -1);
	fullpath = g_build_filename (wd, (gchar *)filename, NULL);
	ShellExecute (NULL, NULL, fullpath, NULL, NULL, 0);
	g_free (wd);
	g_free (fullpath);
#else
	gchar *argv[] = {"xdg-open", NULL, NULL};
	argv[1] = g_build_filename (DOCDIR, (gchar *)filename, NULL);

	if (!g_spawn_async (NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, NULL)) {
		//if we couldn't open the manual page with xdg-open, show an error
		show_error ("Couldn't find a web browser",
			"Look in "DOCDIR" for the help pages");
	}

	g_free (argv[1]);
#endif
}


/*
 * Shows editor options dialog
 */

void callback_seteditoropt (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog, *label, *table, *tabwidthnum, *autoindenton, *wordwrapon,
		*fontbutton, *recentfiles, *searchresultcolor, *undolen;
	GtkAdjustment *adjustment;
	gchar *fontname;

	//build dialog, etc.
	dialog = gtk_dialog_new_with_buttons ("General Settings", GTK_WINDOW (window),
				GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
				GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

	table = gtk_table_new (7, 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE (table), 6);
	gtk_table_set_col_spacings (GTK_TABLE (table), 6);

	label = gtk_label_new ("Editor font:");
	fontname = pango_font_description_to_string (editorfont);
	fontbutton = gtk_font_button_new_with_font (fontname);
	g_free (fontname);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 0, 1);
	gtk_table_attach_defaults (GTK_TABLE (table), fontbutton, 1, 2, 0, 1);

	label = gtk_label_new ("Tab width:");
	adjustment = (GtkAdjustment *)gtk_adjustment_new ((gdouble)tabwidth,
		1.0, 20.0, 1.0, 1.0, 1.0);
	tabwidthnum = gtk_spin_button_new (adjustment, 1.0, 0);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 1, 2);
	gtk_table_attach_defaults (GTK_TABLE (table), tabwidthnum, 1, 2, 1, 2);

	label = gtk_label_new ("Search result color:");
	searchresultcolor = gtk_color_button_new_with_color ((const GdkColor *)&resultcolor);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 2, 3);
	gtk_table_attach_defaults (GTK_TABLE (table), searchresultcolor, 1, 2, 2, 3);

	label = gtk_label_new ("Recent files:");
	adjustment = (GtkAdjustment *)gtk_adjustment_new ((gdouble)numrecentfiles,
		1.0, 20.0, 1.0, 1.0, 1.0);
	recentfiles = gtk_spin_button_new (adjustment, 1.0, 0);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 3, 4);
	gtk_table_attach_defaults (GTK_TABLE (table), recentfiles, 1, 2, 3, 4);

	label = gtk_label_new ("Undo memory size:");
	adjustment = (GtkAdjustment *)gtk_adjustment_new ((gdouble)max_undo_stack_len,
		5.0, 200.0, 1.0, 1.0, 1.0);
	undolen = gtk_spin_button_new (adjustment, 1.0, 0);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 4, 5);
	gtk_table_attach_defaults (GTK_TABLE (table), undolen, 1, 2, 4, 5);

	autoindenton = gtk_check_button_new_with_label ("Indent automatically");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (autoindenton), autoindent);
	gtk_table_attach_defaults (GTK_TABLE (table), autoindenton, 0, 2, 5, 6);

	wordwrapon = gtk_check_button_new_with_label ("Wrap text to window");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wordwrapon), wordwrap);
	gtk_table_attach_defaults (GTK_TABLE (table), wordwrapon, 0, 2, 6, 7);

	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), table);
	gtk_container_set_border_width (GTK_CONTAINER (table), 6);
	gtk_widget_show_all (table);

	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_REJECT) {
		gtk_widget_destroy (dialog);
		return;
	}

	//if the user pressed OK, set the settings
	pango_font_description_free (editorfont);
	editorfont = pango_font_description_from_string (gtk_font_button_get_font_name
		(GTK_FONT_BUTTON (fontbutton)));
	tabwidth = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (tabwidthnum));
	gtk_color_button_get_color (GTK_COLOR_BUTTON (searchresultcolor), &resultcolor);
	numrecentfiles = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (recentfiles));
	autoindent = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (autoindenton));
	wordwrap = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wordwrapon));
	max_undo_stack_len = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (undolen));
	update_editor_settings ();

	gtk_widget_destroy (dialog);
}


/*
 * Shows syntax highlighting options dialog
 */

/*void callback_setsyntaxhltopt_globalchange (GtkToggleButton *togglebutton, gpointer list) {
	gboolean state;
	GList *widget_list;

	state = gtk_toggle_button_get_active (togglebutton);
	widget_list = (GList *)list;

	for (; widget_list; widget_list = widget_list->next)
		gtk_widget_set_sensitive (GTK_WIDGET (widget_list->data), state);
}

void callback_setsyntaxhltopt_change (GtkToggleButton *togglebutton, gpointer data) {

}*/

void callback_setsyntaxhltopt (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog, *label, *table, *labelcoloron, *labelcolorbutton, *commentcoloron,
		*commentcolorbutton, *dircoloron, *dircolorbutton, *strcoloron, *strcolorbutton,
		*opcodecoloron, *opcodecolorbutton, *regcoloron, *regcolorbutton, *syntaxhlton;
//	GList *widget_list = NULL;
	GList *node;
	file *currfile;
	GtkTextIter startiter, enditer;
	gboolean new_dosyntaxhlt, new_labelhlt, new_commenthlt, new_dirhlt, new_strhlt,
		new_opcodehlt, new_reghlt;

	//build dialog, etc.
	dialog = gtk_dialog_new_with_buttons ("Syntax Highlighting Settings", GTK_WINDOW (window),
				GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
				GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

	table = gtk_table_new (7, 3, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE (table), 6);
	gtk_table_set_col_spacings (GTK_TABLE (table), 6);

	label = gtk_label_new ("Use syntax highlighting");
	syntaxhlton = gtk_check_button_new ();
//	g_signal_connect (syntaxhlton, "toggled", G_CALLBACK (callback_setsyntaxhltopt_globalchange),
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (syntaxhlton), dosyntaxhlt);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 0, 1);
	gtk_table_attach_defaults (GTK_TABLE (table), syntaxhlton, 1, 2, 0, 1);

	label = gtk_label_new ("Highlight labels:");
	labelcoloron = gtk_check_button_new ();
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (labelcoloron), labelhlt);
	labelcolorbutton = gtk_color_button_new_with_color (&labelcolor);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 1, 2);
	gtk_table_attach_defaults (GTK_TABLE (table), labelcoloron, 1, 2, 1, 2);
	gtk_table_attach_defaults (GTK_TABLE (table), labelcolorbutton, 2, 3, 1, 2);

	label = gtk_label_new ("Highlight comments:");
	commentcoloron = gtk_check_button_new ();
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (commentcoloron), commenthlt);
	commentcolorbutton = gtk_color_button_new_with_color (&commentcolor);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 2, 3);
	gtk_table_attach_defaults (GTK_TABLE (table), commentcoloron, 1, 2, 2, 3);
	gtk_table_attach_defaults (GTK_TABLE (table), commentcolorbutton, 2, 3, 2, 3);

	label = gtk_label_new ("Highlight directives:");
	dircoloron = gtk_check_button_new ();
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dircoloron), dirhlt);
	dircolorbutton = gtk_color_button_new_with_color (&dircolor);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 3, 4);
	gtk_table_attach_defaults (GTK_TABLE (table), dircoloron, 1, 2, 3, 4);
	gtk_table_attach_defaults (GTK_TABLE (table), dircolorbutton, 2, 3, 3, 4);

	label = gtk_label_new ("Highlight strings:");
	strcoloron = gtk_check_button_new ();
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (strcoloron), strhlt);
	strcolorbutton = gtk_color_button_new_with_color (&strcolor);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 4, 5);
	gtk_table_attach_defaults (GTK_TABLE (table), strcoloron, 1, 2, 4, 5);
	gtk_table_attach_defaults (GTK_TABLE (table), strcolorbutton, 2, 3, 4, 5);

	label = gtk_label_new ("Highlight opcodes:");
	opcodecoloron = gtk_check_button_new ();
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (opcodecoloron), opcodehlt);
	opcodecolorbutton = gtk_color_button_new_with_color (&opcodecolor);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 5, 6);
	gtk_table_attach_defaults (GTK_TABLE (table), opcodecoloron, 1, 2, 5, 6);
	gtk_table_attach_defaults (GTK_TABLE (table), opcodecolorbutton, 2, 3, 5, 6);

	label = gtk_label_new ("Highlight registers:");
	regcoloron = gtk_check_button_new ();
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (regcoloron), reghlt);
	regcolorbutton = gtk_color_button_new_with_color (&regcolor);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 6, 7);
	gtk_table_attach_defaults (GTK_TABLE (table), regcoloron, 1, 2, 6, 7);
	gtk_table_attach_defaults (GTK_TABLE (table), regcolorbutton, 2, 3, 6, 7);

	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), table);
	gtk_container_set_border_width (GTK_CONTAINER (table), 6);
	gtk_widget_show_all (table);

	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_REJECT) {
		gtk_widget_destroy (dialog);
		return;
	}

	//if the user pressed OK, set the settings
	new_dosyntaxhlt = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (syntaxhlton));
	new_labelhlt = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (labelcoloron));
	new_commenthlt = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (commentcoloron));
	new_dirhlt = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dircoloron));
	new_strhlt = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (strcoloron));
	new_opcodehlt = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (opcodecoloron));
	new_reghlt = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (regcoloron));

	gtk_color_button_get_color (GTK_COLOR_BUTTON (labelcolorbutton), &labelcolor);
	gtk_color_button_get_color (GTK_COLOR_BUTTON (commentcolorbutton), &commentcolor);
	gtk_color_button_get_color (GTK_COLOR_BUTTON (dircolorbutton), &dircolor);
	gtk_color_button_get_color (GTK_COLOR_BUTTON (strcolorbutton), &strcolor);
	gtk_color_button_get_color (GTK_COLOR_BUTTON (opcodecolorbutton), &opcodecolor);
	gtk_color_button_get_color (GTK_COLOR_BUTTON (regcolorbutton), &regcolor);

	update_syntaxhlt_settings ();

	//ugh, just look at this monster...
	if (((new_labelhlt != labelhlt && new_labelhlt) || (new_commenthlt != commenthlt && new_commenthlt)
	    || (new_dirhlt != dirhlt && new_dirhlt) || (new_strhlt != strhlt && new_strhlt) ||
	    (new_opcodehlt != opcodehlt && new_opcodehlt) || (new_reghlt != reghlt && new_reghlt)
	    || (new_dosyntaxhlt != dosyntaxhlt && new_dosyntaxhlt)) && new_dosyntaxhlt) {

		//if any new syntax highlighting options have been turned on,
		// run syntax highlighting on all files
		dosyntaxhlt = new_dosyntaxhlt;
		labelhlt = new_labelhlt;
		commenthlt = new_commenthlt;
		dirhlt = new_dirhlt;
		strhlt = new_strhlt;
		opcodehlt = new_opcodehlt;
		reghlt = new_reghlt;

		for (node = open_files; node; node = g_list_next (node))
			run_syntax_hlt_on_file ((file *)(node->data));

	} else if (((new_labelhlt != labelhlt && !new_labelhlt) || (new_commenthlt != commenthlt &&
	           !new_commenthlt) || (new_dirhlt != dirhlt && !new_dirhlt) || (new_strhlt != strhlt &&
	           !new_strhlt) || (new_opcodehlt != opcodehlt && !new_opcodehlt) || (new_reghlt != reghlt
	           && !new_reghlt)) && new_dosyntaxhlt) {

		//if any syntax highlighting options have been turned off, clear those tags
		for (node = open_files; node; node = g_list_next (node)) {
			currfile = (file *)(node->data);
			gtk_text_buffer_get_bounds (currfile->textbuffer, &startiter, &enditer);

			if (!new_labelhlt)
				gtk_text_buffer_remove_tag (currfile->textbuffer, label_tag, &startiter, &enditer);
			if (!new_commenthlt)
				gtk_text_buffer_remove_tag (currfile->textbuffer, comment_tag, &startiter, &enditer);
			if (!new_dirhlt)
				gtk_text_buffer_remove_tag (currfile->textbuffer, directive_tag, &startiter, &enditer);
			if (!new_strhlt)
				gtk_text_buffer_remove_tag (currfile->textbuffer, string_tag, &startiter, &enditer);
			if (!new_opcodehlt)
				gtk_text_buffer_remove_tag (currfile->textbuffer, opcode_tag, &startiter, &enditer);
			if (!new_reghlt)
				gtk_text_buffer_remove_tag (currfile->textbuffer, register_tag, &startiter, &enditer);
		}

	} else if (new_dosyntaxhlt != dosyntaxhlt && !new_dosyntaxhlt) {

		//if syntax highlighting has been turned off, clear all tags for all open files
		for (node = open_files; node; node = g_list_next (node)) {
			currfile = (file *)(node->data);
			gtk_text_buffer_get_bounds (currfile->textbuffer, &startiter, &enditer);
			gtk_text_buffer_remove_all_tags (currfile->textbuffer, &startiter, &enditer);
		}
	}

	dosyntaxhlt = new_dosyntaxhlt;
	labelhlt = new_labelhlt;
	commenthlt = new_commenthlt;
	dirhlt = new_dirhlt;
	strhlt = new_strhlt;
	opcodehlt = new_opcodehlt;
	reghlt = new_reghlt;

	gtk_widget_destroy (dialog);
}


/*
 * Shows assembler options dialog
 */

void callback_setassembleropt_add (GtkButton *button, gpointer data) {
	GtkWidget *dialog;
	gchar *path;

	dialog = gtk_file_chooser_dialog_new ("Choose an include directory", GTK_WINDOW (window),
	                                      GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER, GTK_STOCK_CANCEL,
	                                      GTK_RESPONSE_CANCEL, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
	                                      NULL);
	gtk_file_chooser_set_select_multiple (GTK_FILE_CHOOSER (dialog), FALSE);

	if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_ACCEPT) {
		gtk_widget_destroy (dialog);
		return;
	}

	path = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
	gtk_list_store_insert_with_values (include_dirs, NULL, -1, 0, path, -1);
	g_free (path);
	gtk_widget_destroy (dialog);
}

void callback_setassembleropt_remove (GtkButton *button, gpointer list) {
	GtkTreeModel *templist;
	GtkTreeIter iter;

	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (list)),
	                                     &templist, &iter)) {
		gtk_list_store_remove (GTK_LIST_STORE (templist), &iter);
	}
}

void callback_setassembleropt (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog, *label, *vbox, *includedirlist, *cslabelson, *outputtypelist,
		*viewport, *button, *buttonbox, *autosaveon;
	GtkCellRenderer *textrenderer;
	GtkTreeViewColumn *column;
	GtkTreePath *path;
	GtkTreeIter selectediter;

	//build dialog, etc.
	dialog = gtk_dialog_new_with_buttons ("Assembler Settings", GTK_WINDOW (window),
				GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
				GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
	gtk_window_set_resizable (GTK_WINDOW (dialog), TRUE);

	vbox = gtk_vbox_new (FALSE, 6);

	cslabelson = gtk_check_button_new_with_label ("Labels are case-sensitive");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (cslabelson), labels_casesensitive);
	gtk_box_pack_start (GTK_BOX (vbox), cslabelson, FALSE, FALSE, 0);

	autosaveon = gtk_check_button_new_with_label ("Save before build without asking");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (autosaveon), save_before_build);
	gtk_box_pack_start (GTK_BOX (vbox), autosaveon, FALSE, FALSE, 0);

	label = gtk_label_new ("Include directories:");
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);

	includedirlist = gtk_tree_view_new_with_model (GTK_TREE_MODEL (include_dirs));
	textrenderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("", textrenderer, "text", 0, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (includedirlist), column);
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (includedirlist), FALSE);
	gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (includedirlist), TRUE);
	gtk_tree_selection_set_mode (gtk_tree_view_get_selection (GTK_TREE_VIEW (includedirlist)),
		GTK_SELECTION_BROWSE);
	viewport = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (viewport), GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (viewport), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (viewport), includedirlist);
	gtk_box_pack_start (GTK_BOX (vbox), viewport, TRUE, TRUE, 0);

	buttonbox = gtk_hbutton_box_new ();
	gtk_box_set_spacing (GTK_BOX (buttonbox), 6);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (buttonbox), GTK_BUTTONBOX_SPREAD);
	button = gtk_button_new_from_stock (GTK_STOCK_ADD);
	g_signal_connect (button, "clicked", callback_setassembleropt_add, NULL);
	gtk_container_add (GTK_CONTAINER (buttonbox), button);
	button = gtk_button_new_from_stock (GTK_STOCK_REMOVE);
	g_signal_connect (button, "clicked", callback_setassembleropt_remove, (gpointer)includedirlist);
	gtk_container_add (GTK_CONTAINER (buttonbox), button);
	gtk_box_pack_start (GTK_BOX (vbox), buttonbox, FALSE, FALSE, 0);

	label = gtk_label_new ("Output type:");
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);

	outputtypelist = gtk_tree_view_new_with_model (GTK_TREE_MODEL (output_types));
	textrenderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("", textrenderer, "text", 0, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (outputtypelist), column);
	column = gtk_tree_view_column_new_with_attributes ("", textrenderer, "text", 1, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (outputtypelist), column);
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (outputtypelist), FALSE);
	gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (outputtypelist), TRUE);
	gtk_tree_selection_set_mode (gtk_tree_view_get_selection (GTK_TREE_VIEW (outputtypelist)),
		GTK_SELECTION_BROWSE);
	path = gtk_tree_path_new_from_indices (output_type, -1);
	gtk_tree_selection_select_path (gtk_tree_view_get_selection (GTK_TREE_VIEW (outputtypelist)),
		path);
	gtk_tree_path_free (path);
	viewport = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (viewport), GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (viewport), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (viewport), outputtypelist);
	gtk_box_pack_start (GTK_BOX (vbox), viewport, TRUE, TRUE, 0);

	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), vbox);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 6);
	gtk_widget_show_all (vbox);

	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_REJECT) {
		gtk_widget_destroy (dialog);
		return;
	}

	//now, apply the settings
	gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (outputtypelist)),
		NULL, &selectediter);
	//(why does all this GtkTreeModel/GtkTreePath/GtkTreeIter crap have to be so complex???)
	path = gtk_tree_model_get_path (GTK_TREE_MODEL (output_types), &selectediter);
	output_type = *(gtk_tree_path_get_indices (path));
	gtk_tree_path_free (path);

	labels_casesensitive = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (cslabelson));
	save_before_build = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (autosaveon));

	gtk_widget_destroy (dialog);
}


/*
 * Toggles visibility of assembler
 * output window
 */

void callback_toggle_outputwin (GtkCheckMenuItem *menuitem, gpointer data) {

	if (menuitem->active)
		gtk_widget_show_all (outputwinvbox);
	else
		gtk_widget_hide (outputwinvbox);
}


/*
 * Stops syntax highlighting
 * on current file
 */

void callback_stop_syntaxhlt (GtkWidget *widget, gpointer data) {

	if (currfile->syntaxhltiter) {
		g_idle_remove_by_data ((gpointer)currfile);
		g_slice_free (GtkTextIter, currfile->syntaxhltiter);
		currfile->syntaxhltiter = NULL;
		gtk_text_view_set_editable (GTK_TEXT_VIEW (currfile->textview), TRUE);
	}
}
