/*
	Wabbitcode - An IDE for the Z80 line of TI calculators
	Copyright (C) 2007 Donald Straney

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <string.h>
#include <stdlib.h>
#include "asmide.h"
#include "file.h"
#include "callbacks.h"
#include "undo.h"
#include "project.h"
#include "macros.h"

//tab label for unsaved file
static const gchar *unsaved_str = "(Unsaved file)";
//recognized keywords
/* NOTE: run_syntax_hlt will stop once it finds the first match, so longer keywords must
	come before shorter ones in the same list.  For example, lddr must come before ldd,
	which must come before ld */
//TODO: put most-used opcodes first?
static const gchar *dot_directives[] = {"org", "db", "dw", "echo", "list", "nolist", "end",
	"fill", "block", "addinstr", "word", "byte", "show", "error", "seek", NULL};
static const gchar *pound_directives[] = {"define", "include", "ifdef", "ifndef", "if", "else",
	"endif", "endcomment", "defcont", "undefine", "comment", "macro", "endmacro", NULL};
static const gchar *opcodes[] = {"adc", "add", "and", "bit", "call", "ccf", "cpdr", "cpd", "cpir",
	"cpi", "cpl", "cp", "daa", "dec", "di", "djnz", "ei", "exx", "ex", "halt", "im", "inc",
	"indr", "ind", "inir", "ini", "in", "jp", "jr", "lddr", "ldd", "ldir", "ldi", "ld", "neg",
	"nop", "or", "otdr", "otir", "outd", "outi", "out", "pop", "push", "res", "reti", "retn",
	"ret", "rla", "rlca", "rlc", "rld", "rl", "rra", "rrca", "rrc", "rrd", "rr", "rst", "sbc",
	"scf", "set", "sla", "sll", "sra", "srl", "sub", "xor", NULL};
static const gchar *registers[] = {"af", "bc", "de", "hl", "ixh", "ixl", "iyh", "iyl", "ix",
	"iy", "a", "f", "b", "c", "d", "e", "h", "l", "r", "sp", "pc", "i", "af'", NULL};


/*
 * Gets word at iter (enditer is
 * end of line) and moves
 * iter to end of word, returns
 * allocated text
 */

gchar *get_word (GtkTextIter *iter, GtkTextIter *enditer) {
	GtkTextIter wordstartiter;
	gunichar currchar;
	gchar *temp_word, *word;

	//save the start of the word
	wordstartiter = *iter;
	//now find the end
	currchar = gtk_text_iter_get_char (iter);
	while ((g_unichar_isalpha (currchar) ||
		   currchar == underscore_unichar) &&
		   !gtk_text_iter_equal (iter, enditer)) {

		gtk_text_iter_forward_char (iter);
		currchar = gtk_text_iter_get_char (iter);
	}
	//get the word
//	return gtk_text_buffer_get_text (gtk_text_iter_get_buffer (iter), &wordstartiter, iter, FALSE);
	temp_word = gtk_text_buffer_get_text (gtk_text_iter_get_buffer (iter), &wordstartiter, iter, FALSE);
	word = g_ascii_strdown (temp_word, -1);
	g_free (temp_word);

	return word;
}


/*
 * Tries to match word with
 * entry in wordlist, returns
 * TRUE if successful
 */

gboolean match_word (gchar *word, const gchar *wordlist[]) {
	int currword;

	currword = 0;
	while (wordlist[currword]) {
		if (!strcmp (wordlist[currword], word))
			return TRUE;

		currword++;
	}

	return FALSE;
}


/*
 * Runs syntax highlighting
 * on line with iter
 */

void run_syntax_hlt (GtkTextIter *iter) {
	GtkTextIter startiter, enditer, tempiter;
	GtkTextBuffer *textbuf;
	gunichar currchar;
	gchar *temp_word;
	gboolean opcode_can_come_next = FALSE, directive_can_come_next = TRUE,
		register_can_come_next = FALSE;

	textbuf = gtk_text_iter_get_buffer (iter);

	//get the start and end of the current line
	startiter = *iter;
	gtk_text_iter_set_line_offset (&startiter, 0);
	enditer = startiter;
	if (!gtk_text_iter_ends_line (&enditer))
		gtk_text_iter_forward_to_line_end (&enditer);

	//if there's nothing on the line, we don't have anything to do, so return!
	if (gtk_text_iter_equal (&startiter, &enditer))
		return;

	//first, clear all existing tags
	gtk_text_buffer_remove_all_tags (textbuf, &startiter, &enditer);

	//now process the line, looking for each enabled tag
	//if the first char isn't a space,
	currchar = gtk_text_iter_get_char (&startiter);
	if (g_unichar_isalnum (currchar) || currchar == underscore_unichar) {
		//then we've got a label on our hands!
		//store the beginning of the label and find its end
		tempiter = startiter;
		gtk_text_iter_forward_char (&startiter);
		currchar = gtk_text_iter_get_char (&startiter);
		while (currchar != colon_unichar &&
			   !gtk_text_iter_equal (&startiter, &enditer) &&
			   !g_unichar_isspace (currchar)) {

			gtk_text_iter_forward_char (&startiter);
			currchar = gtk_text_iter_get_char (&startiter);
		}

		if (currchar == colon_unichar)
			gtk_text_iter_forward_char (&startiter);

		//then apply the label tag
		if (labelhlt)
			gtk_text_buffer_apply_tag (textbuf, label_tag, &tempiter, &startiter);

		directive_can_come_next = FALSE;
	}

	//now loop, looking for any other keywords in the line
	while (!gtk_text_iter_equal (&startiter, &enditer)) {
		//skip whitespace
		currchar = gtk_text_iter_get_char (&startiter);
		while (g_unichar_isspace (currchar) && !gtk_text_iter_equal (&startiter, &enditer)) {
			gtk_text_iter_forward_char (&startiter);
			currchar = gtk_text_iter_get_char (&startiter);
			opcode_can_come_next = register_can_come_next = directive_can_come_next = TRUE;
		}

		if (gtk_text_iter_equal (&startiter, &enditer))
			break;

		if (currchar == leftparen_unichar || currchar == comma_unichar) {
			gtk_text_iter_forward_char (&startiter);
			register_can_come_next = TRUE;
			opcode_can_come_next = directive_can_come_next = FALSE;
			continue;
		}

		if (currchar == rightparen_unichar) {
			gtk_text_iter_forward_char (&startiter);
			register_can_come_next = opcode_can_come_next = directive_can_come_next = FALSE;
			continue;
		}

		//if there's a word starting with a . or #, it could be a directive
		if ((currchar == dot_unichar || currchar == pound_unichar) && directive_can_come_next) {
			tempiter = startiter;

			if (gtk_text_iter_forward_char (&startiter)) {
				temp_word = get_word (&startiter, &enditer);

				if (dirhlt && currchar == dot_unichar && match_word (temp_word, dot_directives))
					gtk_text_buffer_apply_tag (textbuf, directive_tag, &tempiter, &startiter);

				if (dirhlt && currchar == pound_unichar && match_word (temp_word, pound_directives))
					gtk_text_buffer_apply_tag (textbuf, directive_tag, &tempiter, &startiter);

				register_can_come_next = opcode_can_come_next = directive_can_come_next = FALSE;
				g_free (temp_word);
				continue;
			}
		}

		//if there's a quote, it's a string
		if (currchar == quote_unichar) {
			//save the beginning of the string
			tempiter = startiter;
			//then find where it ends
			gtk_text_iter_forward_char (&startiter);
			while (gtk_text_iter_get_char (&startiter) != quote_unichar &&
				   !gtk_text_iter_equal (&startiter, &enditer)) {

				gtk_text_iter_forward_char (&startiter);
			}

			if (gtk_text_iter_get_char (&startiter) == quote_unichar)
				gtk_text_iter_forward_char (&startiter);

			//finally, apply the tag
			if (strhlt)
				gtk_text_buffer_apply_tag (textbuf, string_tag, &tempiter, &startiter);

			register_can_come_next = opcode_can_come_next = directive_can_come_next = FALSE;
			continue;
		}

		//if there's a ;, just make everything to the end of the line a comment
		if (currchar == semicolon_unichar) {
			if (commenthlt)
				gtk_text_buffer_apply_tag (textbuf, comment_tag, &startiter, &enditer);

			//that's all for now, folks!
			return;
		}

		//if it's not anything, try seeing if it's a opcode or register
		if (g_unichar_isalpha (currchar)) {
			tempiter = startiter;
			temp_word = get_word (&startiter, &enditer);

			if (opcodehlt && opcode_can_come_next && match_word (temp_word, opcodes))
				gtk_text_buffer_apply_tag (textbuf, opcode_tag, &tempiter, &startiter);

			if (reghlt && register_can_come_next && match_word (temp_word, registers))
				gtk_text_buffer_apply_tag (textbuf, register_tag, &tempiter, &startiter);

			g_free (temp_word);

		} else {
			//if it's nothing we recognize, just skip this char
			gtk_text_iter_forward_char (&startiter);
		}
		register_can_come_next = opcode_can_come_next = directive_can_come_next = FALSE;
	}
}


/*
 * Idle callback for running
 * syntax highlighting on a file
 */

gboolean callback_run_syntax_hlt (gpointer currfile) {
	GtkTextIter *textiter;

	textiter = ((file *)currfile)->syntaxhltiter;
	run_syntax_hlt (textiter);

	if (gtk_text_iter_forward_line (textiter)) {
		return TRUE;
	} else {
		g_slice_free (GtkTextIter, textiter);
		((file *)currfile)->syntaxhltiter = NULL;
		gtk_text_view_set_editable (GTK_TEXT_VIEW (((file *)currfile)->textview), TRUE);
		return FALSE;
	}
}


/*
 * Runs syntax highlighting
 * on an entire file by setting
 * up an idle callback to do
 * one line at a time
 */

void run_syntax_hlt_on_file (file *newfile) {
	GtkTextIter *newiter;

	newiter = g_slice_new (GtkTextIter);
	newfile->syntaxhltiter = newiter;
	gtk_text_buffer_get_start_iter (newfile->textbuffer, newiter);
	gtk_text_view_set_editable (GTK_TEXT_VIEW (newfile->textview), FALSE);
	g_idle_add_full (G_PRIORITY_DEFAULT_IDLE - 70, callback_run_syntax_hlt, (gpointer)newfile, NULL);
}


/*
 * Adds appropriate filters to file chooser
 */

void add_file_chooser_filters (GtkFileChooser *filechooser) {
	GtkFileFilter *filter_code, *filter_inc, *filter_all,
		*filter_supported, *filter_project;

	//one for all supported types...
	filter_supported = gtk_file_filter_new ();
	gtk_file_filter_set_name (filter_supported, "Supported files");
	gtk_file_filter_add_pattern (filter_supported, "*.asm");
	gtk_file_filter_add_pattern (filter_supported, "*.z80");
	gtk_file_filter_add_pattern (filter_supported, "*.inc");
	gtk_file_filter_add_pattern (filter_supported, "*.h");
	gtk_file_filter_add_pattern (filter_supported, "*"PROJECT_EXTENSION);

	//one for projects...
	filter_project = gtk_file_filter_new ();
	gtk_file_filter_set_name (filter_project, "Projects");
	gtk_file_filter_add_pattern (filter_project, "*"PROJECT_EXTENSION);

	//one for code files...
	filter_code = gtk_file_filter_new ();
	gtk_file_filter_set_name (filter_code, "Source code");
	gtk_file_filter_add_pattern (filter_code, "*.asm");
	gtk_file_filter_add_pattern (filter_code, "*.z80");

	//one for include files...
	filter_inc = gtk_file_filter_new ();
	gtk_file_filter_set_name (filter_inc, "Include files");
	gtk_file_filter_add_pattern (filter_inc, "*.inc");
	gtk_file_filter_add_pattern (filter_inc, "*.h");

	//...and one for everything, just in case
	filter_all = gtk_file_filter_new ();
	gtk_file_filter_set_name (filter_all, "All files");
	gtk_file_filter_add_pattern (filter_all, "*");

	//now add the filters
	gtk_file_chooser_add_filter (filechooser, filter_supported);
	gtk_file_chooser_add_filter (filechooser, filter_project);
	gtk_file_chooser_add_filter (filechooser, filter_code);
	gtk_file_chooser_add_filter (filechooser, filter_inc);
	gtk_file_chooser_add_filter (filechooser, filter_all);
	//let's set it on all supported types by default
	gtk_file_chooser_set_filter (filechooser, filter_supported);
}


/*
 * Returns static string to use
 * on tab label for file
 */

const gchar *get_file_label (file *openfile) {
	const gchar *path;
	int pos;

	path = openfile->path;
	if (path) {
		//go backwards through the path until we find a dir separator
		for (pos = strlen (path) - 1; pos >= 0 && !G_IS_DIR_SEPARATOR (path[pos]); pos--);
		path += pos + 1;

	} else {
		path = unsaved_str;
	}

	return path;
}


/*
 * Called when file's name or
 * modified status changed,
 * updates title in tab and
 * window title
 */

void callback_update_tab_label (GtkTextBuffer *widget, gpointer data) {
	GtkWidget *tablabel;
	gchar *labeltext;

	tablabel = ((file *)data)->tablabel;
	if (gtk_text_buffer_get_modified (widget)) {
		labeltext = g_strconcat ("*", get_file_label ((file *)data), NULL);
		gtk_label_set_text (GTK_LABEL (tablabel), labeltext);
		gtk_window_set_title (GTK_WINDOW (window), labeltext);
		g_free (labeltext);
	} else {
		labeltext = (gchar *)get_file_label ((file *)data);
		gtk_label_set_text (GTK_LABEL (tablabel), labeltext);
		gtk_window_set_title (GTK_WINDOW (window), labeltext);
	}
}


/*
 * Called when cursor moved -
 * updates line and column
 * numbers in statusbar
 */

void callback_cursor_moved (GtkTextBuffer *textbuffer, GtkTextIter *iter,
							GtkTextMark *mark, gpointer data) {

	const gchar *markname;

	//make sure it's the cursor that moved
	markname = gtk_text_mark_get_name (mark);
	if (markname && !strcmp (markname, "insert"))
		update_statusbar (iter);
}


/*
 * Called when text inserted -
 * adds action to undo stack
 * and macro
 */

void callback_text_inserted (GtkTextBuffer *textbuffer, GtkTextIter *start, gchar *text,
							 gint len, gpointer data) {

	if (!ignore_actions)
		add_undo_action (ACTION_INSERT, gtk_text_iter_get_offset (start), g_strndup (text, len));

	if (recording_macro)
		add_macro_insert (text);
}


/*
 * Called after text inserted -
 * updates syntax highlighting
 */

void callback_text_inserted_after (GtkTextBuffer *textbuffer, GtkTextIter *start, gchar *text,
								   gint len, gpointer data) {

	int currpos, numlines = 1;
	GtkTextIter tempiter, startiter, enditer;
	gunichar curr_char, newline_char;
	gchar *indent_text;

	if (dosyntaxhlt) {
		//we need to find out how many lines have been inserted
		// (pasted from another program, for example)
		for (currpos = 0; currpos < len; currpos++) {
			if (text[currpos] == '\n')
				numlines++;
		}

		if (numlines == 1) {
			//if there's only one line, no need for the extra overhead
			run_syntax_hlt (start);
		} else {
			//otherwise, go through and run the syntax highlighting on every inserted line
			tempiter = *start;
			while (numlines > 0) {
				run_syntax_hlt (&tempiter);
				gtk_text_iter_backward_line (&tempiter);
				numlines--;
			}
		}
	}

	//do auto-indent if needed
	if (!ignore_actions && autoindent && text[0] == '\n' && text[1] == '\0') {
		gtk_text_buffer_get_iter_at_line (textbuffer, &startiter, gtk_text_iter_get_line (start) - 1);
		enditer = startiter;

		//find the end of the indent
		curr_char = gtk_text_iter_get_char (&enditer);
		newline_char = g_utf8_get_char ("\n");
		while (g_unichar_isspace (curr_char) && curr_char != newline_char) {
			gtk_text_iter_forward_char (&enditer);
			curr_char = gtk_text_iter_get_char (&enditer);
		}

		//if there actually was an indent on the last line,
		if (!gtk_text_iter_equal (start, &enditer)) {
			//then copy it and insert it at the cursor
			indent_text = gtk_text_buffer_get_text (textbuffer, &startiter, &enditer, FALSE);
			gtk_text_buffer_insert_at_cursor (textbuffer, indent_text, -1);
			g_free (indent_text);
		}

		//the cursor position has changed, so the "start" iter isn't valid anymore
		update_statusbar (NULL);
	} else {
		//update cursor position too
		update_statusbar (start);
	}
}


/*
 * Called when text deleted -
 * adds action to undo stack
 */

void callback_text_deleted (GtkTextBuffer *textbuffer, GtkTextIter *start,
							GtkTextIter *end, gpointer data) {

	if (!ignore_actions)
		add_undo_action (ACTION_DELETE, gtk_text_iter_get_offset (start),
			gtk_text_buffer_get_text (textbuffer, start, end, FALSE));
}


/*
 * Called after text deleted -
 * updates syntax highlighting
 */

void callback_text_deleted_after (GtkTextBuffer *textbuffer, GtkTextIter *start,
								  GtkTextIter *end, gpointer data) {

	//run syntax highlighting if necessary
	if (dosyntaxhlt)
		run_syntax_hlt (start);

	//update statusbar text with cursor position
	update_statusbar (start);
}


/*
 * Called when key pressed -
 * waits for tab to take care
 * of indenting/unindenting
 * multiple lines
 */

gboolean callback_key_pressed (GtkWidget *widget, GdkEventKey *event, gpointer data) {
	guint mask;
	int endline;
	GtkTextIter startiter, enditer;
	gunichar unichar_tab;

	//if we're recording a macro, add this keypress
	if (recording_macro && event->state == 0)
		add_macro_keypress (event->keyval);

	if ((event->keyval == GDK_Tab || event->keyval == GDK_KP_Tab ||
	    event->keyval == GDK_ISO_Left_Tab) &&
	    gtk_text_buffer_get_has_selection (currfile->textbuffer)) {

		//handle indenting/un-indenting selected lines with tab/shift-tab
		mask = event->state & (GDK_SHIFT_MASK | GDK_LOCK_MASK | GDK_CONTROL_MASK | GDK_MOD1_MASK);
		if (mask == 0) {
			//indent
			gtk_text_buffer_begin_user_action (currfile->textbuffer);
			gtk_text_buffer_get_selection_bounds (currfile->textbuffer, &startiter, &enditer);
			endline = gtk_text_iter_get_line (&enditer);
			gtk_text_iter_set_line_offset (&startiter, 0);
			do {
				gtk_text_buffer_insert (currfile->textbuffer, &startiter, "\t", -1);
			} while (gtk_text_iter_forward_line (&startiter) &&
			         gtk_text_iter_get_line (&startiter) <= endline);

			gtk_text_buffer_end_user_action (currfile->textbuffer);
			return TRUE;

		} else if (mask == GDK_SHIFT_MASK) {
			//unindent
			gtk_text_buffer_begin_user_action (currfile->textbuffer);
			gtk_text_buffer_get_selection_bounds (currfile->textbuffer, &startiter, &enditer);
			endline = gtk_text_iter_get_line (&enditer);
			unichar_tab = g_utf8_get_char ("\t");
			gtk_text_iter_set_line_offset (&startiter, 0);
			do {
				if (gtk_text_iter_get_char (&startiter) == unichar_tab) {
					enditer = startiter;
					gtk_text_iter_forward_char (&enditer);
					gtk_text_buffer_delete (currfile->textbuffer, &startiter, &enditer);
				}
			} while (gtk_text_iter_forward_line (&startiter) &&
			         gtk_text_iter_get_line (&startiter) <= endline);

			gtk_text_buffer_end_user_action (currfile->textbuffer);
			return TRUE;
		}
	}

	return FALSE;
}


/*
 * Used to make tab
 * labels thin, copied
 * from gedit 2.18.1
 */

void tab_label_style_set_cb (GtkWidget *hbox, GtkStyle *previous_style, gpointer data) {
	gint h, w;

	gtk_icon_size_lookup_for_settings (gtk_widget_get_settings (hbox), GTK_ICON_SIZE_MENU, &w, &h);
	gtk_widget_set_size_request (GTK_WIDGET (data), w + 2, h + 2);
}


/*
 * Creates new open file and
 * sets contents, path
 */

void add_file (gchar *contents, gint len, gchar *path) {
	file *newfile;
	GtkWidget *image, *hbox, *closebutton;
	GtkRcStyle *rcstyle;

	//create a new file struct
	newfile = g_slice_new (file);
	if (path)
		newfile->path = strdup (path);
	else
		newfile->path = NULL;

	//create a new textview widget for it
	newfile->textbuffer = gtk_text_buffer_new (tagtable);
	newfile->textview = gtk_text_view_new_with_buffer (newfile->textbuffer);
	gtk_widget_modify_font (newfile->textview, editorfont);
	gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (newfile->textview),
		wordwrap ? GTK_WRAP_WORD_CHAR : GTK_WRAP_NONE);
	gtk_text_view_set_tabs (GTK_TEXT_VIEW (newfile->textview), tabstops);
	gtk_text_buffer_set_text (newfile->textbuffer, contents, len);
	gtk_text_buffer_set_modified (newfile->textbuffer, FALSE);
	newfile->scrolledwin = gtk_scrolled_window_new (NULL, NULL);
	gtk_container_add (GTK_CONTAINER (newfile->scrolledwin), newfile->textview);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (newfile->scrolledwin), GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (newfile->scrolledwin), GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC);

	//init its undo stack
	newfile->undo_stack = newfile->undo_stack_end = NULL;
	newfile->undo_stack_pos = newfile->undo_stack_len = 0;

	//add its tab label + close button
	newfile->tablabel = gtk_label_new (get_file_label (newfile));
	closebutton = gtk_button_new ();
	image = gtk_image_new_from_stock (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
	gtk_container_add (GTK_CONTAINER (closebutton), image);
	gtk_button_set_relief (GTK_BUTTON (closebutton), GTK_RELIEF_NONE);
	gtk_button_set_focus_on_click (GTK_BUTTON (closebutton), FALSE);
	//some stuff to make sure the close button's as small as possible - copied from gedit 2.18.1
	rcstyle = gtk_rc_style_new ();
	rcstyle->xthickness = rcstyle->ythickness = 0;
	gtk_widget_modify_style (closebutton, rcstyle);
	gtk_rc_style_unref (rcstyle);
	gtk_misc_set_padding (GTK_MISC (newfile->tablabel), 0, 0);

	hbox = gtk_hbox_new (FALSE, 10);
	gtk_box_pack_start (GTK_BOX (hbox), newfile->tablabel, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (hbox), closebutton, FALSE, FALSE, 0);

	//more hack-ish stuff from gedit to get those thin tab labels
	g_signal_connect (hbox, "style-set", G_CALLBACK (tab_label_style_set_cb), (gpointer)closebutton);

	gtk_widget_show_all (newfile->scrolledwin);
	gtk_widget_show_all (hbox);

	//add this to the list of open files
	open_files = g_list_prepend (open_files, newfile);

	//make sure file options are active
	update_file_widgets (TRUE);

	//add a notebook tab for this file
	//(yeah, yeah, I know it should check the return value, but when's this going to fail?)
	gtk_notebook_insert_page (GTK_NOTEBOOK (notebook), newfile->scrolledwin, hbox, -1);
	gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), -1);
	gtk_widget_grab_focus (newfile->textview);

	//add callbacks for file
	g_signal_connect (newfile->textbuffer, "modified-changed", G_CALLBACK (callback_update_tab_label),
		(gpointer)newfile);
	g_signal_connect (closebutton, "clicked", G_CALLBACK (callback_close), (gpointer)newfile);
	g_signal_connect (newfile->textbuffer, "mark-set", G_CALLBACK (callback_cursor_moved), NULL);
	g_signal_connect (newfile->textbuffer, "begin-user-action",
		G_CALLBACK (callback_start_user_action), NULL);
	g_signal_connect (newfile->textbuffer, "end-user-action",
		G_CALLBACK (callback_end_user_action), NULL);
	g_signal_connect (newfile->textbuffer, "delete-range", G_CALLBACK (callback_text_deleted), NULL);
	g_signal_connect (newfile->textbuffer, "insert-text", G_CALLBACK (callback_text_inserted), NULL);
	g_signal_connect_after (newfile->textbuffer, "delete-range",
		G_CALLBACK (callback_text_deleted_after), NULL);
	g_signal_connect_after (newfile->textbuffer, "insert-text",
		G_CALLBACK (callback_text_inserted_after), NULL);
	g_signal_connect (newfile->textview, "key-press-event", G_CALLBACK (callback_key_pressed), NULL);

	//now run syntax highlighting on it if it's needed
	if (dosyntaxhlt)
		run_syntax_hlt_on_file (newfile);
	else
		newfile->syntaxhltiter = NULL;
}


/*
 * Removes file, prompting for
 * save if unsaved changes -
 * returns TRUE if file was
 * removed successfully
 */

gboolean remove_file (file *oldfile) {
	GtkWidget *dialog;
	gint response;

	if (gtk_text_buffer_get_modified (oldfile->textbuffer)) {
		//if the file has unsaved changes, ask the user what to do
		dialog = gtk_message_dialog_new (GTK_WINDOW (window), GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION,
			GTK_BUTTONS_NONE, "Do you want to save the changes\nto %s before closing?",
			get_file_label (oldfile));
		gtk_dialog_add_buttons (GTK_DIALOG (dialog), "Close _without saving", GTK_RESPONSE_NO,
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_YES, NULL);
		response = gtk_dialog_run (GTK_DIALOG (dialog));
		gtk_widget_destroy (dialog);

		if (response == GTK_RESPONSE_CANCEL) {
			return FALSE;
		} else if (response == GTK_RESPONSE_YES) {
			if (!save_file (oldfile, FALSE))
				//well, I guess they don't want to save after all!
				return FALSE;
		}
	}

	//take the page out of the notebook,
	gtk_notebook_remove_page (GTK_NOTEBOOK (notebook),
		gtk_notebook_page_num (GTK_NOTEBOOK (notebook), oldfile->scrolledwin));
	//stop the syntax highlighting callback (if any),
	if (oldfile->syntaxhltiter) {
		g_idle_remove_by_data ((gpointer)oldfile);
		g_slice_free (GtkTextIter, oldfile->syntaxhltiter);
	}
	//remove the file from the list,
	open_files = g_list_remove (open_files, oldfile);
	//free its data,
	free_undo_stack (oldfile->undo_stack);
	g_free (oldfile->path);
	g_slice_free (file, oldfile);

	//and hide the assembler output window, clear the status bar, and deactivate
	// the file-specific options if no other files are open
	if (!open_files) {
		currfile = NULL;
		set_statusbar_text ("");
		update_file_widgets (FALSE);
		gtk_window_set_title (GTK_WINDOW (window), APP_NAME);
	}
	return TRUE;
}


/*
 * Tries to close all files, returns
 * TRUE if successful or FALSE if canceled
 */

gboolean remove_all_files () {

	//while there are still open files,
	while (open_files) {
		//try to close the first one in the list
		if (!remove_file (open_files->data))	//TODO: g_list_get_first or something?
			//if the user canceled, stop
			return FALSE;
	}

	return TRUE;
}


/*
 * Saves file, showing "save" dialog
 * if path isn't set already (unless
 * save_as is set, then always shown) -
 * returns TRUE if saved successfully
 */

gboolean save_file (file *savefile, gboolean save_as) {
	GtkWidget *dialog;
	GError *error = NULL;
	GtkTextIter startiter, enditer;
	gint response;
	gchar *path, *contents, *fullpath;
	gboolean showdialog;

	//decide if we need to show the save dialog
	showdialog = !savefile->path || save_as;

	if (showdialog) {
		//if the file hasn't been saved yet, give the user a chance to do that
		dialog = gtk_file_chooser_dialog_new ("Save File", GTK_WINDOW (window),
			GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);

		if (savefile->path)
			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (dialog), savefile->path);
		else if (project_open)
			gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), project_path);

		gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
		add_file_chooser_filters (GTK_FILE_CHOOSER (dialog));
		response = gtk_dialog_run (GTK_DIALOG (dialog));

		if (response != GTK_RESPONSE_ACCEPT) {
			gtk_widget_destroy (dialog);
			return FALSE;
		}

		path = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
		gtk_widget_destroy (dialog);

	} else {
		path = savefile->path;
	}

	//now actually try to save it
	gtk_text_buffer_get_bounds (savefile->textbuffer, &startiter, &enditer);
	contents = gtk_text_buffer_get_text (savefile->textbuffer, &startiter, &enditer, FALSE);
	if (!g_file_set_contents (path, contents, -1, &error)) {

		g_free (contents);
		if (showdialog)
			g_free (path);

		//if it didn't work, show an error
		show_error ("Couldn't save the file", error->message);
		g_error_free (error);
		return FALSE;
	}

	g_free (contents);

	//update the file's name if it needs it
	if (showdialog) {
		if (savefile->path)
			g_free (savefile->path);
		savefile->path = path;
		fullpath = g_filename_to_uri (path, NULL, NULL);
		gtk_recent_manager_add_item (recentfiles, fullpath);
		g_free (fullpath);
		//make sure that the labels get updated when saved under a different name even if no changes
		if (!gtk_text_buffer_get_modified (savefile->textbuffer))
			callback_update_tab_label (savefile->textbuffer, (gpointer)savefile);
		//if a project's open, refresh the sidebar to make sure any new project files are shown
		if (project_open)
			callback_refresh_project_files (NULL, NULL);
	}
	//and show that it's been saved successfully
	gtk_text_buffer_set_modified (savefile->textbuffer, FALSE);
	return TRUE;
}


/*
 * Reads in a file, returns
 * TRUE if successful
 */

gboolean read_file (gchar *path, gchar **contents, guint *len) {
	gchar *fullpath;
	GError *error = NULL;

	//first try to actually open it
	if (!g_file_get_contents (path, contents, len, &error)) {
		show_error ("Couldn't open file", error->message);
		g_error_free (error);
		return FALSE;
	}

	//then check the encoding
	if (!g_utf8_validate (*contents, *len, NULL)) {
		show_error ("This doesn't appear to be a text file", "It's not in UTF-8 encoding");
		g_free (*contents);
		return FALSE;
	}

	//if all that worked, add it to the recent files list
	fullpath = g_filename_to_uri (path, NULL, NULL);
	gtk_recent_manager_add_item (recentfiles, fullpath);
	g_free (fullpath);

	return TRUE;
}


/*
 * Opens a file, adds to list
 * of open files if normal
 * source file, handles appropriately
 * if project file
 */

void open_file (gchar *path) {
	gchar *contents;
	guint len;

	if (!is_project_file (path)) {
		//if it's a normal file, just read it in and add to list of open files
		if (read_file (path, &contents, &len)) {
			add_file (contents, len, path);
			g_free (contents);
		} else {
			return;
		}
	} else {
		//if it's a project, let the project-specific code take care of it
		open_project (path);
	}
}

