/*
	Wabbitcode - An IDE for the Z80 line of TI calculators
	Copyright (C) 2007 Donald Straney

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glib/gstdio.h>
#include <string.h>
#include "asmide.h"
#include "macros.h"

//name of directory where macros are stored
#define MACRO_DIR ".wabbitcode-macros"

//TRUE if recording macro
gboolean recording_macro;
//list of actions in current macro
macro_step *curr_macro;
macro_step *curr_macro_tail;
//string representing current macro
gchar *curr_macro_string;


/*
 * Frees current macro
 */

void free_curr_macro () {
	macro_step *next_step;

	while (curr_macro) {
		if (curr_macro->action == MACRO_INSERT)
			g_free (curr_macro->text);
		next_step = curr_macro->next;
		g_slice_free (macro_step, curr_macro);
		curr_macro = next_step;
	}

	if (curr_macro_string)
		g_free (curr_macro_string);
}


/*
 * Creates macro directory
 * if it doesn't exist
 */

void create_macro_dir () {
	gchar *path;

	path = g_build_filename (g_get_home_dir (), MACRO_DIR, NULL);
	g_mkdir_with_parents (path, 0755);
	g_free (path);
}


/*
 * Called by "record" menu item
 */

void callback_record_macro_toggled_ind (GtkCheckMenuItem *menuitem, gpointer widget) {
	/* toggles macro recording indirectly
	   by toggling the "record macro"
	   toolbar button (so the button and
	   menu option don't get out-of-sync) */
	gtk_toggle_tool_button_set_active (GTK_TOGGLE_TOOL_BUTTON (widget), menuitem->active);
}


/*
 * Starts or stops recording macro
 * when "record" toolbar button is pressed
 */

void callback_record_macro_toggled (GtkToggleToolButton *toolbutton, gpointer data) {
	GtkWidget *dialog, *hbox, *label, *entry;
	gchar *fullpath;
	macro_step *next_step;
	FILE *macrofile;
	gboolean get_macro_name = TRUE;

	recording_macro = gtk_toggle_tool_button_get_active (GTK_TOGGLE_TOOL_BUTTON (toolbutton));

	if (recording_macro) {
		//if recording has started, init the action list
		curr_macro = curr_macro_tail = NULL;
		curr_macro_string = NULL;
	} else {
		while (get_macro_name) {
			//if recording has stopped, get a name for the macro
			dialog = gtk_dialog_new_with_buttons ("Enter a name for the macro", GTK_WINDOW (window),
						GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
						GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
			gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
			gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

			hbox = gtk_hbox_new (FALSE, 0);
			label = gtk_label_new ("Name for this macro:");
			entry = gtk_entry_new ();
			gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);
			gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 6);
			gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 6);
			gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, TRUE, TRUE, 6);
			gtk_widget_show_all (hbox);

			if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_ACCEPT) {
				gtk_widget_destroy (dialog);
				free_curr_macro ();
				return;
			}

			fullpath = g_build_filename (g_get_home_dir (), MACRO_DIR,
			                             gtk_entry_get_text (GTK_ENTRY (entry)), NULL);
			gtk_widget_destroy (dialog);
			create_macro_dir ();

			get_macro_name = FALSE;
			if (g_file_test (fullpath, G_FILE_TEST_EXISTS)) {

				//if there's already one with the same name, ask the user what to do
				dialog = gtk_message_dialog_new (GTK_WINDOW (window), GTK_DIALOG_MODAL,
				                                 GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO,
				                                 "There is already a macro with this name");
				gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
				                                          "Would you like to replace it?");

				if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_YES) {
					get_macro_name = TRUE;
					g_free (fullpath);
				}

				gtk_widget_destroy (dialog);
			}
		}

		//and save it
		macrofile = fopen (fullpath, "w");
		g_free (fullpath);

		if (!macrofile) {
			show_error ("Couldn't save macro", "Dude, your system is screwed");
			free_curr_macro ();
			return;
		}

		while (curr_macro) {
			fputc (curr_macro->action, macrofile);
			if (curr_macro->action == MACRO_INSERT) {
				fputc (strlen (curr_macro->text), macrofile);
				fputs (curr_macro->text, macrofile);
				g_free (curr_macro->text);
			}
			//since we're going to free it next anyways, might as well do it now and save time
			next_step = curr_macro->next;
			g_slice_free (macro_step, curr_macro);
			curr_macro = next_step;
		}

		fclose (macrofile);

		if (curr_macro_string)
			g_free (curr_macro_string);
		update_statusbar (NULL);
	}
}


/*
 * Creates GtkListStore filled with
 * names of available macros, returns
 * pointer or NULL on error
 */

GtkListStore *get_macro_list () {
	GDir *macro_dir;
	GError *error = NULL;
	gchar *path, *name;
	GtkListStore *macro_list;

	//first, make sure the macro directory is there
	create_macro_dir ();

	//now scan for macros
	path = g_build_filename (g_get_home_dir (), MACRO_DIR, NULL);
	macro_dir = g_dir_open (path, 0, &error);
	g_free (path);
	if (!macro_dir) {
		show_error ("Can't read macros", error->message);
		g_error_free (error);
		return NULL;
	}

	name = (gchar *)g_dir_read_name (macro_dir);
	if (!name) {
		show_info ("No macros have been recorded yet");
		return NULL;
	}
	macro_list = gtk_list_store_new (1, G_TYPE_STRING);
	while (name) {
		gtk_list_store_insert_with_values (macro_list, NULL, -1, 0, name, -1);
		name = (gchar *)g_dir_read_name (macro_dir);
	}
	g_dir_close (macro_dir);

	return macro_list;
}


/*
 * Called when "run macro" commands
 * on toolbar or menu are used -
 * prompts for macro to run and
 * number of times to run
 */

void callback_run_macro (GtkWidget *widget, gpointer data) {
	GMappedFile *macrofile;
	GError *error = NULL;
	gsize len;
	gchar *start, *curr_pos, *path, *name;
	gint insert_len, offset, num_cycles;
	GtkTextIter iter, iter_bound;
	GtkWidget *dialog, *vbox, *hbox, *label, *spinbutton, *list, *viewport;
	GtkListStore *macro_list;
	GtkCellRenderer *textrenderer;
	GtkTreeViewColumn *column;
	GtkTreePath *treepath;
	GtkTreeIter treeiter;

	//get the list of macros first
	macro_list = get_macro_list ();
	if (!macro_list)
		return;

	//build the dialog
	dialog = gtk_dialog_new_with_buttons ("Run macro", GTK_WINDOW (window), GTK_DIALOG_MODAL,
	                                      GTK_STOCK_OK, GTK_RESPONSE_OK, GTK_STOCK_CANCEL,
	                                      GTK_RESPONSE_CANCEL, NULL);
	gtk_window_set_resizable (GTK_WINDOW (dialog), TRUE);
	vbox = gtk_vbox_new (FALSE, 6);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 6);

	label = gtk_label_new ("Choose a macro to run:");
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);

	list = gtk_tree_view_new_with_model (GTK_TREE_MODEL (macro_list));
	textrenderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("", textrenderer, "text", 0, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (list), column);
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (list), FALSE);
	gtk_tree_selection_set_mode (gtk_tree_view_get_selection (GTK_TREE_VIEW (list)),
	                             GTK_SELECTION_BROWSE);
	treepath = gtk_tree_path_new_first ();
	gtk_tree_selection_select_path (gtk_tree_view_get_selection (GTK_TREE_VIEW (list)),
	                                treepath);
	gtk_tree_path_free (treepath);
	viewport = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (viewport), GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (viewport), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (viewport), list);
	gtk_box_pack_start (GTK_BOX (vbox), viewport, TRUE, TRUE, 0);

	hbox = gtk_hbox_new (FALSE, 6);
	label = gtk_label_new ("Run it");
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
	spinbutton = gtk_spin_button_new_with_range (1.0, 1000.0, 1.0);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON (spinbutton), 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton), 1.0);
	gtk_box_pack_start (GTK_BOX (hbox), spinbutton, FALSE, FALSE, 0);
	label = gtk_label_new ("time(s)");
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), vbox);
	gtk_widget_show_all (vbox);

	if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_OK) {
		gtk_widget_destroy (dialog);
		return;
	}

	num_cycles = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (spinbutton));
	gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (list)), NULL,
	                                 &treeiter);
	gtk_tree_model_get (GTK_TREE_MODEL (macro_list), &treeiter, 0, &name, -1);
	path = g_build_filename (g_get_home_dir (), MACRO_DIR, name, NULL);
	g_free (name);
	gtk_widget_destroy (dialog);
	macrofile = g_mapped_file_new (path, FALSE, &error);
	g_free (path);

	if (!macrofile) {
		show_error ("Couldn't open macro", error->message);
		g_error_free (error);
		return;
	}

	gtk_text_buffer_begin_user_action (currfile->textbuffer);

	len = g_mapped_file_get_length (macrofile);
	start = g_mapped_file_get_contents (macrofile);

	for (; num_cycles > 0; num_cycles--) {
		curr_pos = start;

		while (curr_pos - start < len) {
			switch (*(curr_pos++)) {
				case MACRO_UP:
					gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, &iter,
					                                  gtk_text_buffer_get_insert (currfile->textbuffer));
					offset = gtk_text_iter_get_line_offset (&iter);
					gtk_text_iter_backward_line (&iter);
					if (offset >= gtk_text_iter_get_chars_in_line (&iter))
						gtk_text_iter_forward_to_line_end (&iter);
					else
						gtk_text_iter_set_line_offset (&iter, offset);
					gtk_text_buffer_place_cursor (currfile->textbuffer, &iter);
					break;
				case MACRO_DOWN:
					gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, &iter,
					                                  gtk_text_buffer_get_insert (currfile->textbuffer));
					offset = gtk_text_iter_get_line_offset (&iter);
					gtk_text_iter_forward_line (&iter);
					if (offset >= gtk_text_iter_get_chars_in_line (&iter) &&
					    !gtk_text_iter_ends_line (&iter))

						gtk_text_iter_forward_to_line_end (&iter);
					else
						gtk_text_iter_set_line_offset (&iter, offset);
					gtk_text_buffer_place_cursor (currfile->textbuffer, &iter);
					break;
				case MACRO_LEFT:
					gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, &iter,
					                                  gtk_text_buffer_get_insert (currfile->textbuffer));
					gtk_text_iter_backward_char (&iter);
					gtk_text_buffer_place_cursor (currfile->textbuffer, &iter);
					break;
				case MACRO_RIGHT:
					gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, &iter,
					                                  gtk_text_buffer_get_insert (currfile->textbuffer));
					gtk_text_iter_forward_char (&iter);
					gtk_text_buffer_place_cursor (currfile->textbuffer, &iter);
					break;
				case MACRO_HOME:
					gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, &iter,
					                                  gtk_text_buffer_get_insert (currfile->textbuffer));
					gtk_text_iter_set_line_offset (&iter, 0);
					gtk_text_buffer_place_cursor (currfile->textbuffer, &iter);
					break;
				case MACRO_END:
					gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, &iter,
					                                  gtk_text_buffer_get_insert (currfile->textbuffer));
					if (!gtk_text_iter_ends_line (&iter))
						gtk_text_iter_forward_to_line_end (&iter);
					gtk_text_buffer_place_cursor (currfile->textbuffer, &iter);
					break;
				case MACRO_DELETE:
					gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, &iter,
					                                  gtk_text_buffer_get_insert (currfile->textbuffer));
					iter_bound = iter;
					gtk_text_iter_forward_char (&iter_bound);
					gtk_text_buffer_delete (currfile->textbuffer, &iter, &iter_bound);
					break;
				case MACRO_BACKSPACE:
					gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, &iter,
					                                  gtk_text_buffer_get_insert (currfile->textbuffer));
					iter_bound = iter;
					gtk_text_iter_backward_char (&iter_bound);
					gtk_text_buffer_delete (currfile->textbuffer, &iter, &iter_bound);
					break;
				case MACRO_INSERT:
					insert_len = (gint) *(curr_pos++);
					gtk_text_buffer_insert_at_cursor (currfile->textbuffer, curr_pos, insert_len);
					curr_pos += insert_len;
					break;
			}
		}
	}

	gtk_text_buffer_end_user_action (currfile->textbuffer);
	g_mapped_file_free (macrofile);
}


/*
 * Callbacks for manage macro dialog
 */

void callback_delete_macro (GtkWidget *widget, gpointer list) {
	GtkWidget *dialog;
	gchar *name, *path;
	gint response;
	GtkTreeIter treeiter;
	GtkTreeModel *macro_list;

	if (!gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (list)), NULL,
	                                      &treeiter))
		return;

	macro_list = gtk_tree_view_get_model (GTK_TREE_VIEW (list));

	dialog = gtk_message_dialog_new (GTK_WINDOW (window), GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION,
	                                 GTK_BUTTONS_NONE, "Are you sure you want to delete this macro?");
	gtk_dialog_add_buttons (GTK_DIALOG (dialog), "_Delete", GTK_RESPONSE_YES, GTK_STOCK_CANCEL,
	                        GTK_RESPONSE_CANCEL, NULL);
	response = gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);

	if (response != GTK_RESPONSE_YES)
		return;

	gtk_tree_model_get (macro_list, &treeiter, 0, &name, -1);
	path = g_build_filename (g_get_home_dir (), MACRO_DIR, name, NULL);
	g_free (name);
	if (g_unlink (path))
		show_error ("Couldn't delete macro", "Your system's REALLY screwed, man");
	else
		gtk_list_store_remove (GTK_LIST_STORE (macro_list), &treeiter);

	g_free (path);
}


void callback_rename_macro (GtkWidget *widget, gpointer list) {
	GtkWidget *dialog, *hbox, *label, *entry;
	GtkTreeIter treeiter;
	GtkTreeModel *macro_list;
	gchar *oldpath, *newpath, *name;

	if (!gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (list)), NULL,
	                                      &treeiter))
		return;

	macro_list = gtk_tree_view_get_model (GTK_TREE_VIEW (list));

	dialog = gtk_dialog_new_with_buttons ("Renaming macro", GTK_WINDOW (window),
				GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
				GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

	hbox = gtk_hbox_new (FALSE, 0);
	label = gtk_label_new ("New name for macro:");
	entry = gtk_entry_new ();
	gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 6);
	gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 6);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, TRUE, TRUE, 6);
	gtk_widget_show_all (hbox);

	if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_ACCEPT) {
		gtk_widget_destroy (dialog);
		return;
	}

	newpath = g_build_filename (g_get_home_dir (), MACRO_DIR,
	                            gtk_entry_get_text (GTK_ENTRY (entry)), NULL);
	gtk_tree_model_get (macro_list, &treeiter, 0, &name, -1);
	oldpath = g_build_filename (g_get_home_dir (), MACRO_DIR, name, NULL);
	g_free (name);

	if (g_rename (oldpath, newpath))
		show_error ("Couldn't rename macro", "You have the worst luck");
	else
		gtk_list_store_set (GTK_LIST_STORE (macro_list), &treeiter, 0,
		                    gtk_entry_get_text (GTK_ENTRY (entry)), -1);

	g_free (oldpath);
	g_free (newpath);
	gtk_widget_destroy (dialog);
}


/*
 * Shows dialog for managing macros
 */

void callback_show_macro_manager (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog, *vbox, *hbox, *list, *label, *viewport, *button, *buttonbox;
	GtkListStore *macro_list;
	GtkCellRenderer *textrenderer;
	GtkTreeViewColumn *column;
	GtkTreePath *treepath;

	//get the list of macros first
	macro_list = get_macro_list ();
	if (!macro_list)
		return;

	//build the dialog
	dialog = gtk_dialog_new_with_buttons ("Manage macros", GTK_WINDOW (window), GTK_DIALOG_MODAL,
	                                      GTK_STOCK_CLOSE, GTK_RESPONSE_OK, NULL);
	gtk_window_set_resizable (GTK_WINDOW (dialog), TRUE);
	vbox = gtk_vbox_new (FALSE, 6);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 6);

	label = gtk_label_new ("Existing macros:");
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);

	hbox = gtk_hbox_new (FALSE, 6);
	list = gtk_tree_view_new_with_model (GTK_TREE_MODEL (macro_list));
	textrenderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("", textrenderer, "text", 0, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (list), column);
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (list), FALSE);
	gtk_tree_selection_set_mode (gtk_tree_view_get_selection (GTK_TREE_VIEW (list)),
	                             GTK_SELECTION_BROWSE);
	treepath = gtk_tree_path_new_first ();
	gtk_tree_selection_select_path (gtk_tree_view_get_selection (GTK_TREE_VIEW (list)),
	                                treepath);
	gtk_tree_path_free (treepath);
	viewport = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (viewport), GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (viewport), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (viewport), list);
	gtk_box_pack_start (GTK_BOX (hbox), viewport, TRUE, TRUE, 0);

	buttonbox = gtk_vbutton_box_new ();
	gtk_button_box_set_layout (GTK_BUTTON_BOX (buttonbox), GTK_BUTTONBOX_START);
	button = gtk_button_new_from_stock (GTK_STOCK_DELETE);
	gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
	g_signal_connect (button, "clicked", G_CALLBACK (callback_delete_macro), list);
	gtk_container_add (GTK_CONTAINER (buttonbox), button);
	button = gtk_button_new_with_mnemonic ("_Rename");
	gtk_button_set_image (GTK_BUTTON (button), gtk_image_new_from_stock (GTK_STOCK_EDIT,
	                      GTK_ICON_SIZE_BUTTON));
	gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
	g_signal_connect (button, "clicked", G_CALLBACK (callback_rename_macro), list);
	gtk_container_add (GTK_CONTAINER (buttonbox), button);
	gtk_box_pack_start (GTK_BOX (hbox), buttonbox, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);

	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), vbox);
	gtk_widget_show_all (vbox);

	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
}


/*
 * Adds action text to the end of
 * the current macro string
 */

void append_macro_string (const gchar *text) {
	gchar *new_curr_macro_string;

	if (!curr_macro_string) {
		curr_macro_string = g_strdup (text);
	} else {
		new_curr_macro_string = g_strconcat (curr_macro_string, "+", text, NULL);
		g_free (curr_macro_string);
		curr_macro_string = new_curr_macro_string;
	}

	set_statusbar_text (curr_macro_string);
}


/*
 * Adds a keypress to the macro,
 * if it's one of the useful ones
 */

void add_macro_keypress (guint key) {
	macro_step *new_step;
	macro_action new_action;
	gchar *new_text;

	switch (key) {
		//if it's one of the keys we're looking for, get the macro action code
		case GDK_Left:
		case GDK_KP_Left:
			new_action = MACRO_LEFT;
			new_text = "Left";
			break;
		case GDK_Right:
		case GDK_KP_Right:
			new_action = MACRO_RIGHT;
			new_text = "Right";
			break;
		case GDK_Up:
		case GDK_KP_Up:
			new_action = MACRO_UP;
			new_text = "Up";
			break;
		case GDK_Down:
		case GDK_KP_Down:
			new_action = MACRO_DOWN;
			new_text = "Down";
			break;
		case GDK_Home:
		case GDK_KP_Home:
			new_action = MACRO_HOME;
			new_text = "Home";
			break;
		case GDK_End:
		case GDK_KP_End:
			new_action = MACRO_END;
			new_text = "End";
			break;
		case GDK_Delete:
		case GDK_KP_Delete:
			new_action = MACRO_DELETE;
			new_text = "Delete";
			break;
		case GDK_BackSpace:
			new_action = MACRO_BACKSPACE;
			new_text = "Backspace";
			break;
		default:
			//if it's none of the keys we're looking for, ignore it
			return;
	}

	//now create the new action and add it to the list
	new_step = g_slice_new (macro_step);
	new_step->action = new_action;
	new_step->next = NULL;
	if (curr_macro_tail)
		curr_macro_tail->next = new_step;
	curr_macro_tail = new_step;
	if (!curr_macro)
		curr_macro = new_step;

	append_macro_string (new_text);
}


/*
 * Adds a "text inserted" action to
 * the macro
 */

void add_macro_insert (const gchar *text) {
	macro_step *new_step;
	gchar *new_text;

	new_step = g_slice_new (macro_step);
	new_step->action = MACRO_INSERT;
	new_step->text = g_strdup (text);
	new_step->next = NULL;
	if (curr_macro_tail)
		curr_macro_tail->next = new_step;
	curr_macro_tail = new_step;
	if (!curr_macro)
		curr_macro = new_step;

	if (strlen (text) > 10) {
		new_text = g_strdup (text);
		strcpy (new_text + 6, "...");
		append_macro_string (new_text);
		g_free (new_text);
	} else {
		append_macro_string (text);
	}
}

