/*
	Wabbitcode - An IDE for the Z80 line of TI calculators
	Copyright (C) 2007 Donald Straney

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <string.h>
#include "asmide.h"
#include "file.h"
#include "settings.h"
#include "callbacks.h"
#include "undo.h"
#include "search.h"
#include "build.h"
#include "project.h"
#include "macros.h"

//main widgets that need to be manipulated by routines
GtkWidget *window, *output, *outputwinvbox, *outputwinlabel, *statusbar, *notebook, *recentmenu,
	*killbutton, *showasmoutput, *outputwin;
//list of widgets that can only be active when a file's open
GList *file_widgets = NULL;
//list of widgets that can only be active when a project's open
GList *project_widgets = NULL;
//list of widgets that can't be active when the assembler's running
GList *assemble_widgets = NULL;
//list of widgets for build button/menu item
GList *build_widgets = NULL;
//lists of undo/redo widgets
GList *undo_widgets = NULL;
GList *redo_widgets = NULL;
//list of open files
GList *open_files = NULL;
//currently edited file
file *currfile = NULL;
//tag table for syntax highlighting
GtkTextTagTable *tagtable;
//initial window dimensions
gint winwidth, winheight;
//shared statusbar message context
guint statusbarctx;
//text tags for editing
GtkTextTag *result_tag, *label_tag, *comment_tag, *directive_tag,
	*opcode_tag, *string_tag, *register_tag, *red_tag, *green_tag,
	*orange_tag, *blue_tag;
//colors of tags
GdkColor resultcolor, labelcolor, commentcolor, dircolor,
	opcodecolor, strcolor, regcolor, redcolor, greencolor,
	orangecolor, bluecolor;
//flags for syntax highlighting
gboolean dosyntaxhlt, labelhlt, commenthlt, dirhlt, strhlt, opcodehlt, reghlt;
//recent file manager
GtkRecentManager *recentfiles;
//list of recent searches, recent search/replaces
GtkListStore *recent_search_list;
//initial number of recent files to show
gint numrecentfiles;
//initial tab width (in spaces)
gint tabwidth;
//initial word wrap setting
gboolean wordwrap;
//initial auto-indent setting
gboolean autoindent;
//initial editor font
PangoFontDescription *editorfont;
//tab array for files
PangoTabArray *tabstops;
//unicode characters for comparing in syntax highlighting
gunichar colon_unichar, leftparen_unichar, rightparen_unichar, dot_unichar, pound_unichar,
	semicolon_unichar, quote_unichar, comma_unichar, underscore_unichar;
//number of actions to keep in undo stack
gint max_undo_stack_len;
//are labels case-sensitive or not?
gboolean labels_casesensitive;
//output type
gint output_type;
//list of output types
GtkListStore *output_types;
//list of include dirs
GtkListStore *include_dirs;
//TRUE to automatically save before building
gboolean save_before_build;

//default height of assembler output window
#define DEFAULT_OUTPUTWIN_HEIGHT 250
//default width of project sidebar
#define DEFAULT_SIDEBAR_WIDTH 200
//number of tab stops to set
#define NUM_TAB_STOPS 20
//strings for assembler window label

#define ASSEMBLER_STOPPED_STR "Assembler is not running"
#define ASSEMBLER_RUNNING_STR "Assembler is working..."
//colors for assembler output window
#define RED_COLOR 0xFF0000
#define GREEN_COLOR 0x308C57
#define BLUE_COLOR 0x0000FF
#define ORANGE_COLOR 0xB7621B //FFA600


/*
 * Called when a recently-opened
 * file is selected - opens it
 * again without adding a duplicate
 * entry to the list
 */

void callback_open_recent_file (GtkRecentChooser *recentchooser, gpointer data) {
	GtkRecentInfo *recentfile;
	gchar *path;

	recentfile = gtk_recent_chooser_get_current_item (recentchooser);
	path = gtk_recent_info_get_uri_display (recentfile);

	open_file (path);

	g_free (path);
	gtk_recent_info_unref (recentfile);
}


/*
 * Builds main window + widgets
 */

void build_gui () {

	GtkWidget *vpaned, *hpaned, *toolbar, *menubar, *menu, *menuitem, *vbox,
		*hbox, *sidebar;
	GtkToolItem *toolbutton, *recordmacrobutton;
	GtkAccelGroup *accelgrp;
	GtkRecentFilter *recentfilter;

	//create the toplevel window
	gtk_window_set_default_icon_name ("wabbitcode");
	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	g_signal_connect (window, "destroy", gtk_main_quit, NULL);
	g_signal_connect (window, "delete_event", G_CALLBACK (callback_close_window), NULL);
	gtk_window_set_title (GTK_WINDOW (window), APP_NAME);
	gtk_container_set_border_width (GTK_CONTAINER (window), 0);

	//create the notebook for switching between files
	notebook = gtk_notebook_new ();
	gtk_notebook_set_scrollable (GTK_NOTEBOOK (notebook), TRUE);
	//TODO: try using a different signal - it seems like this one is activated even before
	// everything's drawn for some reason
	g_signal_connect (notebook, "switch-page", G_CALLBACK (callback_switch_file), NULL);

	//and set up the assembler output window
	output = gtk_text_view_new ();
	outputwin = gtk_scrolled_window_new (NULL, NULL);
	outputwinvbox = gtk_vbox_new (FALSE, 0);
	killbutton = gtk_button_new_with_label ("Stop assembler");
	gtk_button_set_image (GTK_BUTTON (killbutton),
		gtk_image_new_from_stock (GTK_STOCK_STOP, GTK_ICON_SIZE_MENU));
	gtk_button_set_relief (GTK_BUTTON (killbutton), GTK_RELIEF_NONE);
	gtk_widget_set_sensitive (killbutton, FALSE);
	g_signal_connect (killbutton, "clicked", G_CALLBACK (callback_stop_assembler), NULL);
	outputwinlabel = gtk_label_new (ASSEMBLER_STOPPED_STR);
	gtk_misc_set_alignment (GTK_MISC (outputwinlabel), 0, 0.5);
	hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (hbox), outputwinlabel, TRUE, TRUE, 2);
	gtk_box_pack_end (GTK_BOX (hbox), killbutton, FALSE, FALSE, 2);
//	gtk_box_pack_start (GTK_BOX (hbox), throbber, FALSE, FALSE, 2);
	gtk_container_add (GTK_CONTAINER (outputwin), output);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (outputwin), GTK_SHADOW_IN);
	gtk_text_view_set_editable (GTK_TEXT_VIEW (output), FALSE);
	gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (output), FALSE);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (outputwin), GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start (GTK_BOX (outputwinvbox), hbox, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (outputwinvbox), outputwin, TRUE, TRUE, 0);

	//a paned window divides the files and the output window
	vpaned = gtk_vpaned_new ();
	gtk_paned_pack1 (GTK_PANED (vpaned), notebook, TRUE, TRUE);
	gtk_paned_pack2 (GTK_PANED (vpaned), outputwinvbox, TRUE, TRUE);

	//create another paned window to separate the editor and project sidebar
	hpaned = gtk_hpaned_new ();
	sidebar = init_project_gui ();
	gtk_paned_pack1 (GTK_PANED (hpaned), sidebar, TRUE, TRUE);
	gtk_paned_pack2 (GTK_PANED (hpaned), vpaned, TRUE, TRUE);

	//...create the statusbar...
	statusbar = gtk_statusbar_new ();
	statusbarctx = gtk_statusbar_get_context_id (GTK_STATUSBAR (statusbar), "");

	//and the toolbar
	toolbar = gtk_toolbar_new ();
#ifdef G_OS_WIN32
	//make sure toolbar icon sizes come out right on Windows
	gtk_toolbar_set_icon_size (GTK_TOOLBAR (toolbar), GTK_ICON_SIZE_LARGE_TOOLBAR);
#endif
	toolbutton = gtk_tool_button_new_from_stock (GTK_STOCK_NEW);
	g_signal_connect (toolbutton, "clicked", G_CALLBACK (callback_new), NULL);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);
	toolbutton = gtk_tool_button_new_from_stock (GTK_STOCK_OPEN);
	g_signal_connect (toolbutton, "clicked", G_CALLBACK (callback_open), NULL);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);
	toolbutton = gtk_tool_button_new_from_stock (GTK_STOCK_SAVE);
	file_widgets = g_list_prepend (file_widgets, toolbutton);
	assemble_widgets = g_list_prepend (assemble_widgets, toolbutton);
	g_signal_connect (toolbutton, "clicked", G_CALLBACK (callback_save), NULL);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);
	toolbutton = gtk_separator_tool_item_new ();
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	toolbutton = gtk_tool_button_new_from_stock (GTK_STOCK_UNDO);
	file_widgets = g_list_prepend (file_widgets, toolbutton);
	undo_widgets = g_list_prepend (undo_widgets, toolbutton);
	g_signal_connect (toolbutton, "clicked", G_CALLBACK (callback_undo), NULL);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);
	toolbutton = gtk_tool_button_new_from_stock (GTK_STOCK_REDO);
	file_widgets = g_list_prepend (file_widgets, toolbutton);
	redo_widgets = g_list_prepend (redo_widgets, toolbutton);
	g_signal_connect (toolbutton, "clicked", G_CALLBACK (callback_redo), NULL);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);
	toolbutton = gtk_tool_button_new_from_stock (GTK_STOCK_COPY);
	file_widgets = g_list_prepend (file_widgets, toolbutton);
	g_signal_connect (toolbutton, "clicked", G_CALLBACK (callback_copy), NULL);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);
	toolbutton = gtk_tool_button_new_from_stock (GTK_STOCK_CUT);
	file_widgets = g_list_prepend (file_widgets, toolbutton);
	g_signal_connect (toolbutton, "clicked", G_CALLBACK (callback_cut), NULL);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);
	toolbutton = gtk_tool_button_new_from_stock (GTK_STOCK_PASTE);
	file_widgets = g_list_prepend (file_widgets, toolbutton);
	g_signal_connect (toolbutton, "clicked", G_CALLBACK (callback_paste), NULL);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);
	toolbutton = gtk_tool_button_new_from_stock (GTK_STOCK_FIND);
	g_signal_connect (toolbutton, "clicked", G_CALLBACK (callback_find), NULL);
	file_widgets = g_list_prepend (file_widgets, toolbutton);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);
	toolbutton = gtk_separator_tool_item_new ();
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	toolbutton = gtk_tool_button_new (gtk_image_new_from_stock (GTK_STOCK_EXECUTE,
		GTK_ICON_SIZE_SMALL_TOOLBAR), "Build");
	build_widgets = g_list_prepend (build_widgets, toolbutton);
	assemble_widgets = g_list_prepend (assemble_widgets, toolbutton);
	g_signal_connect (toolbutton, "clicked", G_CALLBACK (callback_build), NULL);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);
	toolbutton = gtk_tool_button_new (gtk_image_new_from_stock ("wabbitcode-code-counter",
		GTK_ICON_SIZE_SMALL_TOOLBAR), "Count");
	file_widgets = g_list_prepend (file_widgets, toolbutton);
	assemble_widgets = g_list_prepend (assemble_widgets, toolbutton);
	g_signal_connect (toolbutton, "clicked", G_CALLBACK (callback_countcode), NULL);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);
	toolbutton = gtk_separator_tool_item_new ();
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	toolbutton = gtk_tool_button_new (gtk_image_new_from_stock ("wabbitcode-run-macro",
		GTK_ICON_SIZE_SMALL_TOOLBAR), "Run");
	file_widgets = g_list_prepend (file_widgets, toolbutton);
	assemble_widgets = g_list_prepend (assemble_widgets, toolbutton);
	g_signal_connect (toolbutton, "clicked", G_CALLBACK (callback_run_macro), NULL);
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);
	toolbutton = gtk_toggle_tool_button_new ();
	gtk_tool_button_set_icon_widget (GTK_TOOL_BUTTON (toolbutton),
		gtk_image_new_from_stock ("wabbitcode-record-macro", GTK_ICON_SIZE_SMALL_TOOLBAR));
	gtk_tool_button_set_label (GTK_TOOL_BUTTON (toolbutton), "Record");
	file_widgets = g_list_prepend (file_widgets, toolbutton);
	assemble_widgets = g_list_prepend (assemble_widgets, toolbutton);
	g_signal_connect (toolbutton, "toggled", G_CALLBACK (callback_record_macro_toggled), NULL);
	recordmacrobutton = toolbutton;
	gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolbutton, -1);
	gtk_tool_item_set_homogeneous (GTK_TOOL_ITEM (toolbutton), FALSE);

	//set up for key accelerators
	accelgrp = gtk_accel_group_new ();
	gtk_window_add_accel_group (GTK_WINDOW (window), accelgrp);

	//get the recent files menu
	recentmenu = gtk_recent_chooser_menu_new_for_manager (recentfiles);
	gtk_recent_chooser_set_sort_type (GTK_RECENT_CHOOSER (recentmenu), GTK_RECENT_SORT_MRU);
	recentfilter = gtk_recent_filter_new ();
	gtk_recent_filter_add_application (recentfilter, APP_NAME);
	gtk_recent_chooser_set_filter (GTK_RECENT_CHOOSER (recentmenu), recentfilter);
	g_signal_connect (recentmenu, "item-activated", G_CALLBACK (callback_open_recent_file), NULL);

	//and create the menus
	menubar = gtk_menu_bar_new ();
	menu = gtk_menu_new ();
	gtk_menu_set_accel_group (GTK_MENU (menu), accelgrp);
	gtk_menu_set_accel_path (GTK_MENU (menu), "<Wabbitcode-Document>/File");
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_NEW, accelgrp);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_new), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_OPEN, accelgrp);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_open), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_label ("Recent files");
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), recentmenu);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_SAVE, accelgrp);
	file_widgets = g_list_prepend (file_widgets, menuitem);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_save), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_SAVE_AS, accelgrp);
	file_widgets = g_list_prepend (file_widgets, menuitem);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_saveas), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_REVERT_TO_SAVED, accelgrp);
	file_widgets = g_list_prepend (file_widgets, menuitem);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_revert), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_CLOSE, accelgrp);
	file_widgets = g_list_prepend (file_widgets, menuitem);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_close), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_PRINT, accelgrp);
	file_widgets = g_list_prepend (file_widgets, menuitem);
//	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_print), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_QUIT, accelgrp);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_quit), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("_File");
	assemble_widgets = g_list_prepend (assemble_widgets, menuitem);
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);
	gtk_menu_shell_append (GTK_MENU_SHELL (menubar), menuitem);

	menu = gtk_menu_new ();
	gtk_menu_set_accel_group (GTK_MENU (menu), accelgrp);
	gtk_menu_set_accel_path (GTK_MENU (menu), "<Wabbitcode-Document>/Edit");
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_UNDO, accelgrp);
	undo_widgets = g_list_prepend (undo_widgets, menuitem);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_undo), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_REDO, accelgrp);
	redo_widgets = g_list_prepend (redo_widgets, menuitem);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_redo), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_COPY, accelgrp);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_copy), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_CUT, accelgrp);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_cut), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_PASTE, accelgrp);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_paste), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_SELECT_ALL, accelgrp);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_select_all), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("Make lower_case");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_change_case), (gpointer)LOWER);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("Make _uppercase");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_change_case), (gpointer)UPPER);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("Make _title case");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_change_case), (gpointer)TITLE);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("_Invert case");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_change_case), (gpointer)INVERT);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_FIND, accelgrp);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_find), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_FIND_AND_REPLACE, accelgrp);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_replace), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_with_mnemonic ("_Go To Line");
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_JUMP_TO, GTK_ICON_SIZE_MENU));
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_goto_line), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("Go To _Label");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_goto_label), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_with_mnemonic ("_Stop Syntax Highlighting");
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_STOP, GTK_ICON_SIZE_MENU));
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_stop_syntaxhlt), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("_Edit");
	file_widgets = g_list_prepend (file_widgets, menuitem);
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);
	gtk_menu_shell_append (GTK_MENU_SHELL (menubar), menuitem);

	menu = gtk_menu_new ();
	gtk_menu_set_accel_group (GTK_MENU (menu), accelgrp);
	gtk_menu_set_accel_path (GTK_MENU (menu), "<Wabbitcode-Document>/Project");
	menuitem = gtk_image_menu_item_new_with_mnemonic ("_New project");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_new_project), NULL);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_NEW, GTK_ICON_SIZE_MENU));
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_with_mnemonic ("_Save project");
	project_widgets = g_list_prepend (project_widgets, menuitem);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_save_project), NULL);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_SAVE, GTK_ICON_SIZE_MENU));
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_with_mnemonic ("_Close project");
	project_widgets = g_list_prepend (project_widgets, menuitem);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_close_project), NULL);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU));
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_separator_menu_item_new ();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_with_mnemonic ("Project _options");
	project_widgets = g_list_prepend (project_widgets, menuitem);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_show_project_options), NULL);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_PROPERTIES, GTK_ICON_SIZE_MENU));
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("_Project");
	assemble_widgets = g_list_prepend (assemble_widgets, menuitem);
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);
	gtk_menu_shell_append (GTK_MENU_SHELL (menubar), menuitem);

	menu = gtk_menu_new ();
	gtk_menu_set_accel_group (GTK_MENU (menu), accelgrp);
	gtk_menu_set_accel_path (GTK_MENU (menu), "<Wabbitcode-Document>/Settings");
	menuitem = gtk_image_menu_item_new_with_mnemonic ("_General");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_seteditoropt), NULL);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_EDIT, GTK_ICON_SIZE_MENU));
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_with_mnemonic ("_Syntax Highlighting");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_setsyntaxhltopt), NULL);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_SELECT_COLOR, GTK_ICON_SIZE_MENU));
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_with_mnemonic ("_Assembler");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_setassembleropt), NULL);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_EXECUTE, GTK_ICON_SIZE_MENU));
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_check_menu_item_new_with_mnemonic ("Show Assembler _Output");
	showasmoutput = menuitem;
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_toggle_outputwin), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("_Settings");
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);
	gtk_menu_shell_append (GTK_MENU_SHELL (menubar), menuitem);

	menu = gtk_menu_new ();
	gtk_menu_set_accel_group (GTK_MENU (menu), accelgrp);
	gtk_menu_set_accel_path (GTK_MENU (menu), "<Wabbitcode-Document>/Macros");
	menuitem = gtk_image_menu_item_new_with_mnemonic ("_Run Macro");
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock ("wabbitcode-run-macro", GTK_ICON_SIZE_MENU));
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_run_macro), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_check_menu_item_new_with_mnemonic ("Re_cord Macro");
	g_signal_connect (menuitem, "toggled", G_CALLBACK (callback_record_macro_toggled_ind),
		(gpointer)recordmacrobutton);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_with_mnemonic ("_Manage Macros");
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_PROPERTIES, GTK_ICON_SIZE_MENU));
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_show_macro_manager), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("_Macros");
	build_widgets = g_list_prepend (build_widgets, menuitem);
	assemble_widgets = g_list_prepend (assemble_widgets, menuitem);
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);
	gtk_menu_shell_append (GTK_MENU_SHELL (menubar), menuitem);

	menu = gtk_menu_new ();
	assemble_widgets = g_list_prepend (assemble_widgets, menu);
	gtk_menu_set_accel_group (GTK_MENU (menu), accelgrp);
	gtk_menu_set_accel_path (GTK_MENU (menu), "<Wabbitcode-Document>/Assemble");
	menuitem = gtk_image_menu_item_new_with_mnemonic ("_Build Program");
	build_widgets = g_list_prepend (build_widgets, menuitem);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_EXECUTE, GTK_ICON_SIZE_MENU));
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_build), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("Create Symbol _Table");
	file_widgets = g_list_prepend (file_widgets, menuitem);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_createsymtable), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_with_mnemonic ("Run _Code Counter");
	file_widgets = g_list_prepend (file_widgets, menuitem);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock ("wabbitcode-code-counter", GTK_ICON_SIZE_MENU));
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_countcode), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_image_menu_item_new_with_mnemonic ("Show Program _Statistics");
	file_widgets = g_list_prepend (file_widgets, menuitem);
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_INFO, GTK_ICON_SIZE_MENU));
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_printstats), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("Generate _Listing");
	file_widgets = g_list_prepend (file_widgets, menuitem);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_createlisting), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("_Assemble");
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);
	gtk_menu_shell_append (GTK_MENU_SHELL (menubar), menuitem);

	menu = gtk_menu_new ();
	gtk_menu_set_accel_group (GTK_MENU (menu), accelgrp);
	gtk_menu_set_accel_path (GTK_MENU (menu), "<Wabbitcode-Document>/Help");
	menuitem = gtk_image_menu_item_new_with_mnemonic ("_Manual");
	gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem),
		gtk_image_new_from_stock (GTK_STOCK_HELP, GTK_ICON_SIZE_MENU));
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_show_help), "manual.html");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("_Opcode Reference");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_show_help), "opcodes.html");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("_Assembler readme");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_show_help), "spasm_readme");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("_Assembler syntax and macros");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_show_help), "spasm_manual.html");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic (".Option _reference");
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_show_help), "spasm_options.html");
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);

	menuitem = gtk_image_menu_item_new_from_stock (GTK_STOCK_ABOUT, accelgrp);
	g_signal_connect (menuitem, "activate", G_CALLBACK (callback_show_about), NULL);
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), menuitem);
	menuitem = gtk_menu_item_new_with_mnemonic ("_Help");
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menu);
	gtk_menu_shell_append (GTK_MENU_SHELL (menubar), menuitem);

	//add some extra key accelerators
	gtk_accel_map_add_entry ("<Wabbitcode-Document>/Edit/Select All", GDK_A, GDK_CONTROL_MASK);
	gtk_accel_map_add_entry ("<Wabbitcode-Document>/Edit/Go To Line", GDK_G, GDK_CONTROL_MASK);
	gtk_accel_map_add_entry ("<Wabbitcode-Document>/Edit/Go To Label", GDK_L, GDK_CONTROL_MASK);
	gtk_accel_map_add_entry ("<Wabbitcode-Document>/Edit/Undo", GDK_Z, GDK_CONTROL_MASK);
	gtk_accel_map_add_entry ("<Wabbitcode-Document>/Edit/Redo", GDK_Z,
		GDK_CONTROL_MASK | GDK_SHIFT_MASK);
	gtk_accel_map_add_entry ("<Wabbitcode-Document>/Edit/Stop Syntax Highlighting", GDK_Escape, 0);
	gtk_accel_map_add_entry ("<Wabbitcode-Document>/Assemble/Build Program", GDK_F5, 0);
	gtk_accel_map_add_entry ("<Wabbitcode-Document>/Assemble/Create Symbol Table", GDK_F6, 0);
	gtk_accel_map_add_entry ("<Wabbitcode-Document>/Assemble/Code Counter", GDK_F7, 0);
	gtk_accel_map_add_entry ("<Wabbitcode-Document>/Assemble/Program Statistics", GDK_F8, 0);
	gtk_accel_map_add_entry ("<Wabbitcode-Document>/Assemble/Generate Listing", GDK_F9, 0);
	gtk_accel_map_add_entry ("<Wabbitcode-Document>/Help/Using Assembly IDE", GDK_F1, 0);

	//shove it all in!
	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (window), vbox);
	gtk_box_pack_start (GTK_BOX (vbox), menubar, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), toolbar, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hpaned, TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), statusbar, FALSE, FALSE, 0);

	//set window size and default position of paned window
	gtk_window_set_default_size (GTK_WINDOW (window), winwidth, winheight);
	gtk_paned_set_position (GTK_PANED (vpaned), winheight - DEFAULT_OUTPUTWIN_HEIGHT);
	gtk_paned_set_position (GTK_PANED (hpaned), DEFAULT_SIDEBAR_WIDTH);

	//finally, show everything
	gtk_widget_show_all (vbox);
	gtk_widget_hide (outputwinvbox);
	gtk_widget_hide (sidebar);

	gtk_widget_show (window);
}


/*
 * Adds some custom icons for actions
 * that aren't in GTK's stock icons
 */

void add_custom_icons () {
	GtkIconFactory *iconfactory;
	GtkIconSet *iconset;
	GdkPixbuf *pixbuf;
	gint curr_icon;
	gchar *curr_icon_path;
#ifdef G_OS_WIN32
    //in Windows, load the standard GNOME/Mist icons to avoid those ugly GTK stock icons
	const gchar *icon_list[] = {"new.png", GTK_STOCK_NEW, "open.png", GTK_STOCK_OPEN, "save.png",
		GTK_STOCK_SAVE, "save_as.png", GTK_STOCK_SAVE_AS, "revert_to_saved.png",
		GTK_STOCK_REVERT_TO_SAVED, "close.png", GTK_STOCK_CLOSE, "print.png", GTK_STOCK_PRINT,
		"undo.png", GTK_STOCK_UNDO, "redo.png", GTK_STOCK_REDO, "copy.png", GTK_STOCK_COPY,
		"cut.png", GTK_STOCK_CUT, "paste.png", GTK_STOCK_PASTE, "find.png", GTK_STOCK_FIND,
		"quit.png", GTK_STOCK_QUIT, "find_and_replace.png", GTK_STOCK_FIND_AND_REPLACE,
		"properties.png", GTK_STOCK_PROPERTIES, "edit.png", GTK_STOCK_EDIT, "help.png",
		GTK_STOCK_HELP, "select_all.png", GTK_STOCK_SELECT_ALL, "codecounter.png",
		"wabbitcode-code-counter", "runmacro.png", "wabbitcode-run-macro", "recordmacro.png",
        "wabbitcode-record-macro", "refresh.png", GTK_STOCK_REFRESH, NULL};

    //there are also some theme icons we need to add
    gtk_icon_theme_append_search_path (gtk_icon_theme_get_default (), "./");
#else
	const gchar *icon_list[] = {"codecounter.png", "wabbitcode-code-counter", "runmacro.png",
		"wabbitcode-run-macro", "recordmacro.png", "wabbitcode-record-macro", NULL};
#endif

	iconfactory = gtk_icon_factory_new ();
	gtk_icon_factory_add_default (iconfactory);

	for (curr_icon = 0; icon_list[curr_icon]; curr_icon++) {
		curr_icon_path = g_build_filename (DATADIR, icon_list[curr_icon++], NULL);
		pixbuf = gdk_pixbuf_new_from_file (curr_icon_path, NULL);
		if (pixbuf) {
			iconset = gtk_icon_set_new_from_pixbuf (pixbuf);
			gtk_icon_factory_add (iconfactory, icon_list[curr_icon], iconset);
		}
		g_free (curr_icon_path);
	}
}


/*
 * Updates sensitivity of build button/menu option
 * depending on whether a file or project is open
 */

void update_build_widgets () {
	GList *currwidget;

	for (currwidget = build_widgets; currwidget; currwidget = currwidget->next)
		gtk_widget_set_sensitive (GTK_WIDGET (currwidget->data), project_open || open_files);
}


/*
 * Updates sensitivity of file-specific menu options
 * depending on whether a file is open or not
 */

void update_file_widgets (gboolean sensitive) {
	GList *currwidget;

	for (currwidget = file_widgets; currwidget; currwidget = currwidget->next)
		gtk_widget_set_sensitive (GTK_WIDGET (currwidget->data), sensitive);

	update_build_widgets ();
}


/*
 * Makes undo widgets sensitive or insensitive
 */

void set_undo_sensitive (gboolean sensitive) {
	GList *currwidget;

	for (currwidget = undo_widgets; currwidget; currwidget = currwidget->next)
		gtk_widget_set_sensitive (GTK_WIDGET (currwidget->data), sensitive);
}


/*
 * Makes redo widgets sensitive or insensitive
 */

void set_redo_sensitive (gboolean sensitive) {
	GList *currwidget;

	for (currwidget = redo_widgets; currwidget; currwidget = currwidget->next)
		gtk_widget_set_sensitive (GTK_WIDGET (currwidget->data), sensitive);
}


/*
 * Gets average width of chars
 * for current editor font
 */

int get_font_width () {
	const char *lang;
	PangoFontMetrics *fontmetrics;
	int fontwidth;

	if (g_getenv ("LANG"))
		lang = g_getenv ("LANG");
	else if (g_getenv ("LC_ALL"))
		lang = g_getenv ("LC_ALL");
	else if (g_getenv ("LC_CTYPE"))
		lang = g_getenv ("LC_CTYPE");
	else
		lang = "en_US";

	fontmetrics = pango_context_get_metrics (gtk_widget_get_pango_context (output),
//		editorfont, pango_language_get_default ());
		editorfont, pango_language_from_string (lang));

	fontwidth = PANGO_PIXELS (pango_font_metrics_get_approximate_digit_width (fontmetrics));
	pango_font_metrics_unref (fontmetrics);

	return fontwidth;
}


/*
 * Re-sets editor settings
 */

void update_editor_settings () {
	int fontwidth, currtab;
	GList *node;
	file *currfile;

	//set result colors
	g_object_set (result_tag, "background-gdk", &resultcolor, NULL);
	//set number of recent files shown
	gtk_recent_chooser_set_limit (GTK_RECENT_CHOOSER (recentmenu), numrecentfiles);
	//set tabs
	fontwidth = get_font_width ();
	for (currtab = 0; currtab < NUM_TAB_STOPS; currtab++)
		pango_tab_array_set_tab (tabstops, currtab, PANGO_TAB_LEFT,
			(currtab + 1) * tabwidth * fontwidth);

	//now for each file...
	for (node = open_files; node; node = g_list_next (node)) {
		currfile = (file *)(node->data);
		//set word wrap
		gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (currfile->textview),
			wordwrap ? GTK_WRAP_WORD_CHAR : GTK_WRAP_NONE);
		//set font
		gtk_widget_modify_font (currfile->textview, editorfont);
		//set tabs
		gtk_text_view_set_tabs (GTK_TEXT_VIEW (currfile->textview), tabstops);
	}
}


/*
 * Re-sets syntax highlighting settings
 */

void update_syntaxhlt_settings () {

	//set tag colors
	g_object_set (label_tag, "foreground-gdk", &labelcolor, NULL);
	g_object_set (comment_tag, "foreground-gdk", &commentcolor, NULL);
	g_object_set (directive_tag, "foreground-gdk", &dircolor, NULL);
	g_object_set (opcode_tag, "foreground-gdk", &opcodecolor, NULL);
	g_object_set (string_tag, "foreground-gdk", &strcolor, NULL);
	g_object_set (register_tag, "foreground-gdk", &regcolor, NULL);
}


int main (int argc, char **argv) {
	int currarg;
	GtkTextTagTable *output_tagtable;

	//init GTK (what else would this do? create a group of monkeys?)
	gtk_init (&argc, &argv);
	g_set_application_name (APP_NAME);
	recording_macro = FALSE;
	//set up tags for searches and syntax highlighting
	tagtable = gtk_text_tag_table_new ();
	label_tag = gtk_text_tag_new ("label");
	comment_tag = gtk_text_tag_new ("comment");
	directive_tag = gtk_text_tag_new ("directive");
	opcode_tag = gtk_text_tag_new ("opcode");
	register_tag = gtk_text_tag_new ("register");
	string_tag = gtk_text_tag_new ("string");
	result_tag = gtk_text_tag_new ("result");
	gtk_text_tag_table_add (tagtable, label_tag);
	gtk_text_tag_table_add (tagtable, comment_tag);
	gtk_text_tag_table_add (tagtable, directive_tag);
	gtk_text_tag_table_add (tagtable, opcode_tag);
	gtk_text_tag_table_add (tagtable, register_tag);
	gtk_text_tag_table_add (tagtable, string_tag);
	gtk_text_tag_table_add (tagtable, result_tag);
	colon_unichar = g_utf8_get_char (":");
	leftparen_unichar = g_utf8_get_char ("(");
	rightparen_unichar = g_utf8_get_char (")");
	dot_unichar = g_utf8_get_char (".");
	pound_unichar = g_utf8_get_char ("#");
	semicolon_unichar = g_utf8_get_char (";");
	quote_unichar = g_utf8_get_char ("\"");
	comma_unichar = g_utf8_get_char (",");
	underscore_unichar = g_utf8_get_char ("_");
	//build list of output types
	output_types = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
	gtk_list_store_insert_with_values (output_types, NULL, 0, 0, "TI-86 program", 1, "86p", -1);
	gtk_list_store_insert_with_values (output_types, NULL, 0, 0, "TI-86 string", 1, "86s", -1);
	gtk_list_store_insert_with_values (output_types, NULL, 0, 0, "TI-85 program", 1, "85p", -1);
	gtk_list_store_insert_with_values (output_types, NULL, 0, 0, "TI-85 string", 1, "85s", -1);
	gtk_list_store_insert_with_values (output_types, NULL, 0, 0, "TI-83+/84 program", 1, "8xp", -1);
	gtk_list_store_insert_with_values (output_types, NULL, 0, 0, "TI-83+/84 application", 1, "8xk", -1);
	gtk_list_store_insert_with_values (output_types, NULL, 0, 0, "TI-83 program", 1, "83p", -1);
	gtk_list_store_insert_with_values (output_types, NULL, 0, 0, "TI-82 program", 1, "82p", -1);
	gtk_list_store_insert_with_values (output_types, NULL, 0, 0, "TI-73 program", 1, "73p", -1);
	gtk_list_store_insert_with_values (output_types, NULL, 0, 0, "Binary file", 1, "bin", -1);
	//load program settings
	load_settings ();
	//add the custom icons
	add_custom_icons ();
	//show the main window
	recentfiles = gtk_recent_manager_get_default ();
	recent_search_list = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
	build_gui ();
	update_file_widgets (FALSE);
	update_project_view (FALSE);
	//set up tags for assembler output window
	INIT_COLOR (redcolor, RED_COLOR);
	INIT_COLOR (greencolor, GREEN_COLOR);
	INIT_COLOR (bluecolor, BLUE_COLOR);
	INIT_COLOR (orangecolor, ORANGE_COLOR);
	output_tagtable = gtk_text_buffer_get_tag_table (gtk_text_view_get_buffer (GTK_TEXT_VIEW (output)));
	red_tag = gtk_text_tag_new ("red");
	green_tag = gtk_text_tag_new ("green");
	blue_tag = gtk_text_tag_new ("blue");
	orange_tag = gtk_text_tag_new ("orange");
	g_object_set (red_tag, "foreground-gdk", &redcolor, NULL);
	g_object_set (green_tag, "foreground-gdk", &greencolor, NULL);
	g_object_set (blue_tag, "foreground-gdk", &bluecolor, NULL);
	g_object_set (orange_tag, "foreground-gdk", &orangecolor, NULL);
	gtk_text_tag_table_add (output_tagtable, red_tag);
	gtk_text_tag_table_add (output_tagtable, green_tag);
	gtk_text_tag_table_add (output_tagtable, blue_tag);
	gtk_text_tag_table_add (output_tagtable, orange_tag);
	//apply current settings
	tabstops = pango_tab_array_new (NUM_TAB_STOPS, TRUE);
	update_editor_settings ();
	update_syntaxhlt_settings ();

	//finally, check if there are any filenames passed in as args that we need to open
	if (argc > 1) {
		for (currarg = 1; currarg < argc; currarg++)
			open_file (argv[currarg]);
	}

	gtk_main ();
	return 0;
}


/*
 * Shows error dialog with message
 */

void show_error (gchar *primary_text, gchar *secondary_text) {
	GtkWidget *dialog;

	dialog = gtk_message_dialog_new (GTK_WINDOW (window),
		GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
		primary_text);
	if (secondary_text)
		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog), secondary_text);
	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
}


/*
 * Shows info dialog with message
 */

void show_info (gchar *text) {
	GtkWidget *dialog;

	dialog = gtk_message_dialog_new (GTK_WINDOW (window),
		GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_INFO, GTK_BUTTONS_OK, text);
	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
}


/*
 * Sets statusbar message
 */

void set_statusbar_text (gchar *text) {

	gtk_statusbar_pop (GTK_STATUSBAR (statusbar), statusbarctx);
	gtk_statusbar_push (GTK_STATUSBAR (statusbar), statusbarctx, text);
}


/*
 * Updates statusbar text to show
 * current line and column numbers,
 * optionally with a GtkTextIter
 * for the cursor position
 */

void update_statusbar (GtkTextIter *iter) {
	gchar *messagetext;
	GtkTextMark *insertmark;
	GtkTextIter insertiter;

	if (recording_macro)
		return;

	if (!iter) {
		insertmark = gtk_text_buffer_get_insert (currfile->textbuffer);
		iter = &insertiter;
		gtk_text_buffer_get_iter_at_mark (currfile->textbuffer, iter, insertmark);
	}

	messagetext = g_strdup_printf ("Line %d, Col %d", gtk_text_iter_get_line (iter) + 1,
					gtk_text_iter_get_line_offset (iter) + 1);
	set_statusbar_text (messagetext);
	g_free (messagetext);
}


/*
 * Clears assembler output window
 */

void clear_assembler_outputwin () {
	GtkTextIter startiter, enditer;
	GtkTextBuffer *textbuffer;

	textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (output));
	gtk_text_buffer_get_bounds (textbuffer, &startiter, &enditer);
	gtk_text_buffer_delete (textbuffer, &startiter, &enditer);
}


/*
 * Updates assembler output window
 * for assembler state
 */

void update_assembler_outputwin (gboolean assembler_running) {
	GList *currwidget;

	if (assembler_running) {
		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (showasmoutput), TRUE);
		gtk_widget_set_sensitive (killbutton, TRUE);
		gtk_label_set_text (GTK_LABEL (outputwinlabel), ASSEMBLER_RUNNING_STR);

		//if a project is open, don't clear the output first
		//we want to be able to see all the messages
		if (!project_open)
			clear_assembler_outputwin ();

	} else {
		gtk_widget_set_sensitive (killbutton, FALSE);
		gtk_label_set_text (GTK_LABEL (outputwinlabel), ASSEMBLER_STOPPED_STR);
	}

	for (currwidget = assemble_widgets; currwidget; currwidget = currwidget->next)
		gtk_widget_set_sensitive (GTK_WIDGET (currwidget->data), !assembler_running);

	if (!assembler_running && !open_files)
		update_file_widgets (FALSE);
}


/*
 * Adds text to assembler output window,
 * in color if specified
 */

void append_assembler_output (const gchar *text, output_color color) {
	GtkTextBuffer *textbuffer;
	GtkTextIter enditer;
	GtkAdjustment *vscrollbar;

	textbuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (output));
	gtk_text_buffer_get_end_iter (textbuffer, &enditer);

	if (color == RED)
		gtk_text_buffer_insert_with_tags (textbuffer, &enditer, text, -1, red_tag, NULL);
	else if (color == GREEN)
		gtk_text_buffer_insert_with_tags (textbuffer, &enditer, text, -1, green_tag, NULL);
	else if (color == ORANGE)
		gtk_text_buffer_insert_with_tags (textbuffer, &enditer, text, -1, orange_tag, NULL);
	else if (color == BLUE)
		gtk_text_buffer_insert_with_tags (textbuffer, &enditer, text, -1, blue_tag, NULL);
	else
		gtk_text_buffer_insert (textbuffer, &enditer, text, -1);

	//because gtk_text_view_place_cursor_onscreen doesn't fuckin' work, move the scrollbar manually
	// ...and not even this works quite right - FUCK GTK
	vscrollbar = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (outputwin));
	if (vscrollbar)
		gtk_adjustment_set_value (vscrollbar, vscrollbar->upper - vscrollbar->page_size);
}


/*
 * Gets output type extension as text
 */

gchar *get_output_type_extension (int output_type) {
	gchar *extension;
	GtkTreePath *treepath;
	GtkTreeIter iter;

	treepath = gtk_tree_path_new_from_indices (output_type, -1);
	gtk_tree_model_get_iter (GTK_TREE_MODEL (output_types), &iter, treepath);
	gtk_tree_path_free (treepath);
	gtk_tree_model_get (GTK_TREE_MODEL (output_types), &iter, 1, &extension, -1);

	return extension;
}

