/*
	Wabbitcode - An IDE for the Z80 line of TI calculators
	Copyright (C) 2007 Donald Straney

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef G_OS_WIN32
#include <windows.h>
#endif

#include <gtk/gtk.h>
#include <string.h>
#include <ctype.h>
#include "asmide.h"
#include "project.h"
#include "file.h"

//lists of known filename extensions for choosing icons in sidebar
static const gchar *source_file_extensions[] = {".asm", ".z80", ".inc", ".h", NULL};
static const gchar *text_file_extensions[] = {".txt", ".text", ".c", ".cpp", ".hpp", NULL};
static const gchar *image_file_extensions[] = {".png", ".jpg", ".jpeg", ".bmp", ".xcf", ".ico",
                                               ".gif", ".tga", NULL};
static const gchar *html_file_extensions[] = {".htm", ".html", ".php", ".xhtml", ".shtm", ".shtml", NULL};
static const gchar *archive_file_extensions[] = {".zip", ".gz", ".bz2", ".tar", ".tgz", NULL};
//text for each build action
static const gchar *action_text[] = {"Assemble %s as %s", "Create symbol table for %s", "Run '%s'"};

//important sidebar widgets
GtkWidget *sidebarvbox, *sidebarviewlist;
//tree of files in project
GtkTreeStore *project_tree;
//true if project is open
gboolean project_open;
//path of project
gchar *project_path;
//path and filename of project
gchar *full_project_path;
//build sequence for project
build_step *build_sequence;
build_step *build_sequence_tail;
//true if project options are unsaved
gboolean project_changed;


/*
 * Sort function for
 * file list in
 * project sidebar
 */

gint project_tree_sort_func (GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer data) {
	gint retval;
	gchar *icon_a, *icon_b, *name_a, *name_b;
	gboolean folder_a, folder_b;

	gtk_tree_model_get (model, a, 0, &icon_a, 1, &name_a, -1);
	gtk_tree_model_get (model, b, 0, &icon_b, 1, &name_b, -1);

	folder_a = !strcmp (icon_a, "folder");
	folder_b = !strcmp (icon_b, "folder");

	//folders come first
	if (folder_a && !folder_b)
		retval = -1;
	else if (folder_b && !folder_a)
		retval = 1;
	else
		//if they're both the same type, go by names
		retval = strcasecmp (name_a, name_b);

	g_free (icon_a); g_free (icon_b);
	g_free (name_a); g_free (name_b);
	return retval;
}


/*
 * Tries to match filename's
 * extension to list of
 * extensions, returns TRUE
 * if successful
 */

gboolean match_filename_extension (const gchar *filename, const gchar **extension_list) {
	int len, extension_len, curr_extension;
	const gchar *extension;

	len = strlen (filename);

	curr_extension = 0;
	for (curr_extension = 0; extension_list[curr_extension]; curr_extension++) {
		extension_len = strlen (extension_list[curr_extension]);
		if (len < extension_len + 1)
			continue;

		extension = filename + len - extension_len;

		if (!strcmp (extension, extension_list[curr_extension]))
			return TRUE;
	}

	return FALSE;
}


/*
 * Refreshes project file list
 * in sidebar
 */

void refresh_project_files (GtkTreeIter *iter, gchar *path) {
	GDir *curr_dir;
	const gchar *curr_filename;
	gchar *full_filename, *icon;
	GtkTreeIter temp_iter;

	curr_dir = g_dir_open (path, 0, NULL);
	//TODO: show different icon for 'unreadable dir'?
	if (!curr_dir)
		return;

	//look at each file in the directory
	while ((curr_filename = g_dir_read_name (curr_dir))) {
		//skip it if it's hidden
		if (curr_filename[0] == '.')
			continue;

		full_filename = g_build_filename (path, curr_filename, NULL);

		if (g_file_test (full_filename, G_FILE_TEST_IS_DIR)) {
			//if it's a directory, go recursive
			gtk_tree_store_insert_with_values (project_tree, &temp_iter,
				iter, 0, 0, "folder", 1, curr_filename, -1);
			refresh_project_files (&temp_iter, full_filename);
		} else {
			//otherwise, just put in the name
			//...if it's the type of file we're looking for
			if (gtk_combo_box_get_active (GTK_COMBO_BOX (sidebarviewlist)) == 0 &&
			    !match_filename_extension (curr_filename, source_file_extensions)) {

				g_free (full_filename);
				continue;
			}

			//find an icon for it by its extension
			//TODO: include program icons, check if executable and have icon for that too
			if (match_filename_extension (curr_filename, text_file_extensions))
				icon = "text-x-generic";
			else if (match_filename_extension (curr_filename, image_file_extensions))
				icon = "image-x-generic";
			else if (match_filename_extension (curr_filename, html_file_extensions))
				icon = "text-html";
			else if (match_filename_extension (curr_filename, archive_file_extensions))
				icon = "package";
			else
				icon = "unknown";

			gtk_tree_store_insert_with_values (project_tree, &temp_iter,
				iter, 0, 0, icon, 1, curr_filename, -1);
		}
		g_free (full_filename);
	}

	g_dir_close (curr_dir);
}


/*
 * Called when the refresh button
 * on the project sidebar is pressed
 * - re-generates tree of project
 * files in sidebar
 */

void callback_refresh_project_files (GtkWidget *widget, gpointer data) {
	gtk_tree_store_clear (project_tree);
	refresh_project_files (NULL, project_path);
}


/*
 * Called when a file from the
 * sidebar is opened
 */

void callback_row_activated (GtkTreeView *treeview, GtkTreePath *treepath,
                             GtkTreeViewColumn *treecol, gpointer data) {

	GtkTreeModel *treemodel;
	gint *path_indices, curr_index, depth;
	gchar *curr_path, *temp, *name, *file_contents;
#ifndef G_OS_WIN32
	gchar *argv[] = {"xdg-open", NULL, NULL};
#endif
	GtkTreeIter curriter, olditer;
	guint file_len;

	//get the filename
	treemodel = gtk_tree_view_get_model (treeview);
	curr_path = g_strdup (project_path);
	path_indices = gtk_tree_path_get_indices (treepath);
	depth = gtk_tree_path_get_depth (treepath);
	for (curr_index = 0; curr_index < depth; curr_index++) {
		//for each component in the path, get its name
		if (curr_index == 0)
			gtk_tree_model_iter_nth_child (treemodel , &curriter, NULL, path_indices[curr_index]);
		else
			gtk_tree_model_iter_nth_child (treemodel , &curriter, &olditer, path_indices[curr_index]);
		gtk_tree_model_get (treemodel, &curriter, 1, &name, -1);
		olditer = curriter;
		//and stick that on the end of the current path
		temp = g_build_filename (curr_path, name, NULL);
		g_free (name);
		g_free (curr_path);
		curr_path = temp;
	}

	//now that we've got the file's full path, decide what to do with it
	if (match_filename_extension (curr_path, source_file_extensions)) {
		//if it's a source file, open it
		//TODO: what to do if it's already open?
		if (read_file (curr_path, &file_contents, &file_len)) {
			add_file (file_contents, file_len, curr_path);
			g_free (file_contents);
		}
	} else {
		//otherwise, pass it on to the OS to take care of
#ifdef G_OS_WIN32
		ShellExecute (NULL, NULL, curr_path, NULL, NULL, 0);
#else
		argv[1] = curr_path;
		if (!g_spawn_async (NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, NULL)) {
			show_error ("Couldn't open this file", "You need to install xdg-utils");
		}
#endif
	}

	g_free (curr_path);
}


/*
 * Initializes project GUI stuff
 * and returns sidebar widget
 */

GtkWidget *init_project_gui () {
	GtkWidget *sidebarlist, *hbox, *sidebarwin, *label, *button;
	GtkCellRenderer *textrenderer, *imagerenderer;
	GtkTreeViewColumn *column;

	//create tree store for project files
	project_tree = gtk_tree_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
	gtk_tree_sortable_set_default_sort_func (GTK_TREE_SORTABLE (project_tree),
		project_tree_sort_func, NULL, NULL);
	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (project_tree),
		GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID, GTK_SORT_ASCENDING);
	//build gui
	sidebarvbox = gtk_vbox_new (FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (sidebarvbox), 2);
	sidebarwin = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sidebarwin),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sidebarwin), GTK_SHADOW_IN);
	hbox = gtk_hbox_new (FALSE, 0);
	label = gtk_label_new ("Project files");
	button = gtk_button_new ();
	gtk_button_set_image (GTK_BUTTON (button),
		gtk_image_new_from_stock (GTK_STOCK_REFRESH, GTK_ICON_SIZE_MENU));
	gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
	g_signal_connect (button, "clicked", G_CALLBACK (callback_refresh_project_files), NULL);
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
	gtk_box_pack_end (GTK_BOX (hbox), button, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (sidebarvbox), hbox, FALSE, FALSE, 0);
	sidebarlist = gtk_tree_view_new_with_model (GTK_TREE_MODEL (project_tree));
	textrenderer = gtk_cell_renderer_text_new ();
	imagerenderer = gtk_cell_renderer_pixbuf_new ();
	column = gtk_tree_view_column_new_with_attributes ("", imagerenderer, "icon-name", 0, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (sidebarlist), column);
	column = gtk_tree_view_column_new_with_attributes ("", textrenderer, "text", 1, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (sidebarlist), column);
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (sidebarlist), FALSE);
	g_signal_connect (sidebarlist, "row-activated", G_CALLBACK (callback_row_activated), NULL);
	gtk_container_add (GTK_CONTAINER (sidebarwin), sidebarlist);
	gtk_box_pack_start (GTK_BOX (sidebarvbox), sidebarwin, TRUE, TRUE, 2);
	sidebarviewlist = gtk_combo_box_new_text ();
	gtk_combo_box_append_text (GTK_COMBO_BOX (sidebarviewlist), "Show source files only");
	gtk_combo_box_append_text (GTK_COMBO_BOX (sidebarviewlist), "Show all files");
	gtk_combo_box_set_active (GTK_COMBO_BOX (sidebarviewlist), 0);
	g_signal_connect (sidebarviewlist, "changed", G_CALLBACK (callback_refresh_project_files), NULL);
	gtk_box_pack_start (GTK_BOX (sidebarvbox), sidebarviewlist, FALSE, FALSE, 2);

	return sidebarvbox;
}


/*
 * Updates project sidebar depending
 * on whether project is open or not
 */

void update_project_view (gboolean show) {
	GList *currwidget;

	project_open = show;
	if (show) {
		gtk_widget_show (sidebarvbox);
		gtk_tree_store_clear (project_tree);
		refresh_project_files (NULL, project_path);
	} else {
		gtk_widget_hide (sidebarvbox);
	}

	for (currwidget = project_widgets; currwidget; currwidget = currwidget->next)
		gtk_widget_set_sensitive (GTK_WIDGET (currwidget->data), show);

	update_build_widgets ();
}


/*
 * Saves the project options,
 * returns TRUE if successful
 */

gboolean save_project_options () {
	FILE *project_file;
	build_step *curr_step;

	//replace the project file
	project_file = fopen (full_project_path, "w");
	if (!project_file) {
		show_error ("Couldn't save the project file", "Make sure you have permission to write to it");
		return FALSE;
	}

	//write each build step
	for (curr_step = build_sequence; curr_step; curr_step = curr_step->next) {
		switch (curr_step->action) {
			case BUILD_NORMAL:
				fprintf (project_file, "B %X %s\n", curr_step->output_type, curr_step->filename);
				break;
			case BUILD_SYMTABLE:
				fprintf (project_file, "S %s\n", curr_step->filename);
				break;
			case BUILD_COMMAND:
				fprintf (project_file, "E %s\n", curr_step->command);
		}
	}

	//that's it!
	fclose (project_file);
	return TRUE;
}


/*
 * Closes the current project,
 * returns TRUE if successful
 */

gboolean close_project () {
	build_step *next_step;
	GtkWidget *dialog;
	gint response;

	//first try to close all open files
	if (!remove_all_files ())
		return FALSE;

	//then prompt to save the project settings
	if (project_changed) {
		dialog = gtk_message_dialog_new (GTK_WINDOW (window), GTK_DIALOG_MODAL,
		                                 GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
		                                 "There are unsaved changes to the project options");
		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
		                                  "Would you like to save them before closing?");
		gtk_dialog_add_buttons (GTK_DIALOG (dialog), "Close _without saving", GTK_RESPONSE_NO,
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_YES, NULL);
		response = gtk_dialog_run (GTK_DIALOG (dialog));
		gtk_widget_destroy (dialog);

		if (response == GTK_RESPONSE_CANCEL) {
			return FALSE;
		} else if (response == GTK_RESPONSE_YES) {
			if (!save_project_options ())
				return FALSE;
		}
	}

	//finally, free any allocated project vars
	g_free (project_path);
	g_free (full_project_path);
	while (build_sequence) {
		if (build_sequence->filename) g_free (build_sequence->filename);
		if (build_sequence->command) g_free (build_sequence->command);
		next_step = build_sequence->next;
		g_slice_free (build_step, build_sequence);
		build_sequence = next_step;
	}
	return TRUE;
}


/*
 * Returns TRUE if file
 * is a project file
 */

gboolean is_project_file (gchar *path) {
	int pathlen, extensionlen;

	pathlen = strlen (path);
	//TODO: is there a way to do this with a macro?
	extensionlen = strlen (PROJECT_EXTENSION);

	//go by filename extension
	if (pathlen <= extensionlen)
		return FALSE;

	return !strcmp (PROJECT_EXTENSION, path + pathlen - extensionlen);
}


/*
 * Callback for "close project" menu item,
 * prompts to save project and hides sidebar
 */

void callback_close_project (GtkWidget *widget, gpointer data) {
	if (close_project ())
		update_project_view (FALSE);
}


/*
 * Callback for "save project" menu item,
 * saves all unsaved files and project options
 */

void callback_save_project (GtkWidget *widget, gpointer data) {
	GList *curr_node;
	file *curr_file;

	//for each open file,
	for (curr_node = open_files; curr_node; curr_node = g_list_next (curr_node)) {
		curr_file = (file *)(curr_node->data);
		//if it has unsaved changes, save it
		if (gtk_text_buffer_get_modified (curr_file->textbuffer))
			save_file (curr_file, FALSE);
	}

	//now save the project options
	if (save_project_options ())
		project_changed = FALSE;
}


/*
 * Callback for "new project" menu item,
 * prompts for location and name of new
 * project, creates file
 */

void callback_new_project (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog, *hbox, *label, *entry;
	const gchar *project_name;

	//if a project's already open, close it first
	if (project_open && !close_project ())
		return;

	//show the dialog to get a folder to use as the project root
	dialog = gtk_file_chooser_dialog_new ("Choose folder for project", GTK_WINDOW (window),
				GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_CANCEL) {
		gtk_widget_destroy (dialog);
		return;
	}

	project_path = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
	gtk_widget_destroy (dialog);

	//next prompt for the project's name
	dialog = gtk_dialog_new_with_buttons ("Enter the project's name", GTK_WINDOW (window),
				GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
				GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

	hbox = gtk_hbox_new (FALSE, 0);
	label = gtk_label_new ("Project name:");
	entry = gtk_entry_new ();
	gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);
	gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 6);
	gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 6);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, TRUE, TRUE, 6);
	gtk_widget_show_all (hbox);

	if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_ACCEPT) {
		gtk_widget_destroy (dialog);
		return;
	}
	project_name = gtk_entry_get_text (GTK_ENTRY (entry));

	//and create the project file
	full_project_path =
		g_strconcat (project_path, G_DIR_SEPARATOR_S, project_name, PROJECT_EXTENSION, NULL);
	build_sequence = NULL;
	save_project_options ();

	//finally, show the sidebar
	gtk_widget_destroy (dialog);
	update_project_view (TRUE);
	project_changed = FALSE;
}


/*
 * Returns a human-readable string
 * describing a build step
 */

gchar *get_action_text (build_action action, gint output_type, gchar *filename, gchar *command) {
	gchar *extension, *text;

	extension = get_output_type_extension (output_type);
	text = g_strdup_printf (action_text[(int) action], action == BUILD_COMMAND ? command : filename, extension);
	if (extension) g_free (extension);
	return text;
}


/*
 * Callbacks for editing build sequence
 */

typedef struct {
	GtkWidget *label;
	GtkWidget *typelist;
	GtkWidget *typelistlabel;
	GtkWidget *filechooser;
	GtkWidget *command;
} edit_build_step_widgets;


void callback_action_changed (GtkComboBox *actionlist, gpointer all_widgets) {
	edit_build_step_widgets *widgets = (edit_build_step_widgets *)all_widgets;
	int action = gtk_combo_box_get_active (actionlist);

	if (action == BUILD_NORMAL) {
		gtk_widget_show (GTK_WIDGET (widgets->typelist));
		gtk_widget_show (GTK_WIDGET (widgets->typelistlabel));
	} else {
		gtk_widget_hide (GTK_WIDGET (widgets->typelist));
		gtk_widget_hide (GTK_WIDGET (widgets->typelistlabel));
	}

	if (action == BUILD_COMMAND) {
		gtk_widget_hide (GTK_WIDGET (widgets->filechooser));
		gtk_widget_show (GTK_WIDGET (widgets->command));
	} else {
		gtk_widget_show (GTK_WIDGET (widgets->filechooser));
		gtk_widget_hide (GTK_WIDGET (widgets->command));
	}

	gtk_label_set_text (GTK_LABEL (widgets->label), action == BUILD_COMMAND ? "Command:" : "File:");
}


gboolean edit_build_step (GtkListStore *build_step_list, GtkTreeIter *treeiter) {
	GtkWidget *dialog, *actionlist, *typelist, *filechooser, *label, *table, *command;
	GtkCellRenderer *textrenderer;
	gint curr_action, curr_type, project_path_len;
	gchar *curr_filename, *text, *fullpath, *temp_filename, *curr_command;
	edit_build_step_widgets widgets;

	dialog = gtk_dialog_new_with_buttons ("Edit build step", GTK_WINDOW (window), GTK_DIALOG_MODAL,
	                                      GTK_STOCK_OK, GTK_RESPONSE_OK, GTK_STOCK_CANCEL,
	                                      GTK_RESPONSE_CANCEL, NULL);
	gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);

	gtk_tree_model_get (GTK_TREE_MODEL (build_step_list), treeiter, 1, &curr_action, 2, &curr_type,
	                    3, &curr_filename, 4, &curr_command, -1);

	table = gtk_table_new (2, 3, FALSE);
	gtk_table_set_col_spacings (GTK_TABLE (table), 6);
	gtk_table_set_row_spacings (GTK_TABLE (table), 6);
	gtk_container_set_border_width (GTK_CONTAINER (table), 6);

	label = gtk_label_new ("Action:");
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 0, 1);
	actionlist = gtk_combo_box_new_text ();
	gtk_combo_box_append_text (GTK_COMBO_BOX (actionlist), "Assemble file");
	gtk_combo_box_append_text (GTK_COMBO_BOX (actionlist), "Create symbol table from file");
	gtk_combo_box_append_text (GTK_COMBO_BOX (actionlist), "Run command");
	gtk_combo_box_set_active (GTK_COMBO_BOX (actionlist), curr_action);
	gtk_table_attach_defaults (GTK_TABLE (table), actionlist, 1, 2, 0, 1);

	label = gtk_label_new ("Output type:");
	widgets.typelistlabel = label;
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 1, 2);
	typelist = gtk_combo_box_new_with_model (GTK_TREE_MODEL (output_types));
	textrenderer = gtk_cell_renderer_text_new ();
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (typelist), textrenderer, TRUE);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (typelist), textrenderer, "text", 0);
	gtk_combo_box_set_active (GTK_COMBO_BOX (typelist), curr_type);
	gtk_table_attach_defaults (GTK_TABLE (table), typelist, 1, 2, 1, 2);
	widgets.typelist = typelist;

	g_signal_connect (actionlist, "changed", G_CALLBACK (callback_action_changed), (gpointer)&widgets);

	label = gtk_label_new ("File:");
	widgets.label = label;
	gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 2, 3);
	filechooser = gtk_file_chooser_button_new ("Select a file", GTK_FILE_CHOOSER_ACTION_OPEN);
	widgets.filechooser = filechooser;
	if (curr_filename) {
		if (g_path_is_absolute (curr_filename)) {
			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (filechooser), curr_filename);
		} else {
			fullpath = g_build_filename (project_path, curr_filename, NULL);
			gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (filechooser), fullpath);
			g_free (fullpath);
		}
		g_free (curr_filename);
	} else {
		gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (filechooser), project_path);
	}
	gtk_table_attach_defaults (GTK_TABLE (table), filechooser, 1, 2, 2, 3);
//TODO: WILL THIS WORK? (attaching two widgets to the same place in a table)
	command = gtk_entry_new ();
	gtk_entry_set_text (GTK_ENTRY (command), curr_command);
	gtk_table_attach_defaults (GTK_TABLE (table), command, 1, 2, 2, 3);
	widgets.command = command;

	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), table);
	gtk_widget_show_all (table);
	callback_action_changed (GTK_COMBO_BOX (actionlist), (gpointer)&widgets);

	if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_OK) {
		gtk_widget_destroy (dialog);
		return FALSE;
	}

	//set stuff
	curr_action = gtk_combo_box_get_active (GTK_COMBO_BOX (actionlist));
	curr_type = gtk_combo_box_get_active (GTK_COMBO_BOX (typelist));
	curr_filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (filechooser));
	curr_command = gtk_entry_get_text (GTK_ENTRY (command));

	if (curr_action != BUILD_COMMAND && !curr_filename) {
		show_error ("You must choose a file!", NULL);
		gtk_widget_destroy (dialog);
		return FALSE;
	} else if (curr_action == BUILD_COMMAND && !curr_command) {
		show_error ("You must enter a command!", NULL);
		gtk_widget_destroy (dialog);
		return FALSE;
	}

	//if the filename is in the project, then store it as a relative path
	if (curr_filename) {
		project_path_len = strlen (project_path);
		if (!strncmp (curr_filename, project_path, project_path_len)) {
			temp_filename = curr_filename + project_path_len;
			if (G_IS_DIR_SEPARATOR (*temp_filename))
				temp_filename++;
			temp_filename = g_strdup (temp_filename);
			g_free (curr_filename);
			curr_filename = temp_filename;
		}
	}

	text = get_action_text (curr_action, curr_type, curr_filename, curr_command);
	gtk_list_store_set (build_step_list, treeiter, 0, text, 1, curr_action, 2, curr_type,
	                    3, curr_filename, 4, curr_command, -1);
	g_free (text);
	if (curr_filename) g_free (curr_filename);

	gtk_widget_destroy (dialog);
	return TRUE;
}


void callback_add_build_step (GtkWidget *widget, gpointer treeview) {
	GtkTreeIter treeiter;
	GtkListStore *build_step_list;

	build_step_list = GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (treeview)));
	gtk_list_store_append (GTK_LIST_STORE (build_step_list), &treeiter);
	if (!edit_build_step (build_step_list, &treeiter))
		gtk_list_store_remove (build_step_list, &treeiter);
	else
		gtk_tree_selection_select_iter (gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview)),
		                                &treeiter);
}


void callback_remove_build_step (GtkWidget *widget, gpointer treeview) {
	GtkTreeModel *build_step_list;
	GtkTreeIter treeiter;

	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview)),
	                                     &build_step_list, &treeiter)) {
		gtk_list_store_remove (GTK_LIST_STORE (build_step_list), &treeiter);
	}
}


void callback_edit_build_step (GtkWidget *widget, gpointer treeview) {
	GtkTreeModel *build_step_list;
	GtkTreeIter treeiter;

	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview)),
	                                     &build_step_list, &treeiter)) {
		edit_build_step (GTK_LIST_STORE (build_step_list), &treeiter);
	}
}


void callback_move_build_step_up (GtkWidget *widget, gpointer treeview) {
	GtkTreeModel *build_step_list;
	GtkTreeIter treeiter, newtreeiter;
	gchar *curr_text, *curr_filename, *curr_command;
	gint curr_action, curr_type;
	GtkTreePath *pointless_tree_path;

	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview)),
	                                     &build_step_list, &treeiter)) {

		//fuckin' gtk_list_store_swap messes with the order of other rows or does something strange,
		// so I've got to write my own here

		// I FUCKING HATE THIS TREE MODEL SHIT!!!  AGGGHHHHHH!!!!!
		pointless_tree_path = gtk_tree_model_get_path (build_step_list, &treeiter);
		if (gtk_tree_path_prev (pointless_tree_path)) {
			gtk_tree_model_get (build_step_list, &treeiter, 0, &curr_text, 1, &curr_action,
		    	                2, &curr_type, 3, &curr_filename, 4, &curr_command, -1);

			gtk_tree_model_get_iter (build_step_list, &treeiter, pointless_tree_path);
			gtk_list_store_insert_before (GTK_LIST_STORE (build_step_list), &newtreeiter, &treeiter);

			gtk_list_store_set (GTK_LIST_STORE (build_step_list), &newtreeiter, 0, curr_text,
			                    1, curr_action, 2, curr_type, 3, curr_filename, 4, curr_command, -1);
			treeiter = newtreeiter;
			gtk_tree_model_iter_next (build_step_list, &newtreeiter);
			gtk_tree_model_iter_next (build_step_list, &newtreeiter);
			gtk_list_store_remove (GTK_LIST_STORE (build_step_list), &newtreeiter);
			gtk_tree_selection_select_iter (gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview)),
			                                &treeiter);

			g_free (curr_text);
			if (curr_filename) g_free (curr_filename);
			if (curr_command) g_free (curr_command);
		}
		//good riddance to you, pointless tree path!
		gtk_tree_path_free (pointless_tree_path);
	}
}


void callback_move_build_step_down (GtkWidget *widget, gpointer treeview) {
	GtkTreeModel *build_step_list;
	GtkTreeIter treeiter, newtreeiter;
	gchar *curr_text, *curr_filename, *curr_command;
	gint curr_action, curr_type;

	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview)),
	                                     &build_step_list, &treeiter)) {

		gtk_tree_model_get (build_step_list, &treeiter, 0, &curr_text, 1, &curr_action,
		                    2, &curr_type, 3, &curr_filename, 4, &curr_command, -1);

		if (gtk_list_store_remove (GTK_LIST_STORE (build_step_list), &treeiter))
			gtk_list_store_insert_after (GTK_LIST_STORE (build_step_list), &newtreeiter, &treeiter);
		else
			//ugh... WHY WHY WHY?????
			gtk_list_store_append (GTK_LIST_STORE (build_step_list), &newtreeiter);

		gtk_list_store_set (GTK_LIST_STORE (build_step_list), &newtreeiter, 0, curr_text,
		                    1, curr_action, 2, curr_type, 3, curr_filename, 4, curr_command, -1);
		gtk_tree_selection_select_iter (gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview)),
		                                &newtreeiter);

		g_free (curr_text);
		if (curr_filename) g_free (curr_filename);
		if (curr_command) g_free (curr_command);
	}
}


/*
 * Callback for "project options" menu item,
 * shows options dialog (what else?)
 */

void callback_show_project_options (GtkWidget *widget, gpointer data) {
	GtkWidget *dialog, *hbox, *label, *list, *buttonbox, *vbox, *viewport, *button;
	GtkListStore *build_step_list;
	build_step *curr_step;
	GtkCellRenderer *textrenderer;
	GtkTreeViewColumn *column;
	GtkTreePath *path;
	GtkTreeIter treeiter;
	gchar *curr_string, *curr_filename, *curr_command;
	gint curr_action, curr_type;

	//create the dialog
	dialog = gtk_dialog_new_with_buttons ("Project options", GTK_WINDOW (window), GTK_DIALOG_MODAL,
	                                      GTK_STOCK_OK, GTK_RESPONSE_OK, GTK_STOCK_CANCEL,
	                                      GTK_RESPONSE_CANCEL, NULL);

	vbox = GTK_DIALOG (dialog)->vbox;
	gtk_box_set_spacing (GTK_BOX (vbox), 6);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 6);

	label = gtk_label_new ("Build sequence:");
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);

	build_step_list = gtk_list_store_new (5, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING);
	list = gtk_tree_view_new_with_model (GTK_TREE_MODEL (build_step_list));
	textrenderer = gtk_cell_renderer_text_new ();
	column = gtk_tree_view_column_new_with_attributes ("", textrenderer, "text", 0, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (list), column);
	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (list), FALSE);
	gtk_tree_selection_set_mode (gtk_tree_view_get_selection (GTK_TREE_VIEW (list)),
	                             GTK_SELECTION_BROWSE);
	path = gtk_tree_path_new_first ();
	gtk_tree_selection_select_path (gtk_tree_view_get_selection (GTK_TREE_VIEW (list)),
	                                path);

	hbox = gtk_hbox_new (FALSE, 6);
	gtk_tree_path_free (path);
	viewport = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (viewport), GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (viewport), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (viewport), list);
	gtk_box_pack_start (GTK_BOX (hbox), viewport, TRUE, TRUE, 0);
	buttonbox = gtk_vbutton_box_new ();
	gtk_button_box_set_layout (GTK_BUTTON_BOX (buttonbox), GTK_BUTTONBOX_START);
	button = gtk_button_new_from_stock (GTK_STOCK_ADD);
	gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
	g_signal_connect (button, "clicked", G_CALLBACK (callback_add_build_step), list);
	gtk_container_add (GTK_CONTAINER (buttonbox), button);
	button = gtk_button_new_from_stock (GTK_STOCK_REMOVE);
	gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
	g_signal_connect (button, "clicked", G_CALLBACK (callback_remove_build_step), list);
	gtk_container_add (GTK_CONTAINER (buttonbox), button);
	button = gtk_button_new_from_stock (GTK_STOCK_EDIT);
	gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
	g_signal_connect (button, "clicked", G_CALLBACK (callback_edit_build_step), list);
	gtk_container_add (GTK_CONTAINER (buttonbox), button);
	button = gtk_button_new_with_label ("Move up");
	gtk_button_set_image (GTK_BUTTON (button),
	                      gtk_image_new_from_stock (GTK_STOCK_GO_UP, GTK_ICON_SIZE_BUTTON));
	gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
	g_signal_connect (button, "clicked", G_CALLBACK (callback_move_build_step_up), list);
	gtk_container_add (GTK_CONTAINER (buttonbox), button);
	button = gtk_button_new_with_label ("Move down");
	gtk_button_set_image (GTK_BUTTON (button),
	                      gtk_image_new_from_stock (GTK_STOCK_GO_DOWN, GTK_ICON_SIZE_BUTTON));
	gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
	g_signal_connect (button, "clicked", G_CALLBACK (callback_move_build_step_down), list);
	gtk_container_add (GTK_CONTAINER (buttonbox), button);
	gtk_box_pack_start (GTK_BOX (hbox), buttonbox, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);

	//insert each build step into the list model
	for (curr_step = build_sequence; curr_step; curr_step = curr_step->next) {
		curr_string = get_action_text (curr_step->action, curr_step->output_type,
		                               curr_step->filename, curr_step->command);
		gtk_list_store_insert_with_values (build_step_list, NULL, -1, 0, curr_string,
		                                   1, (gint) curr_step->action, 2, curr_step->output_type,
		                                   3, curr_step->filename, 4, curr_step->command, -1);
		g_free (curr_string);
	}

	gtk_widget_show_all (vbox);
	if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_CANCEL) {
		gtk_widget_destroy (dialog);
		return;
	}

	//save the new build sequence (after freeing the old one)
	while (build_sequence) {
		g_free (build_sequence->filename);
		curr_step = build_sequence->next;
		g_slice_free (build_step, build_sequence);
		build_sequence = curr_step;
	}

	build_sequence = build_sequence_tail = NULL;
	if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL (build_step_list), &treeiter)) {
		do {
			curr_step = g_slice_new (build_step);
			gtk_tree_model_get (GTK_TREE_MODEL (build_step_list), &treeiter, 1, &curr_action,
			                    2, &curr_type, 3, &curr_filename, 4, &curr_command, -1);
			curr_step->action = curr_action;
			curr_step->output_type = (build_action) curr_type;
			curr_step->filename = curr_filename;
			curr_step->command = curr_command;
			curr_step->prev = build_sequence_tail;
			curr_step->next = NULL;
			if (build_sequence_tail)
				build_sequence_tail->next = curr_step;
			if (!build_sequence)
				build_sequence = curr_step;
			build_sequence_tail = curr_step;
		} while (gtk_tree_model_iter_next (GTK_TREE_MODEL (build_step_list), &treeiter));
	}

	project_changed = TRUE;
	gtk_widget_destroy (dialog);
}


/*
 * Loads project options
 */

void load_project_options () {
	FILE *project_file;
	gchar buffer[300], *curr_filename, *curr_command;
	build_action curr_action;
	gint curr_type, len;
	build_step *curr_step;

	//get the contents of the project file
	project_file = fopen (full_project_path, "r");
	if (!project_file) {
		show_error ("Couldn't open the project file", "Make sure you have read permissions");
		return;
	}

	//now read in the build sequence
	build_sequence = build_sequence_tail = NULL;
	while (fgets (buffer, 300, project_file)) {
		//take out any newline at the end
		len = strlen (buffer);
		if (len > 0 && buffer[len - 1] == '\n') {
			buffer[len - 1] = '\0';
			len--;
		}
		//...just in case someone's fucked it up by saving from notepad or some shit like that
		if (len > 0 && buffer[len - 1] == '\r') {
			buffer[len - 1] = '\0';
			len--;
		}

		//ignore empty lines
		if (buffer[0] == '\0') continue;

		//get the action and handle the rest of the line depending on that
		switch (buffer[0]) {
			case 'B':
				curr_action = BUILD_NORMAL;
				if (len < 5) continue;

				if (isalpha (buffer[2])) curr_type = buffer[2] - 'A' + 10;
				else if (isdigit (buffer[2])) curr_type = buffer[2] - '0';
				else continue;

				curr_filename = g_strdup (&buffer[4]);
				curr_command = NULL;
				break;
			case 'S':
				curr_action = BUILD_SYMTABLE;
				if (len < 3) continue;

				curr_filename = g_strdup (&buffer[2]);
				curr_command = NULL;
				break;
			case 'E':
				curr_action = BUILD_COMMAND;
				if (len < 3) continue;

				curr_command = g_strdup (&buffer[2]);
				curr_filename = NULL;
				break;
			default:
				//if it's unrecognized, skip this line
				continue;
		}

		//now that we've got all the info, create the build step
		curr_step = g_slice_new (build_step);
		curr_step->filename = curr_filename;
		curr_step->command = curr_command;
		curr_step->action = curr_action;
		curr_step->output_type = curr_type;
		//and stick it on the end
		curr_step->prev = build_sequence_tail;
		curr_step->next = NULL;
		if (build_sequence_tail)
			build_sequence_tail->next = curr_step;
		if (!build_sequence)
			build_sequence = curr_step;
		build_sequence_tail = curr_step;
	}

	fclose (project_file);
}


/*
 * Opens project from filename
 */

void open_project (gchar *filename) {
	gchar *fullpath;

	//first see if another project is open already
	if (project_open && !close_project ())
		//if it is, close it
		return;

	full_project_path = g_strdup (filename);
	//then, get the project's root directory
	project_path = g_path_get_dirname (filename);
	//and show the sidebar
	update_project_view (TRUE);
	//load the project options too
	load_project_options ();
	project_changed = FALSE;

	//and add the project to the list of recent files
	fullpath = g_filename_to_uri (filename, NULL, NULL);
	gtk_recent_manager_add_item (recentfiles, fullpath);
	g_free (fullpath);
}

