/*
	Wabbitcode - An IDE for the Z80 line of TI calculators
	Copyright (C) 2007 Donald Straney

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "asmide.h"

//name of the config file stored in the user's home
#define CONFIG_FILE ".wabbitcode"
//default settings
#define DEFAULT_WINDOW_WIDTH 800
#define DEFAULT_WINDOW_HEIGHT 800
#define DEFAULT_NUM_RECENT_FILES 5
#define DEFAULT_RESULT_COLOR 0xFFFF00
#define DEFAULT_LABEL_COLOR 0x0000FF
#define DEFAULT_COMMENT_COLOR 0x207C47
#define DEFAULT_DIRECTIVE_COLOR 0xA020F0
#define DEFAULT_OPCODE_COLOR 0x603FCD
#define DEFAULT_REGISTER_COLOR 0x3F7CBC
#define DEFAULT_STRING_COLOR 0x289CF8
#define DEFAULT_TAB_WIDTH 4
#define DEFAULT_WORD_WRAP TRUE
#define DEFAULT_AUTO_INDENT TRUE
#define DEFAULT_EDITOR_FONT "Monospace 10"
#define DEFAULT_UNDO_STACK_LEN 50
#define DEFAULT_LABELS_CASE_SENSITIVE TRUE
#define DEFAULT_OUTPUT_TYPE 0
#define DEFAULT_SAVE_BEFORE_BUILD FALSE


/*
 * Loads settings from user's
 * config file, sets global vars
 */

void load_settings () {
	gchar *path, *value;
	FILE *configfile;
	char linebuf[200];
	int lineend;

	//first set default values
	include_dirs = gtk_list_store_new (1, G_TYPE_STRING);
	winwidth = DEFAULT_WINDOW_WIDTH; winheight = DEFAULT_WINDOW_HEIGHT;
	numrecentfiles = DEFAULT_NUM_RECENT_FILES;
	INIT_COLOR (resultcolor, DEFAULT_RESULT_COLOR);
	INIT_COLOR (labelcolor, DEFAULT_LABEL_COLOR);
	INIT_COLOR (commentcolor, DEFAULT_COMMENT_COLOR);
	INIT_COLOR (dircolor, DEFAULT_DIRECTIVE_COLOR);
	INIT_COLOR (opcodecolor, DEFAULT_OPCODE_COLOR);
	INIT_COLOR (strcolor, DEFAULT_STRING_COLOR);
	INIT_COLOR (regcolor, DEFAULT_REGISTER_COLOR);
	tabwidth = DEFAULT_TAB_WIDTH;
	wordwrap = DEFAULT_WORD_WRAP;
	autoindent = DEFAULT_AUTO_INDENT;
	editorfont = pango_font_description_from_string (DEFAULT_EDITOR_FONT);
	dosyntaxhlt = labelhlt = commenthlt = dirhlt = strhlt = opcodehlt = reghlt = TRUE;
	labels_casesensitive = DEFAULT_LABELS_CASE_SENSITIVE;
	output_type = DEFAULT_OUTPUT_TYPE;
	max_undo_stack_len = DEFAULT_UNDO_STACK_LEN;
	save_before_build = DEFAULT_SAVE_BEFORE_BUILD;

	//now get the path of our config file for this user
	path = g_build_filename (g_get_home_dir (), CONFIG_FILE, NULL);

	//try to open it
	configfile = fopen (path, "r");
	g_free (path);
	if (!configfile) {
		g_printerr ("Couldn't open config file, using defaults");
		return;
	}

	//now go through the file line-by-line
	while (!feof (configfile)) {
		fgets (linebuf, 199, configfile);
		lineend = strlen (linebuf) - 1;
		if (linebuf[lineend] == '\n')
			linebuf[lineend] = '\0';
		//for each line, separate it into the key and value
		value = strchr (linebuf, '=');
		//skip lines that don't match
		if (!value)
			continue;
		*value = '\0'; value++;
		//now check against recognized keys
		if (!strcmp (linebuf, "ResultColor"))
			gdk_color_parse ((const gchar *)value, &resultcolor);
		else if (!strcmp (linebuf, "LabelColor"))
			gdk_color_parse ((const gchar *)value, &labelcolor);
		else if (!strcmp (linebuf, "CommentColor"))
			gdk_color_parse ((const gchar *)value, &commentcolor);
		else if (!strcmp (linebuf, "DirectiveColor"))
			gdk_color_parse ((const gchar *)value, &dircolor);
		else if (!strcmp (linebuf, "OpcodeColor"))
			gdk_color_parse ((const gchar *)value, &opcodecolor);
		else if (!strcmp (linebuf, "StringColor"))
			gdk_color_parse ((const gchar *)value, &strcolor);
		else if (!strcmp (linebuf, "RegisterColor"))
			gdk_color_parse ((const gchar *)value, &regcolor);
		else if (!strcmp (linebuf, "WindowWidth"))
			winwidth = atoi (value);
		else if (!strcmp (linebuf, "WindowHeight"))
			winheight = atoi (value);
		else if (!strcmp (linebuf, "NumRecentFiles"))
			numrecentfiles = atoi (value);
		else if (!strcmp (linebuf, "TabWidth"))
			tabwidth = atoi (value);
		else if (!strcmp (linebuf, "WordWrap"))
			wordwrap = (gboolean)atoi (value);
		else if (!strcmp (linebuf, "AutoIndent"))
			autoindent = (gboolean)atoi (value);
		else if (!strcmp (linebuf, "FontFamily"))
			pango_font_description_set_family (editorfont, value);
		else if (!strcmp (linebuf, "FontStyle"))
			pango_font_description_set_style (editorfont, (PangoStyle)atoi (value));
		else if (!strcmp (linebuf, "FontWeight"))
			pango_font_description_set_weight (editorfont, (PangoWeight)atoi (value));
		else if (!strcmp (linebuf, "FontSize"))
			pango_font_description_set_size (editorfont, atoi (value));
		else if (!strcmp (linebuf, "SyntaxHlt"))
			dosyntaxhlt = (gboolean)atoi (value);
		else if (!strcmp (linebuf, "LabelHlt"))
			labelhlt = (gboolean)atoi (value);
		else if (!strcmp (linebuf, "CommentHlt"))
			commenthlt = (gboolean)atoi (value);
		else if (!strcmp (linebuf, "DirectiveHlt"))
			dirhlt = (gboolean)atoi (value);
		else if (!strcmp (linebuf, "StringHlt"))
			strhlt = (gboolean)atoi (value);
		else if (!strcmp (linebuf, "OpcodeHlt"))
			opcodehlt = (gboolean)atoi (value);
		else if (!strcmp (linebuf, "RegisterHlt"))
			reghlt = (gboolean)atoi (value);
		else if (!strcmp (linebuf, "UndoStackLen"))
			max_undo_stack_len = atoi (value);
		else if (!strcmp (linebuf, "LabelsCaseSensitive"))
			labels_casesensitive = (gboolean)atoi (value);
		else if (!strcmp (linebuf, "IncludeDir"))
			gtk_list_store_insert_with_values (include_dirs, NULL, -1, 0, g_strdup (value), -1);
		else if (!strcmp (linebuf, "OutputType"))
			output_type = atoi (value);
		else if (!strcmp (linebuf, "SaveBeforeBuild"))
			save_before_build = (gboolean)atoi (value);
	}
	fclose (configfile);

	//sanity checks
	if (tabwidth < 1 || tabwidth > 20)
		tabwidth = DEFAULT_TAB_WIDTH;
	if (max_undo_stack_len < 5 || max_undo_stack_len > 200)
		max_undo_stack_len = DEFAULT_UNDO_STACK_LEN;
	if (output_type >= NUM_OUTPUT_TYPES || output_type < 0)
		output_type = DEFAULT_OUTPUT_TYPE;
	//I hope no one has their resolution set to >3000x3000...
	if (winwidth < 50 || winwidth > 3000)
		winwidth = DEFAULT_WINDOW_WIDTH;
	if (winheight < 50 || winheight > 3000)
		winheight = DEFAULT_WINDOW_HEIGHT;
}


/*
 * Saves settings from global vars
 * into user's config file
 */

void save_settings () {
	gchar *path;
	FILE *configfile;
	GtkTreeIter iter;
	gboolean more_dirs;

	//get the path of our config file for this user
	path = g_build_filename (g_get_home_dir (), CONFIG_FILE, NULL);
	//and try to open it
	configfile = fopen (path, "w");
	g_free (path);
	if (!configfile) {
		show_error ("Couldn't save configuration",
			"This probably means your\nsystem has serious problems");
		return;
	}

	//then write all the settings to it
	fprintf (configfile, "WindowWidth=%d\n", winwidth);
	fprintf (configfile, "WindowHeight=%d\n", winheight);
	fprintf (configfile, "NumRecentFiles=%d\n", numrecentfiles);
	fprintf (configfile, "TabWidth=%d\n", tabwidth);
	fprintf (configfile, "WordWrap=%d\n", (int)wordwrap);
	fprintf (configfile, "ResultColor=#%.2X%.2X%.2X\n", (resultcolor.red >> 8) & 0xFF,
		(resultcolor.green >> 8) & 0xFF, (resultcolor.blue >> 8) & 0xFF);
	fprintf (configfile, "LabelColor=#%.2X%.2X%.2X\n", (labelcolor.red >> 8) & 0xFF,
		(labelcolor.green >> 8) & 0xFF, (labelcolor.blue >> 8) & 0xFF);
	fprintf (configfile, "CommentColor=#%.2X%.2X%.2X\n", (commentcolor.red >> 8) & 0xFF,
		(commentcolor.green >> 8) & 0xFF, (commentcolor.blue >> 8) & 0xFF);
	fprintf (configfile, "DirectiveColor=#%.2X%.2X%.2X\n", (dircolor.red >> 8) & 0xFF,
		(dircolor.green >> 8) & 0xFF, (dircolor.blue >> 8) & 0xFF);
	fprintf (configfile, "StringColor=#%.2X%.2X%.2X\n", (strcolor.red >> 8) & 0xFF,
		(strcolor.green >> 8) & 0xFF, (strcolor.blue >> 8) & 0xFF);
	fprintf (configfile, "OpcodeColor=#%.2X%.2X%.2X\n", (opcodecolor.red >> 8) & 0xFF,
		(opcodecolor.green >> 8) & 0xFF, (opcodecolor.blue >> 8) & 0xFF);
	fprintf (configfile, "RegisterColor=#%.2X%.2X%.2X\n", (regcolor.red >> 8) & 0xFF,
		(regcolor.green >> 8) & 0xFF, (regcolor.blue >> 8) & 0xFF);
	fprintf (configfile, "AutoIndent=%d\n", (int)autoindent);
	fprintf (configfile, "FontFamily=%s\n", pango_font_description_get_family (editorfont));
	fprintf (configfile, "FontStyle=%d\n", (int)pango_font_description_get_style (editorfont));
	fprintf (configfile, "FontWeight=%d\n", (int)pango_font_description_get_weight (editorfont));
	fprintf (configfile, "FontSize=%d\n", pango_font_description_get_size (editorfont));
	fprintf (configfile, "SyntaxHlt=%d\n", (int)dosyntaxhlt);
	fprintf (configfile, "LabelHlt=%d\n", (int)labelhlt);
	fprintf (configfile, "CommentHlt=%d\n", (int)commenthlt);
	fprintf (configfile, "DirectiveHlt=%d\n", (int)dirhlt);
	fprintf (configfile, "StringHlt=%d\n", (int)strhlt);
	fprintf (configfile, "OpcodeHlt=%d\n", (int)opcodehlt);
	fprintf (configfile, "RegisterHlt=%d\n", (int)reghlt);
	fprintf (configfile, "UndoStackLen=%d\n", max_undo_stack_len);
	fprintf (configfile, "LabelsCaseSensitive=%d\n", (int)labels_casesensitive);
	fprintf (configfile, "OutputType=%d\n", output_type);
	fprintf (configfile, "SaveBeforeBuild=%d\n", (int)save_before_build);

	//write a line for each include dir too
	more_dirs = gtk_tree_model_get_iter_first (GTK_TREE_MODEL (include_dirs), &iter);
	while (more_dirs) {
		gchar *curr_dir;
		gtk_tree_model_get (GTK_TREE_MODEL (include_dirs), &iter, 0, &curr_dir, -1);
		fprintf (configfile, "IncludeDir=%s\n", curr_dir);
		g_free (curr_dir);
		more_dirs = gtk_tree_model_iter_next (GTK_TREE_MODEL (include_dirs), &iter);
	}

	fclose (configfile);
}
