#define __MAIN_C

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "pass_one.h"
#include "pass_two.h"
#include "storage.h"
#include "spasm.h"
#include "utils.h"
#include "big.h"
#include "console.h"

#define LISTING_BUF_SIZE 65536	//initial size of buffer for output listing

void write_file (const unsigned char *, int, const char *);

int main (int argc, char **argv) {

	int curr_arg = 0;
	bool case_sensitive = false;
	char *output_filename, *input_contents;

	use_colors = true;
	exit_code = EXIT_NORMAL;
	save_console_attributes ();
	atexit (restore_console_attributes);
	stats_codesize = stats_datasize = stats_mintime = stats_maxtime = 0;
	
	//if there aren't enough args, show info
	if (argc < 2) {
		puts ("Z80 Assembler by Don Straney and Spencer Putt\n\nspasm [options] <input file> <output file>\n");
		puts ("Options:\n-T = Generate code listing\n-C = Code counter mode\n-L = Symbol table mode\n-S = Stats mode\n-O = Don't write to output file");
		puts ("-I [directory] = Add include directory\n-A = Labels are cAse-sensitive\n-D<name>[=value] = Create a define 'name' [with 'value']");
		puts ("-N = Don't use colors for messages");
		return EXIT_NORMAL;
	}

	//init stuff
	mode = MODE_NORMAL;
	in_macro = 0;
	init_storage();
	
	//otherwise, get any options
	curr_arg = 1;
	curr_input_file = output_filename = NULL;

	while (curr_arg < argc) {
		if (argv[curr_arg][0] == '-') switch (argv[curr_arg][1]) {
			//args for different modes
			case 'O':
				mode = mode & (~MODE_NORMAL);
				break;
			case 'T':
				mode |= MODE_LIST;
				listing_on = true;
				break;
			case 'C':
				mode = MODE_CODE_COUNTER;
				break;
			case 'L':
				mode = MODE_SYMTABLE;
				break;
			case 'S':
				mode = MODE_STATS;
				break;
			//handle no-colors flag
			case 'N':
				use_colors = false;
				break;
			//handle include files too
			case 'I':
			{
				char *dir;
				//make sure there's another argument after it for the include path
				if (strlen(argv[curr_arg]) > 2) {
					dir = strdup (&argv[curr_arg][2]);
				} else {
					if (curr_arg >= argc - 1) {
						printf ("%s used without include path\n", argv[curr_arg]);
						break;
					}
					
					dir = strdup (argv[++curr_arg]);
				}
				
				char *p;
				for (p = strtok (dir, ";,"); p; p = strtok (NULL, ";,")) {
					include_dirs = list_append (include_dirs, strdup(p));
				}

				break;
			}
			//and the case-sensitive flag
			case 'A':
				case_sensitive = true;
				break;
			//handle adding defines
			case 'D':
			{
				char *ptr;
				if (strlen (argv[curr_arg]) > 2) {
					ptr = &argv[curr_arg][2];
				} else {
					if (curr_arg >= argc - 1) {
						printf ("%s used without define name", argv[curr_arg]);
						break;
					}
					
					ptr = argv[++curr_arg];
				}
				
				char name[256];
				define_t *define;
				read_expr (&ptr, name, "=");

				define = add_define (strdup (name), NULL);
				if (*skip_whitespace (ptr) != '\0')
					define->contents = strdup (ptr);
				break;
			}
			default:
				printf ("Unrecognized option %s\n", argv[curr_arg]);

		} else {
			//if it's not a flag, then it must be a filename
			if (curr_input_file)
				output_filename = argv[curr_arg];
			else
				curr_input_file = argv[curr_arg];
		}
		curr_arg++;
	}

	set_case_sensitive (case_sensitive);
	
	//check on filenames
	if (!curr_input_file) {
		puts ("No input file specified");
		return EXIT_FATAL_ERROR;
	}

	if (!output_filename) {
		if (mode & MODE_SYMTABLE)
			output_filename = change_extension (curr_input_file, "lab");
		else
			output_filename = change_extension (curr_input_file, "bin");
	}

	//read in the input file
	input_contents = get_file_contents (curr_input_file);

	if (!input_contents) {
		puts ("Couldn't open input file");
		return EXIT_FATAL_ERROR;
	}

	//and create the output buffer
	if (mode & MODE_NORMAL) {
		output_contents = malloc_chk (OUTPUT_BUF_SIZE);
		out_ptr = output_contents;
	}
	//along with the listing buffer, if required
	if (mode & MODE_LIST) {
		listing_buf = eb_init (LISTING_BUF_SIZE);
		listing_offset = 0;
	}

	set_console_attributes (COLOR_WHITE);
	printf ("Pass one... \n");
	run_first_pass (input_contents);

	//free input file and include dirs when done
	free (input_contents);
	while (include_dirs) {
		list *old_link = include_dirs->next;
		free (include_dirs->data);
		list_free (include_dirs);
		include_dirs = old_link;
	}

	//if there was an error in the first pass, don't bother with the second pass
	if (error_occurred)
		return exit_code;

	//...and if there's output, run the second pass and write it to the output file
	if (mode & MODE_NORMAL || mode & MODE_LIST) {

		set_console_attributes (COLOR_WHITE);
		printf ("Pass two... \n");

		run_second_pass ();

		set_console_attributes (COLOR_WHITE);
		printf ("Done\n");
		
		//run the output through the appropriate program export and write it to a file
		if (mode & MODE_NORMAL) {
			write_file (output_contents, out_ptr - output_contents, output_filename);
			free (output_contents);

			if (error_occurred)
				return exit_code;
		}

		//write the listing file if necessary
		if (mode & MODE_LIST) {
			FILE *file;
			char *name;

			//get file name
			name = change_extension (output_filename, "lst");
			file = fopen (name, "w");
			if (!file) {
				printf ("Couldn't open listing file '%s'", name);
				free (name);
				return EXIT_FATAL_ERROR;
			}
			free (name);
			
			if (fwrite (listing_buf->start, 1, listing_offset, file) != listing_offset) {
				puts ("Error writing to listing file");
				fclose (file);
				return EXIT_FATAL_ERROR;
			}

			fclose (file);
		}
		
		//free the output buffer and all the names of input files
		while (input_files) {
			list *old_link = input_files;
			input_files = list_remove (input_files, input_files);
			free (old_link->data);
			list_free (old_link);
		}
	}

	//if there's info to be dumped, do that
	if (mode & MODE_CODE_COUNTER) {
		fprintf (stderr, "Size: %u\nMin. execution time: %u\nMax. execution time: %u\n",
		         stats_codesize, stats_mintime, stats_maxtime);
	}
	
	if (mode & MODE_SYMTABLE) {
		write_labels (output_filename);
	}

	if (mode & MODE_STATS) {
		fprintf (stderr, "Number of labels: %u\nNumber of defines: %u\nCode size: %u\nData size: %u\nTotal size: %u\n",
		         get_num_labels (), get_num_defines (), stats_codesize, stats_datasize, stats_codesize + stats_datasize);
	}

	return exit_code;
}
