/*
	Wabbitcode - An IDE for the Z80 line of TI calculators
	Copyright (C) 2007 Donald Straney

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <gtk/gtk.h>
#include <string.h>
#include "asmide.h"
#include "undo.h"

//used to keep "undone" actions from being registered themselves
gboolean ignore_actions = FALSE;


/*
 * Frees memory allocated
 * for an action
 */

void free_action (action *oldaction) {

	if (oldaction->text)
		g_free (oldaction->text);

	g_slice_free (action, oldaction);
}


/*
 * Frees entire undo stack for file
 */

void free_undo_stack (action *curraction) {
	action *nextaction;

	while (curraction) {
		nextaction = curraction->next;
		free_action (curraction);
		curraction = nextaction;
	}
}


/*
 * Adds an action to the
 * current file's undo stack,
 * text must be allocated
 * mem, not static or local
 */

void add_undo_action (action_type type, int offset, gchar* text) {
	action *newaction, *nextaction;
	int i;

	newaction = g_slice_new (action);
	newaction->type = type;
	newaction->offset = offset;
	newaction->text = text;

	//if we're not at the beginning of the stack, clear everything before the current position
	//TODO: move this to callback_start_user_action?
	if (currfile->undo_stack_pos > 0) {

		i = 0;
		while (i < currfile->undo_stack_pos) {
			if (currfile->undo_stack->type == ACTION_START)
				currfile->undo_stack_len--;

			nextaction = currfile->undo_stack->next;
			free_action (currfile->undo_stack);
			currfile->undo_stack = nextaction;
			i++;
		}
		if (currfile->undo_stack->next == NULL) {
			free_action (currfile->undo_stack);
			currfile->undo_stack = currfile->undo_stack_end = NULL;
		} else {
			currfile->undo_stack->prev = NULL;
		}
		currfile->undo_stack_pos = 0;

		set_redo_sensitive (FALSE);
		set_undo_sensitive (TRUE);
	} else if (currfile->undo_stack == NULL) {
		set_undo_sensitive (TRUE);
	}

	if (currfile->undo_stack)
		currfile->undo_stack->prev = newaction;
	newaction->prev = NULL;
	newaction->next = currfile->undo_stack;
	currfile->undo_stack = newaction;
	if (currfile->undo_stack_end == NULL)
		currfile->undo_stack_end = newaction;
}


/*
 * Called on start-user-action -
 * adds a "start" action to
 * current file's undo stack
 * to separate high-level actions
 */

void callback_start_user_action (GtkTextBuffer *widget, gpointer data) {

	if (ignore_actions)
		return;

	add_undo_action (ACTION_START, 0, NULL);

	//if there are too many actions in the undo stack, free the extras
	if (++(currfile->undo_stack_len) > max_undo_stack_len) {
		while (currfile->undo_stack_len > max_undo_stack_len) {
			if (currfile->undo_stack_end->type == ACTION_START)
				currfile->undo_stack_len--;

			currfile->undo_stack_end = currfile->undo_stack_end->prev;
			free_action (currfile->undo_stack_end->next);
		}

		while (currfile->undo_stack_end->type != ACTION_START) {
			currfile->undo_stack_end = currfile->undo_stack_end->prev;
//TEST
			if (currfile->undo_stack_end == NULL || currfile->undo_stack_end->next == NULL) {
				if (currfile->undo_stack_end == NULL)
					puts ("Ran off start of undo stack!");
				else
					puts ("Undo stack entry is missing next pointer!");

				printf ("%d %d\n", currfile->undo_stack_len, max_undo_stack_len);
				action *curr_action = currfile->undo_stack_end;
				while (curr_action) {
					if (curr_action->type == ACTION_INSERT)
						printf ("Inserted %s at %d\n", curr_action->text, curr_action->offset);
					else if (curr_action->type == ACTION_DELETE)
						printf ("Deleted %s at %d\n", curr_action->text, curr_action->offset);
					else if (curr_action->type == ACTION_START)
						puts ("Start-user-action");
					else
						puts ("Unrecognized type");

					curr_action = curr_action->prev;
				}

				return;
			}
//END TEST
			free_action (currfile->undo_stack_end->next);
		}

		currfile->undo_stack_end->next = NULL;
	}
}


/*
 * Called on end-user-action -
 * checks for an extra "start" action
 * added by a copy command and removes it
 */

void callback_end_user_action (GtkTextBuffer *widget, gpointer data) {
	action *oldaction;

	if (ignore_actions)
		return;

	/* for some reason, doing a copy on a GtkTextBuffer sends an extra
		pair of start-user-action/end-user-action signals even though
		there's nothing being done between them, so to make the undo
		work right, we've got to wait for an end-user-action and
		remove that extra "start" action on the undo stack if it's there */
	if (currfile->undo_stack && currfile->undo_stack->type == ACTION_START) {
		oldaction = currfile->undo_stack;
		currfile->undo_stack = oldaction->next;
		free_action (oldaction);
		if (currfile->undo_stack)
			currfile->undo_stack->prev = NULL;
		currfile->undo_stack_len--;
	}
}


/*
 * Undoes the last high-level action
 */

void callback_undo (GtkWidget *widget, gpointer data) {
	action *curr_action;
	int i = 0;
	GtkTextIter startiter, enditer;

	ignore_actions = TRUE;

	//enable redo
	set_redo_sensitive (TRUE);

	//find our place in the undo stack
	for (curr_action = currfile->undo_stack; i < currfile->undo_stack_pos;
		i++, curr_action = curr_action->next);

	//then undo each part of the action until we hit the end of the stack
	// or a separator between high-level actions
	while (curr_action->type != ACTION_START) {

		gtk_text_buffer_get_iter_at_offset (currfile->textbuffer, &startiter, curr_action->offset);

		if (curr_action->type == ACTION_INSERT) {
			gtk_text_buffer_get_iter_at_offset (currfile->textbuffer, &enditer,
				curr_action->offset + g_utf8_strlen (curr_action->text, -1));
			gtk_text_buffer_delete (currfile->textbuffer, &startiter, &enditer);
		} else {
			gtk_text_buffer_insert (currfile->textbuffer, &startiter,
				(const gchar *)curr_action->text, -1);
		}

		curr_action = curr_action->next;
		i++;
	}

	if (curr_action->next) {
		//skip the separator
		curr_action = curr_action->next;
		i++;
	} else {
		//if we've hit the end of the undo stack, disable undo
		set_undo_sensitive (FALSE);
	}

	currfile->undo_stack_pos = i;
	ignore_actions = FALSE;
}


/*
 * Redoes the next high-level action
 */

void callback_redo (GtkWidget *widget, gpointer data) {
	action *curr_action;
	GtkTextIter startiter, enditer;
	int i = 0;

	ignore_actions = TRUE;

	//enable undo
	set_undo_sensitive (TRUE);

	//find our place in the undo stack
	for (curr_action = currfile->undo_stack; i < currfile->undo_stack_pos;
		i++, curr_action = curr_action->next);

	if (curr_action->prev->type == ACTION_START) {
		curr_action = curr_action->prev->prev;
		i -= 2;
	} else if (curr_action->type == ACTION_START) {
		curr_action = curr_action->prev;
		i--;
	}

	//then redo each part of the action until we hit the beginning of the stack
	// or the start of the next high-level action
	while (curr_action && curr_action->type != ACTION_START) {

		gtk_text_buffer_get_iter_at_offset (currfile->textbuffer, &startiter, curr_action->offset);

		if (curr_action->type == ACTION_DELETE) {
			gtk_text_buffer_get_iter_at_offset (currfile->textbuffer, &enditer,
				curr_action->offset + g_utf8_strlen (curr_action->text, -1));
			gtk_text_buffer_delete (currfile->textbuffer, &startiter, &enditer);
		} else {
			gtk_text_buffer_insert (currfile->textbuffer, &startiter,
				(const gchar *)curr_action->text, -1);
		}

		curr_action = curr_action->prev;
		i--;
	}
	i++;

	//if we're at the beginning of the undo stack, disable redo
	if (!curr_action)
		set_redo_sensitive (FALSE);

	currfile->undo_stack_pos = i;
	ignore_actions = FALSE;
}

